/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: subr.c,v 1.4 2009/08/14 11:57:37 mickey Exp $ */

#include "vchatctl.h"

#if HAVE_PWD_H
  #include <pwd.h>
#endif

/*
 * Globals.
 */

int			sock;
struct sockaddr_un	myaddr;
struct sockaddr_un	vsaddr;

/*
 * Externals 
 */

extern	char			*prog_name;

/*
 * Initialize the local socket, we will be sending messages from.
 */

int sock_open(void)
{
	int ret = -1;

	if((sock = socket(PF_LOCAL, SOCK_DGRAM, 0)) >= 0)
	{
		/*
		 * Prepare local socket address.
		 */

		myaddr.sun_family = AF_LOCAL;
		snprintf(myaddr.sun_path, sizeof(myaddr.sun_path), "%svchatctl.%d",
			_PATH_TMP, getpid());
#ifdef HAVE_SUN_LEN
		myaddr.sun_len = SUN_LEN(&sa);
#endif
		unlink(myaddr.sun_path);

		/*
		 * Prepare remote socket address.
		 */

		vsaddr.sun_family = AF_LOCAL;
		snprintf(vsaddr.sun_path, sizeof(vsaddr.sun_path), "%svctl",
			DEFAULT_BASEDIR);
#ifdef HAVE_SUN_LEN
		vsaddr.sun_len = SUN_LEN(&vsaddr);
#endif
		/*
		 * Bind local socket.
		 */

		if(bind(sock, (struct sockaddr *) &myaddr, SUN_LEN(&myaddr)) >= 0)
		{
			chmod(myaddr.sun_path, 0666);

			ret = 0;
		}
		else
		{
			fprintf(stderr, "%s: bind failure (%s).\n", prog_name,
				strerror(errno));
		}
	}
	else
	{
		fprintf(stderr, "%s: socket failure (%s)", prog_name,
			strerror(errno));
	}

	return ret;
}

/*
 * Close our local socket.
 */

void sock_close(void)
{
	if(sock >= 0)
		close(sock);

	unlink(myaddr.sun_path);
}

void dumpvctl(VCTL *v)
{
	fprintf(stderr, "VCTL: type=$%04X flags=$%04X len=$%04X version=$%04X "
		"data:\n%s\n", v->type, v->flags, v->len, v->version,
		v->data);
}

int _sendctlmsg(VCTL *vctl)
{
	int ret = -1;

	if(vctl)
	{
		struct timeval to;
		fd_set fds;
		int sr;

		FD_ZERO(&fds);
		FD_SET(sock, &fds);

		to.tv_sec = 4;
		to.tv_usec = 0;

		/*
		 * Wait for the socket to become ready for sending data
		 * or until the timeout limit is reached, whatever happens
		 * first.
		 */

		if(!(sr = select(FD_SETSIZE, 0, &fds, 0, &to)))
			errno = ETIMEDOUT;

		if(sr == 1 && FD_ISSET(sock, &fds))
		{
			/*
			 * Send the message.
			 */

			if(sendto(sock, vctl, vctl->len, 0, (struct sockaddr *)
			&vsaddr, SUN_LEN(&vsaddr)) == vctl->len)
			{
				/* OK */

				ret = 0;
			}
			else
			{
				fprintf(stderr, "Failed to send control "
					"message to '%s' (%s).\n",
					vsaddr.sun_path, strerror(errno));
			}
		}
		else
		{
			fprintf(stderr, "Unable to send control message to "
				"'%s' (%s).\n", vsaddr.sun_path,
				strerror(errno));
		}
	}
	else
	{
		/*
		 * No message - No send.
		 */

		ret = 0;
	}

	return ret;
}

/*
 * Send a message of <type> with <flags> and accompanying
 * <data> of size <len> to the VChat conference server.
 */

int vctl_send(uint16_t type, uint16_t flags, void *data, size_t len)
{
	VCTL vctl;

	/*
	 * Prepare the control message for sending.
	 */

	bzero(&vctl, sizeof(vctl));

	if(data)
	{
		if(len > (sizeof(vctl.data) - 1))
			len = (sizeof(vctl.data) - 1);

		if(len > 0)
			bcopy(data, &vctl.data[0], len);
	}
	else
	{
		len = 0;
	}

	vctl.len	= VCTLSIZE(len);
	vctl.flags	= flags;
	vctl.type	= type;
	vctl.version	= VCTL_PROTOCOL;

	return _sendctlmsg(&vctl);
}

/*
 * Send the string specified by <fmt> and it's arguments as a message
 * of <type> with <flags> to the VChat conference server.
 */

int vctl_sendmsg(uint16_t type, uint16_t flags, char *fmt, ...)
{
	va_list ap;
	VCTL vctl;
	size_t len = 0;

	va_start(ap, fmt);

	/*
	 * Prepare the control message for sending.
	 */

	bzero(&vctl, sizeof(vctl));

	if(fmt)
		len = vsnprintf(vctl.data, sizeof(vctl.data), fmt, ap);

	vctl.len	= VCTLSIZE(len);
	vctl.flags	= flags;
	vctl.type	= type;
	vctl.version	= VCTL_PROTOCOL;

	va_end(ap);

	return _sendctlmsg(&vctl);
}

/*
 * Receive a message from the VChat conference server
 * and return a pointer to it, or NULL if receive failed.
 */

VCTL *vctl_recvmsg(void)
{
	static VCTL vctl;
	struct timeval to;
	struct sockaddr_un fsa;
	socklen_t fsalen = sizeof(fsa);
	fd_set fds;
	int sr;
	size_t msglen;

	bzero(&fsa, sizeof(fsa));
	bzero(&vctl, sizeof(vctl));

	FD_ZERO(&fds);
	FD_SET(sock, &fds);

	to.tv_sec = 4;
	to.tv_usec = 0;

	/*
	 * Wait until a message is waiting on the socket or
	 * the timeout has expired, whatever happens first.
	 */

	if(!(sr = select(FD_SETSIZE, &fds, 0, 0, &to)))
		errno = ETIMEDOUT;

	if(sr == 1 && FD_ISSET(sock, &fds))
	{
		/*
		 * Receive the message.
		 */

		if((msglen = recvfrom(sock, &vctl, sizeof(vctl), 0, (struct sockaddr *)&fsa, &fsalen)) >= 0)
		{
			vctl.data[sizeof(vctl.data)-1] = 0;

			if(msglen >= VCTLSIZE(0))
			{
				if(vctl.version == VCTL_PROTOCOL)
				{
					if(vctl.len == msglen)
					{
						return &vctl;
					}
					else
					{
						fprintf(stderr, "%s: Received "
							"truncated message "
							"($%04X/$%04X bytes)."
							"\n", pqpath, msglen,
							vctl.len);
					}
				}
				else
				{
					fprintf(stderr, "%s: Protocol version "
						"mismatch on received message "
						"($%04X/$%04X).\n",pqpath,
						vctl.version, VCTL_PROTOCOL);
				}
			}
			else
			{
				fprintf(stderr, "%s: Received message of "
					"invalid length ($%04X/$%04X).\n",
					pqpath, msglen, VCTLSIZE(0));
			}
		}
		else
		{
			fprintf(stderr, "%s: Failed to receive message "
				"(%s).\n", pqpath, strerror(errno));
		}
	}
	else
	{
		fprintf(stderr, "%s: Unable to receive message (%s)\n",
			pqpath, strerror(errno));
	}

	return (VCTL *)0;


}

/*
 * Wait for messages of type TEXT to arrive and show them
 * on the users terminal.
 */

void vctl_showreply(void)
{
	VCTL *vctl;

	for( ; (vctl = vctl_recvmsg()) ; )
	{
		if(vctl->type == VCTL_MSG_TEXT)
			printf("%s\n", vctl->data);

		if(vctl->flags & VCTL_FLG_MSGCT)
			break;
	}
}

/*
 * Check whether a nickname is valid, that is it does not contain
 * any invalid characters and does not exceed the maximum length
 * or is empty. Returns 1 if nickname is valid, 0 otherwise.
 *
 * XXX: This function also exists in vserver and should be shared.
 */

int nick_ok(register char *nick)
{
	register size_t len;

	if(!nick || !*nick)
		return 0;

	for(len = 0; *nick; nick++, len++)
	{
		if(!strchr(NICKVALIDCHARS, *nick) || len > VPNICKSIZE)
			return 0;
	}

	return 1;
}

void ulist(int32_t *channel)
{
	if(!vctl_send(VCTL_CMD_ULIST, VCTL_FLG_MSGSA, channel, channel ? sizeof(int32_t) : 0))
	{
		VCTL *vctl;
		int n;

		for(n = 0; (vctl = vctl_recvmsg()); )
		{
			if(vctl->type == VCTL_MSG_UINFO)
			{
				VCTL_UINFO *vu = (VCTL_UINFO *)&vctl->data[0];

				if(!n)
				{
					printf("NICKNAME        CHANNEL   IDLE         FROM\n");
				}

				if(vu->channel >= 0)
				{
					printf("%-15s $%08X %-12s %s\n",
						vu->nick, vu->channel,
						elapsed_now(vu->stamp), vu->host);
				}
				else
				{
					printf("%-15s %-9s %-12s %s\n",
						vu->nick, "??????",
						elapsed_now(vu->stamp), vu->host);
				}

				n++;
			}
			else if(vctl->type == VCTL_MSG_TEXT)
			{
				printf("%s%s\n", n ? "\n" : "", vctl->data);
			}

			if(vctl->flags & VCTL_FLG_MSGCT)
				break;
		}
	}
}

char *getopname(void)
{
	static char buf[64];
	struct passwd *pwe;
	uid_t uid = getuid();

	bzero(buf, sizeof(buf));

	if((pwe = getpwuid(uid)))
	{
		char *optty = NULL;

		if(isatty(0))
			optty = ttyname(0);
		else if(isatty(1))
			optty = ttyname(1);
		else if(isatty(2))
			optty = ttyname(2);

		if(optty && !strncmp(optty, "/dev/", 5))
			optty += 5;

		if(optty)
		{
			snprintf(buf, sizeof(buf), "%s@%s", pwe->pw_name,
				optty);
		}
		else
		{
			sncopy(pwe->pw_name, buf, sizeof(buf));
		}
	}

	return buf;
}

void broadcast(void)
{
	struct passwd *pwe;
	uid_t uid = getuid();

	/*
	 * Get the users name from the password database
	 */

	if((pwe = getpwuid(uid)))
	{
		char buf[1016];
		char *opname = pwe->pw_name;
		char *optty = NULL;

		/*
		 * Get the users login terminal
		 */

		if(isatty(0))
			optty = ttyname(0);
		else if(isatty(1))
			optty = ttyname(1);
		else if(isatty(2))
			optty = ttyname(2);

		if(!optty)
			optty = "<stdin>";


		if(!vctl_sendmsg(VCTL_CMD_BROADCAST, VCTL_FLG_MSGIT, "%s %s ", opname, optty))
		{
			register char *s;

			if(isatty(0))
				printf("Enter broadcast message [CTRL-D to end message]:\n");

			while(fgets(buf, sizeof(buf), stdin))
			{
				while((s = strrchr(buf, '\r')))
					*s = 0;

				if(vctl_sendmsg(VCTL_CMD_BROADCAST, 0, "%s", buf))
				{
					return;
				}
			}

			if(feof(stdin) && !ferror(stdin))
			{
				clearerr(stdin);

				if(!vctl_send(VCTL_CMD_BROADCAST, VCTL_FLG_MSGCT, NULL, 0))
				{
					vctl_showreply();
				}
			}
			else
			{
				fprintf(stderr, "%s: Read error on stdin\n",
					prog_name);
			}
		}
	}
	else
	{
		fprintf(stderr, "%s: Failed to get username for uid %d (%s)\n",
			prog_name, uid, strerror(errno));
	}
}

void killuser(char *nick, int silent)
{
	if(nick_ok(nick))
	{
		VCTL_KILL vck;

		bzero(&vck, sizeof(vck));

		sncopy(nick, vck.nick, sizeof(vck.nick));
		vck.target = VCTL_KILLTARG_USER;

		if(!silent)
		{
			sncopy(getopname(), vck.traitor, sizeof(vck.traitor));
		}

		if(!vctl_send(VCTL_CMD_KILL, VCTL_FLG_MSGSA, &vck, sizeof(vck)))
		{
			vctl_showreply();
		}
	}
	else
	{
		printf("Nickname '%s' is invalid.\n", nick);
	}
}

void killchannel(int32_t channel, int silent)
{
	VCTL_KILL vck;

	bzero(&vck, sizeof(vck));

	vck.channel = channel;
	vck.target = VCTL_KILLTARG_CHANNEL;

	if(!silent)
	{
		sncopy(getopname(), vck.traitor, sizeof(vck.traitor));
	}

	if(!vctl_send(VCTL_CMD_KILL, VCTL_FLG_MSGSA, &vck, sizeof(vck)))
	{
		vctl_showreply();
	}
}

void killall(int silent)
{
	VCTL_KILL vck;

	bzero(&vck, sizeof(vck));

	vck.target = VCTL_KILLTARG_ALL;

	if(!silent)
	{
		sncopy(getopname(), vck.traitor, sizeof(vck.traitor));
	}

	if(!vctl_send(VCTL_CMD_KILL, VCTL_FLG_MSGSA, &vck, sizeof(vck)))
	{
		vctl_showreply();
	}
}

void anonymous(uint8_t *onoff)
{
	VCTL *vctl;

	if(!vctl_send(VCTL_CMD_ANONYMOUS, VCTL_FLG_MSGSA, onoff, onoff ? 1 : 0))
	{
		for( ; (vctl = vctl_recvmsg()) ; )
		{
			if(vctl->type == VCTL_MSG_ONOFF)
			{
				if(vctl->data[0])
				{
					printf("Client addresses/hostnames are NOT visible to others.\n");
				}
				else
				{
					printf("Client addresses/hostnames are visible to others.\n");
				}
			}
			else if(vctl->type == VCTL_MSG_TEXT)
			{
				printf("%s\n", vctl->data);
			}

			if(vctl->flags & VCTL_FLG_MSGCT)
				break;
		}
	}
}

void email(uint8_t *onoff)
{
	VCTL *vctl;

	if(!vctl_send(VCTL_CMD_ENABLEMAIL, VCTL_FLG_MSGSA, onoff, onoff ? 1 : 0))
	{
		for( ; (vctl = vctl_recvmsg()) ; )
		{
			if(vctl->type == VCTL_MSG_ONOFF)
			{
				if(vctl->data[0])
				{
					printf("Sending invitations via E-Mail is allowed.\n");
				}
				else
				{
					printf("Sending invitations via E-Mail is NOT allowed.\n");
				}
			}
			else if(vctl->type == VCTL_MSG_TEXT)
			{
				printf("%s\n", vctl->data);
			}

			if(vctl->flags & VCTL_FLG_MSGCT)
				break;
		}
	}
}

void regonly(uint8_t *onoff)
{
	VCTL *vctl;

	if(!vctl_send(VCTL_CMD_REGONLY, VCTL_FLG_MSGSA, onoff, onoff ? 1 : 0))
	{
		for( ; (vctl = vctl_recvmsg()) ; )
		{
			if(vctl->type == VCTL_MSG_ONOFF)
			{
				if(vctl->data[0])
				{
					printf("Only registered users are allowed to enter.\n");
				}
				else
				{
					printf("Unregistered users are allowed to enter.\n");
				}
			}
			else if(vctl->type == VCTL_MSG_TEXT)
			{
				printf("%s\n", vctl->data);
			}

			if(vctl->flags & VCTL_FLG_MSGCT)
				break;
		}
	}
}

void floodguard(uint8_t *onoff)
{
	VCTL *vctl;

	if(!vctl_send(VCTL_CMD_FLOODGUARD, VCTL_FLG_MSGSA, onoff, onoff ? 1 : 0))
	{
		for( ; (vctl = vctl_recvmsg()) ; )
		{
			if(vctl->type == VCTL_MSG_ONOFF)
			{
				if(vctl->data[0])
				{
					printf("FloodGuard flood protection is enabled.\n");
				}
				else
				{
					printf("FloodGuard flood protection is disabled.\n");
				}
			}
			else if(vctl->type == VCTL_MSG_TEXT)
			{
				printf("%s\n", vctl->data);
			}

			if(vctl->flags & VCTL_FLG_MSGCT)
				break;
		}
	}
}

void fgthreshold(uint32_t *arg)
{
	VCTL *vctl;

	if(!vctl_send(VCTL_CMD_FGTHRESHOLD, VCTL_FLG_MSGSA, arg, arg ? sizeof(uint32_t) : 0))
	{
		for( ; (vctl = vctl_recvmsg()) ; )
		{
			if(vctl->type == VCTL_MSG_FGTHRESHOLD)
			{
				uint32_t val = *(uint32_t *)&vctl->data[0];
				struct timeval tv;

				tv.tv_sec = val / 1000000;
				tv.tv_usec = val % 1000000;

				printf("FloodGuard threshold %s.\n",
					mt_print(&tv));
			}
			else if(vctl->type == VCTL_MSG_TEXT)
			{
				printf("%s\n", vctl->data);
			}

			if(vctl->flags & VCTL_FLG_MSGCT)
				break;
		}
	}
}

void shut_down(void)
{
	if(!vctl_sendmsg(VCTL_CMD_SHUTDOWN, VCTL_FLG_MSGSA, NULL))
	{
		vctl_showreply();
	}
}

void dnschklvl(uint8_t *lvl)
{
	VCTL *vctl;

	if(!vctl_send(VCTL_CMD_DNSCHKLVL, VCTL_FLG_MSGSA, lvl, lvl ? 1 : 0))
	{
		for( ; (vctl = vctl_recvmsg()) ; )
		{
			if(vctl->type == VCTL_MSG_DNSCHKLVL)
			{
				uint8_t curlvl = *(uint8_t *)&vctl->data[0];

				printf("Current DNS check level: %s.\n",
					dnslvlname(curlvl));
			}
			else if(vctl->type == VCTL_MSG_TEXT)
			{
				printf("%s\n", vctl->data);
			}

			if(vctl->flags & VCTL_FLG_MSGCT)
				break;
		}
	}
}

void servername(char *arg)
{
	VCTL *vctl;

	if(!vctl_sendmsg(VCTL_CMD_SERVERNAME, VCTL_FLG_MSGSA, "%s",
		(arg && *arg ? arg : "")))
	{
		for( ; (vctl = vctl_recvmsg()) ; )
		{
			if(vctl->type == VCTL_MSG_SERVERNAME)
			{
				printf("Servername is: %s.\n", vctl->data);
			}
			else if(vctl->type == VCTL_MSG_TEXT)
			{
				printf("%s\n", vctl->data);
			}

			if(vctl->flags & VCTL_FLG_MSGCT)
				break;
		}
	}
}

void uptime(void)
{
	VCTL *vctl;

	if(!vctl_send(VCTL_CMD_UPTIME, VCTL_FLG_MSGSA, NULL, 0))
	{
		for( ; (vctl = vctl_recvmsg()) ; )
		{
			if(vctl->type == VCTL_MSG_UPTIME)
			{
				time_t up = *(time_t *)&vctl->data[0];
				char dtbuf[64];

				strftime(dtbuf, sizeof(dtbuf),
					"%a, %d %b %Y %H:%M:%S %z",
					localtime(&up));
				dtbuf[sizeof(dtbuf)-1] = 0;

				printf("Server up since %s (%s).\n",
					dtbuf, elapsed_now(up));
			}
			else if(vctl->type == VCTL_MSG_TEXT)
			{
				printf("%s\n", vctl->data);
			}

			if(vctl->flags & VCTL_FLG_MSGCT)
				break;
		}
	}
}

void acl_list(void)
{
	if(!vctl_send(VCTL_CMD_ACL_LIST, VCTL_FLG_MSGSA, NULL, 0))
	{
		vctl_showreply();
	}
}

void acl_flush(void)
{
	if(!vctl_send(VCTL_CMD_ACL_FLUSH, VCTL_FLG_MSGSA, NULL, 0))
	{
		vctl_showreply();
	}
}

void acl_delete(uint16_t rule)
{
	if(!vctl_send(VCTL_CMD_ACL_DELETE, VCTL_FLG_MSGSA, &rule, sizeof(rule)))
	{
		vctl_showreply();
	}
}

void acl_add(VCTL_ACL *vacl)
{
	if(!vctl_send(VCTL_CMD_ACL_ADD, VCTL_FLG_MSGSA, vacl, vacl ? sizeof(VCTL_ACL) : 0))
	{
		vctl_showreply();
	}
}

void user_add(char *nick, char *passwd)
{
	VCTL_NEWUSER nu;
	size_t plen; 

	if(!nick_ok(nick))
	{
		printf("Nickname '%s' is invalid.\n", nick);
		return;
	}

	sncopy(nick, nu.nick, sizeof(nu.nick));
	plen = sncopy(passwd, nu.passwd, sizeof(nu.passwd));

	if(plen < 4 || plen > VPPASSWDSIZE)
	{
		printf("Please choose a password 4 to %d characters in "
			"length.\n", VPPASSWDSIZE);
		return;
	}

	if(!vctl_send(VCTL_CMD_USER_ADD, VCTL_FLG_MSGSA, &nu, sizeof(nu)))
	{
		vctl_showreply();
	}
}

void user_delete(char *nick)
{
	if(!vctl_sendmsg(VCTL_CMD_USER_DELETE, VCTL_FLG_MSGSA, "%s", nick))
	{
		vctl_showreply();
	}
}

void cnmessage(int32_t channel, char *nick)
{
	if(!nick || (nick && *nick))
	{
		uid_t uid = getuid();
		struct passwd *pwe;

		/*
		 * Get the users name from the password database.
		 */

		if((pwe = getpwuid(uid)))
		{
			VCTL_CNMSG cn;
			char buf[1016];
			char *opname = pwe->pw_name;
			char *optty = NULL;

			/*
			 * Get the users login terminal
			 */

			if(isatty(0))
				optty = ttyname(0);
			else if(isatty(1))
				optty = ttyname(1);
			else if(isatty(2))
				optty = ttyname(2);

			if(!optty)
				optty = "<stdin>";

			if(!strncmp(optty, "/dev/", 5))
				optty += 5;

			/*
			 * Prepare the message header structure with
			 * either the target nickname or channel.
			 */

			bzero(&cn, sizeof(cn));

			sncopy(opname, cn.from, sizeof(cn.from));
			sncopy(optty, cn.tty, sizeof(cn.tty));
	
			if(nick)
				sncopy(nick, cn.to, sizeof(cn.to));
			else
				cn.channel = channel;

			/*
			 * Send message header structure.
			 */

			if(!vctl_send(VCTL_CMD_CNMSG, VCTL_FLG_MSGIT, &cn, sizeof(cn)))
			{
				register char *s;

				/*
				 * Get the message data and send it.
				 */

				if(isatty(0))
					printf("Enter message [CTRL-D to end message]:\n");

				while(fgets(buf, sizeof(buf), stdin))
				{
					while((s = strrchr(buf, '\r')))
						*s = 0;

					if(vctl_sendmsg(VCTL_CMD_CNMSG, 0, "%s", buf))
					{
						return;
					}
				}

				if(feof(stdin) && !ferror(stdin))
				{
					clearerr(stdin);

					if(!vctl_send(VCTL_CMD_CNMSG, VCTL_FLG_MSGCT, NULL, 0))
					{
						vctl_showreply();
					}
				}
				else
				{
					fprintf(stderr, "%s: Read error on stdin\n",
						prog_name);
				}
			}
		}
		else
		{
			fprintf(stderr, "%s: Failed to get username for uid %d (%s)\n",
				prog_name, uid, strerror(errno));
		}
	}
	else
	{
		printf("Empty recipient nickname.\n");
	}
}

void lastlog(VCTL_LLQUERY *query, int wide)
{
	if(!vctl_send(VCTL_CMD_LASTLOG, VCTL_FLG_MSGSA, query, sizeof(VCTL_LLQUERY)))
	{
		register VCTL *vctl;
		register VCTL_LLRESP *r;
		char dtbuf1[24];
		char dtbuf2[8];
		uint32_t n;

		for(n = 0; (vctl = vctl_recvmsg()); )
		{
			if(vctl->type == VCTL_MSG_LASTLOG)
			{
				r = (VCTL_LLRESP *)&vctl->data[0];

				if(!n)
				{
					if(wide)
					{
						printf("UNAME           NICKNAME        FROM                                     LOGIN@               OUT   TIME         INFO\n");
					}
					else
					{
						printf("NICKNAME        FROM                    LOGIN@               OUT   TIME\n");
					}
				}

				strftime(dtbuf1, sizeof(dtbuf1),
					"%a %d.%m.%y %H:%M",
					localtime(&r->signon));
				dtbuf1[sizeof(dtbuf1)-1] = 0;

				strftime(dtbuf2, sizeof(dtbuf2),
					"%H:%M", localtime(&r->signoff));
				dtbuf2[sizeof(dtbuf2)-1] = 0;

				if(wide)
				{
					r->host[40] = 0;

					printf("%-15s %-15s %-40s %s - %s %-12s %s\n",
						r->user, r->nick, r->host,
						dtbuf1, dtbuf2,
						elapsed(r->signon, r->signoff),
						r->info);
				}
				else
				{
					r->host[23] = 0;

					printf("%-15s %-23s %s - %s %s\n",
						r->nick, r->host, dtbuf1,
						dtbuf2, elapsed(r->signon,
						r->signoff));
				}

				n++;
			}
			else if(vctl->type == VCTL_MSG_TEXT)
			{
				printf("%s\n", vctl->data);
			}

			if(vctl->flags & VCTL_FLG_MSGCT)
				break;
		}
	}
}

/*
 * Provide a user listing of a format that fulfills the requirements
 * of webchat.de for listing chat servers on their website.
 */

int webchatlist(int argc, char *argv[], PQLINK *pql)
{
	/*
	 * Output mime type header
	 */

	printf
	(
		"Content-type: text/plain; charset=iso-8859-15\n"
		"Cache-control: no-cache\n"
		"Pragma: no-cache\n"
		"\n"
	);

	fflush(stdout);

	/*
	 * Send request to server control socket
	 */

	if(!vctl_send(VCTL_CMD_ULIST, VCTL_FLG_MSGSA, NULL, 0))
	{
		VCTL *vctl;

		for( ; (vctl = vctl_recvmsg()); )
		{
			if(vctl->type == VCTL_MSG_UINFO)
			{
				VCTL_UINFO *vu = (VCTL_UINFO *)&vctl->data[0];

				if(vu->channel >= 0)
				{
					printf("(%s)(%d)(#%d)\n", vu->nick,
						vu->stamp, vu->channel);
				}
				else
				{
					printf("(%s)(%d)(#%s)\n", vu->nick,
						vu->stamp, "??????");
				}
			}


			if(vctl->flags & VCTL_FLG_MSGCT)
				break;
		}
	}

	fflush(stdout);

	return 0;
}
