/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: acl.c,v 1.15 2009/08/14 11:54:49 mickey Exp $ */

#include "vserver.h"

/****************
 *** ACL code ***
 ****************/

/*
 * Initialize the ACL database with the contents of the database file.
 */

int acldb_open(void)
{
	/*
	 * Open the ACL database and prime it with the ACL database
	 * file. If the file does not exist, it is created.
	 */

	if(!rms_open(RMS_ACL, sizeof(ACLRULE), fn_acldb, RMS_O_CREATE, 0600))
	{
		/*
		 * Sort database in the order of ascending ACL rule numbers.
		 */

		acldb_sort();

		/*
		 * Log the current ACL database status.
		 */

		acldb_log_status();

		return 0;
	}

	vlog(VLOG_CRIT, "Failed to open ACL database '%s' (%s)",
		fn_acldb, strerror(errno));

	return -1;
}

/*
 * Close the ACL database and release any memory used by it.
 */

int acldb_close(void)
{
	return rms_close(RMS_ACL);
}

/*
 * Save back the contents of the ACL database to the database file.
 */

ssize_t acldb_save(void)
{
	return rms_save(RMS_ACL);
}

/*
 * Comparator function used for sorting ACL rules in ascending
 * order of their rule numbers.
 */

int aclcmp(const void *a, const void *b)
{
	return ((ACLRULE *)a)->rule == ((ACLRULE *)b)->rule ? 0
		: ((ACLRULE *)a)->rule < ((ACLRULE *)b)->rule ? -1 : 1;
}

/*
 * Sort the contents of the ACL database in ascending order
 * of the ACL rule numbers.
 */

void acldb_sort(void)
{
	int records = rms_records(RMS_ACL);

	if(records > 0)
	{
		void *base = rms_addr(RMS_ACL, 0);

		if(base)
		{
			qsort(base, records, sizeof(ACLRULE), aclcmp);
		}
	}
}

/*
 * Add the ACL rule pointed to by <rule> to the ACL database. Upon
 * successful completion a value of 0 is returned. Otherwise a value
 * of -1 is returned and the global variable errno is set to indicate
 * the error.
 */

int acldb_add(ACLRULE *rule)
{
	int ret = -1;

	if(rule)
	{
		if(rule->action == ACL_ACCEPT || rule->action == ACL_REJECT)
		{
			time_t now;

			time(&now);

			if(!rule->expires || rule->expires > now)
			{
				ret = (rms_add_rec(RMS_ACL, rule) ? 0 : -1);

				if(!ret)
				{
					acldb_sort();

					ret = (acldb_save() >= 0 ? 0 : -1);
				}
			}
			else
			{
				errno = EINVAL;
			}
		}
		else
		{
			errno = EINVAL;
		}
	}
	else
	{
		errno = EINVAL;
	}

	return ret;
}

/*
 * Delete the first occurance of an ACL rule whose rule number
 * matches <rule>. If the ACL rule has been deleted successfully,
 * a value of 0 is returned. When no ACL rule matching <rule>
 * could be found, a value of 1 is returned. If an error occurs
 * the value returned is -1 and the global variable errno is
 * set to indicate the error.
 */

int acldb_delete(uint16_t rule)
{
	int32_t idx = acldb_find(rule);
	int ret = 1;

	if(idx >= 0)
	{
		if(!(ret = rms_del_rec(RMS_ACL, idx)))
			ret = (rms_save(RMS_ACL) >= 0 ? 0 : -1);
	}

	return ret;
}

/*
 * Get the number of ACL rules currently stored within
 * the ACL database.
 */

int acldb_records(void)
{
	return rms_records(RMS_ACL);
}

/*
 * Get the memory address of the ACL database record that is
 * indexed by <idx>. On successful completion the address is
 * returned. Otherwise a value of NULL is returned and the global
 * variable errno is set to indicate the error.
 */

ACLRULE *acldb_addr(int32_t idx)
{
	return (ACLRULE *)rms_addr(RMS_ACL, idx);
}

/*
 * Find the ACL rule specified by <rule> within the ACL database and
 * return the index of the ACL rule. If successful, the non-negative
 * index of the record is returned. Otherwise a value of -1 is returned
 * and the global variable errno is set to indicate the error.
 */

int acldb_find(uint16_t rule)
{
	int records = rms_records(RMS_ACL);

	if(records > 0)
	{
		register ACLRULE *rp = (ACLRULE *)rms_addr(RMS_ACL, 0);

		if(rp)
		{
			register int32_t i;

			for(i = 0; i < records; i++, rp++)
			{
				if(rp->rule == rule)
					return i;
			}
		}
	}

	errno = ENOENT;

	return -1;
}

/*
 * Find the ACL rule specified by <rule> within the ACL database and
 * return the memory address of the ACL rule. If successful, the address
 * of the record is returned. Otherwise a value of NULL is returned and
 * the global variable errno is set to indicate the error.
 */

ACLRULE *acldb_findaddr(uint16_t rule)
{
	int records = rms_records(RMS_ACL);

	if(records > 0)
	{
		register ACLRULE *rp = (ACLRULE *)rms_addr(RMS_ACL, 0);

		if(rp)
		{
			register int32_t i;

			for(i = 0; i < records; i++, rp++)
			{
				if(rp->rule == rule)
					return rp;
			}
		}
	}

	errno = ENOENT;

	return (ACLRULE *)0;
}

/*
 * Check the ACL database if the internet address specified by <address>
 * is allowed access. Return 1 if access is granted or 0 if access is
 * denied. If the specified <buf> argument is a non NULL pointer then
 * the rule granting/denying the access is written to the buffer it points
 * to, not exceeding <len> bytes including the terminating NUL character.
 */

int acldb_authorize(struct in_addr address, char *buf, size_t len)
{
	int records = rms_records(RMS_ACL);

	if(records > 0)
	{
		register ACLRULE *rp = (ACLRULE *)rms_addr(RMS_ACL, 0);

		if(rp)
		{
			register int32_t i;
			uint32_t a1;
			uint32_t a2;

			for(i = 0; rp && i < records; i++, rp++)
			{
				a1 = address.s_addr;
				a1 &= rp->netmask.s_addr;

				a2 = rp->address.s_addr;
				a2 &= rp->netmask.s_addr;

				/*
				 * Check if rule applies to address.
				 */

				if((a1 == a2) || a2 == 0)
				{
					if(buf && len > 0)
						acldb_print(rp, buf, len);

					return rp->action == ACL_ACCEPT ? 1 : 0;
				}
			}
		}
	}

	if(buf && len > 0)
		snprintf(buf, len, "0xFFFF accept any");

	return 1;	/* Allow */
}

/*
 * Delete all ACL rules from the database that have expired.
 */

void acldb_delete_expired(void)
{
	int records = rms_records(RMS_ACL);

	if(records > 0)
	{
		register ACLRULE *rp = (ACLRULE *)rms_addr(RMS_ACL, records-1);

		if(rp)
		{
			register int32_t i;
			register uint32_t deleted = 0;
			time_t now;

			time(&now);

			for(i = records - 1; i >= 0; i--, rp--)
			{
				if(rp->expires && rp->expires <= now)
				{
					rms_del_rec(RMS_ACL, i);

					++deleted;
				}
			}

			if(deleted)
				rms_save(RMS_ACL);
		}
	}
}

/*
 * Flush the ACL database.
 */

void acldb_flush(void)
{
	rms_discard(RMS_ACL);
	rms_save(RMS_ACL);
}

/*
 * Log the current ACL database status to the server log.
 */

void acldb_log_status(void)
{
	int records = rms_records(RMS_ACL);

	if(records >= 0)
	{
		records++;
		vlog(VLOG_INFO, "ACL status: %d rule%s active", records,
			records == 1 ? "" : "s");
	}
	else
	{
		vlog(VLOG_ERR, "ACL status: database failure (%s)",
			strerror(errno));
	}
}

/*
 * Print the ACL rule specified by <rule> in a convenient human readable
 * format into the buffer pointed to by <buf>, not exceeding <len> bytes
 * including the terminating NUL character.
 */

void acldb_print(ACLRULE *rule, char *buf, size_t len)
{
	int	bits;

	/*
	 * No buffer - no action.
	 */

	if(!buf || len <= 0)
		return;

	/*
	 * No rule - no action.
	 */

	if(!rule)
	{
		snprintf(buf, len, "<null>");
		return;
	}

	/*
	 * Format rule into buffer.
	 */

	bits = netmask_bits(rule->netmask);

	if(rule->expires > 0)
	{
		time_t now;
		char *ttl;

		time(&now);
		ttl = elapsed(now, rule->expires);

		if(rule->address.s_addr == INADDR_ANY)
		{
			snprintf(buf, len, "0x%04X %s any (Expires: %s)",
				rule->rule, (rule->action == ACL_ACCEPT) ?
				"accept" : "reject", ttl);
		}
		else if(bits == (CHAR_BIT * sizeof(uint32_t)))
		{
			snprintf(buf, len, "0x%04X %s %s (Expires: %s)",
				rule->rule, (rule->action == ACL_ACCEPT) ?
				"accept" : "reject", inet_ntoa(rule->address),
				ttl);
		}
		else
		{
			snprintf(buf, len, "0x%04X %s %s/%d (Expires: %s)",
				rule->rule, (rule->action == ACL_ACCEPT) ?
				"accept" : "reject", inet_ntoa(rule->address),
				bits, ttl);
		}
	}
	else
	{
		if(rule->address.s_addr == INADDR_ANY)
		{
			snprintf(buf, len, "0x%04X %s any", rule->rule,
				(rule->action == ACL_ACCEPT) ?
				"accept" : "reject");
		}
		else if(bits == (CHAR_BIT * sizeof(uint32_t)))
		{
			snprintf(buf, len, "0x%04X %s %s", rule->rule,
				(rule->action == ACL_ACCEPT) ?
				"accept" : "reject",
				inet_ntoa(rule->address));
		}
		else
		{
			snprintf(buf, len, "0x%04X %s %s/%d", rule->rule,
				(rule->action == ACL_ACCEPT) ?
				"accept" : "reject",
				inet_ntoa(rule->address), bits);
		}
	}
}
