/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: cldb.c,v 1.5 2009/08/14 11:54:49 mickey Exp $ */

#include "vserver.h"

/*
 * Initialize the client database.
 */

int cldb_open(void)
{
	return rms_open(RMS_CLIENT, sizeof(CLDATA), NULL, 0, 0);
}

/*
 * Dispose of the client database.
 */

void cldb_close(void)
{
	rms_close(RMS_CLIENT);
}

/*
 * Add the new client data record pointed to by <rec> to the client database.
 * Upon successful operation a pointer to the location where the record has
 * been stored into the database is returned. Otherwise a value of NULL is
 * returned.
 */

CLDATA *cldb_add(CLDATA *rec)
{
	return (CLDATA *)rms_add_rec(RMS_CLIENT, rec);
}

/*
 * Delete the client data record pointed to by <rec> from the client database.
 * Upon successful operation a value of 0 is returned. Otherwise a value of -1
 * is returned.
 */

int cldb_delete(CLDATA *rec)
{
	int idx = rms_index(RMS_CLIENT, rec);

	if(idx >= 0)
	{
		return rms_del_rec(RMS_CLIENT, idx);
	}

	return -1;
}

/*
 * Get the number of client data records currently contained within the
 * client database.
 */

uint32_t cldb_records(void)
{
	return rms_records(RMS_CLIENT);
}

/*
 * Get the memory address of the client data record stored at index <idx>.
 * In case <idx> does not denote a valid record index a value of NULL is
 * returned.
 */

CLDATA *cldb_addr(int idx)
{
	return (CLDATA *)rms_addr(RMS_CLIENT, idx);
}

/*
 * Find the client data record associated with file deescriptor <fd>.
 * Returns the index of the corresponding client data record. In case no
 * matching client data record could be located, a value of -1 is returned
 * and the global variable errno is set to indicate the error.
 */

int cldb_find(int fd)
{
	uint32_t records = rms_records(RMS_CLIENT);
	register CLDATA *cp;

	if(records && (cp = (CLDATA *)rms_addr(RMS_CLIENT, 0)))
	{
		register int i;

		for(i = 0; i < records; i++, cp++)
		{
			if(cp->fd == fd)
			{
				/* FOUND */

				errno = 0;

				return i;
			}
		}
	}

	errno = ENOENT;

	return -1;
}

/*
 * Find the client data record associated with file deescriptor <fd>. Returns
 * the memory address of the corresponding client data record. In case no
 * matching client data record could be located, a value of NULL is returned
 * and the global variable errno is set to indicate the error.
 */

CLDATA *cldb_findaddr(int fd)
{
	uint32_t records = rms_records(RMS_CLIENT);
	register CLDATA *cp;

	if(records && (cp = (CLDATA *)rms_addr(RMS_CLIENT, 0)))
	{
		register int i;

		for(i = 0; i < records; i++, cp++)
		{
			if(cp->fd == fd)
			{
				/* FOUND */

				errno = 0;

				return cp;
			}
		}
	}

	errno = ENOENT;

	return (CLDATA *)0;
}

/*
 * Check if the user specified by <nick> is currently on channel <channel>.
 * Returns 1 if the specified user is on the specified channel, 0 otherwise.
 */

uint32_t cldb_ison(char *nick, int32_t channel)
{
	uint32_t records = rms_records(RMS_CLIENT);
	register CLDATA *cp;
	uint32_t ret = 0;

	if(records && (cp  = (CLDATA *)rms_addr(RMS_CLIENT, 0)))
	{
		register int i;

		for(i = 0; i < records; i++, cp++)
		{
			if(cp->channel == channel && !strcmp(cp->nick, nick))
			{
				ret = 1;
				break;
			}
		}
	}

	return ret;
}

/*
 * Check if the user specified by <nick> is currently online and thus present
 * in the client database. Returns 1 if the specified user is online, otherwise
 * a value of 0 is returned.
 */

uint32_t cldb_chkpresent(char *nick)
{
	uint32_t records = rms_records(RMS_CLIENT);
	register CLDATA *cp;
	uint32_t ret = 0;

	if(records && (cp = (CLDATA *)rms_addr(RMS_CLIENT, 0)))
	{
		register int i;

		for(i = 0; i < records; i++, cp++)
		{
			if(!strcmp(cp->nick, nick))
			{
				ret = 1;
				break;
			}
		}
	}

	return ret;
}

/*
 * Kill the user specified by <nick> off the server. If <traitor> is non-NULL
 * then the string it points to is used as the name of the person committing
 * the kill. Returns the number of instances of user <nick> that has been
 * killed, or -1 in case an attempt was made to kill the VChat super-user.
 */

int cldb_kill(char *nick, char *traitor)
{
	uint32_t records = rms_records(RMS_CLIENT);
	register CLDATA *cp;
	int ret = 0;

	if(records && (cp = (CLDATA *)rms_addr(RMS_CLIENT, 0)))
	{
		register int i;

		for(i = 0; i < records; i++, cp++)
		{
			if(strcmp(cp->nick, nick))
				continue;

			if(cp->accperm == AUTH_ROOT
			&& cp->modperm == AUTH_ROOT)
			{
				ret = -1;
				break;
			}

			if(traitor && *traitor)
			{
				sncopy(traitor, cp->killinfo, sizeof(cp->killinfo));
				cp->flags |= CFLAG_KILL;
			}
			else
			{
#ifdef WIN32
				cp->error = EIO;
#else
				cp->error = ECONNRESET;
#endif
				cp->flags |= CFLAG_ERROR;
			}

			ret++;
		}
	}

	return ret;
}

/*
 * Kill all users on the specified <channel> off the server. If <traitor> is
 * non-NULL then the string it points to is used as the name of the person
 * committing the kill. Returns the number of users that has been killed.
 */

int cldb_killchannel(int32_t channel, char *traitor)
{
	uint32_t records = rms_records(RMS_CLIENT);
	register CLDATA *cp;
	int ret = 0;

	if(records && (cp = (CLDATA *)rms_addr(RMS_CLIENT, 0)))
	{
		register int i;

		for(i = 0; i < records; i++, cp++)
		{
			if(cp->channel != channel)
				continue;

			if(cp->accperm == AUTH_ROOT
			&& cp->modperm == AUTH_ROOT)
			{
				continue;
			}

			if(traitor && *traitor)
			{
				sncopy(traitor, cp->killinfo, sizeof(cp->killinfo));
				cp->flags |= CFLAG_KILL;
			}
			else
			{
#ifdef WIN32
				cp->error = EIO;
#else
				cp->error = ECONNRESET;
#endif
				cp->flags |= CFLAG_ERROR;
			}

			ret++;
		}
	}

	return ret;
}

/*
 * Kill all users off the server. If <traitor> is non-NULL then the string it
 * points to is used as the name of the person committing the kill. Returns
 * the number of users that has been killed.
 */

int cldb_killall(char *traitor)
{
	uint32_t records = rms_records(RMS_CLIENT);
	register CLDATA *cp;
	int ret = 0;

	if(records && (cp = (CLDATA *)rms_addr(RMS_CLIENT, 0)))
	{
		register int i;

		for(i = 0; i < records; i++, cp++)
		{
			if(cp->accperm == AUTH_ROOT
			&& cp->modperm == AUTH_ROOT)
			{
				continue;
			}

			if(traitor && *traitor)
			{
				sncopy(traitor, cp->killinfo, sizeof(cp->killinfo));
				cp->flags |= CFLAG_KILL;
			}
			else
			{
#ifdef WIN32
				cp->error = EIO;
#else
				cp->error = ECONNRESET;
#endif
				cp->flags |= CFLAG_ERROR;
			}

			ret++;
		}
	}

	return ret;
}

/*
 * Re-verify that all clients that are currently connected fulfill the
 * requirements for DNS check level <lvl>. All non-complying clients are
 * killed off the server with an appropriate message.
 */

uint32_t cldb_dnschklvl(uint8_t lvl)
{
	uint32_t records = rms_records(RMS_CLIENT);
	register CLDATA *cp;
	uint32_t ret = 0;

	if(records && (cp = (CLDATA *)rms_addr(RMS_CLIENT, 0)))
	{
		register int i;

		for(i = 0; i < records; i++, cp++)
		{
			if(cp->dnschklvl >= lvl)
				continue;

			cp->flags |= CFLAG_SIGNOFF;
			snprintf(cp->offmsg, sizeof(cp->offmsg),
				"level %d DNS clearance required", lvl);

			ret++;
		}
	}

	return ret;
}
