/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: cmd_chaccperm.c,v 1.11 2009/08/14 11:54:49 mickey Exp $ */

#include "vserver.h"

/*
 * Change the access permissions of registered nicknames
 */

void cmd_chaccperm(CLDATA *ptr, void *data, size_t len)
{
	change_perm(ptr, 'p', data, len);
}

void change_perm(CLDATA *ptr, char which, void *data, size_t len)
{
	TABLE			cltab;
	char			*xargv[VCMD_MAXARGS];
	register USERDATA	*udat;
	register int		i;
	uint32_t		md;
	int			xargc;
	uint32_t		givemask	= 0;
	uint32_t		takemask	= 0;
	char			modified	= 0;
	char			*arg;

	/*
	 * Check for correct permission set specifier
	 */

	if(which != 'p' && which != 'm')
	{
		vlog(VLOG_ERR, "Unknown permission set specifier '%c' encountered", which);
		return;
	}

	/*
	 * Check for minimum argument size
	 */

	if(len < (2 * sizeof(uint32_t)))
	{
		vcp_msg(ptr, "* Supplied datasize is too short");
		return;
	}
	else if(len == (2 * sizeof(uint32_t)))
	{
		vcp_msg(ptr, "* No nickname(s) specified");
		return;
	}

	/*
	 * Get givemask and takemask from supplied data packet
	 * and set the pointer for the start of the nicknamelist.
	 */

	givemask = ntohl(*((uint32_t *)(data)));
	takemask = ntohl(*((uint32_t *)(data + sizeof(uint32_t))));

	/*
	 * Initialize output table.
	 */

	table_init(&cltab, ptr);

	/*
	 * Parse nickname list into argument vector
	 */

	data += 2 * sizeof(uint32_t);

	if(!(xargc = nicksplit(&cltab, (char *)data, xargv, VCMD_MAXARGS, 1)))
		return;

	/*
	 * Check for permissions compensating each other.
	 */

	md = (givemask ^ takemask);
	givemask &= md;
	takemask &= md;

	/*
	 * Check if any mode change would take place.
	 */

	if(givemask == 0 && takemask == 0)
	{
		table_xprintf(&cltab, "\n* Null permission change specified");
		table_flush(&cltab);
		return;
	}

	/*
	 * Check if the client is allowed to change the specified modes.
	 */

	md = (~ptr->modperm & givemask);
	md |= (~ptr->modperm & takemask);

	if(md)
	{
		table_printf(&cltab, "* %sYou are not authorized to change the following %s permission(s):%s\n*%s",
			S_UL, (which == 'p') ? "access" : "modify", S_OFF, S_FG_RED);

		if(md & AUTH_ADD)
			table_printf(&cltab, " ADD");
		if(md & AUTH_DELETE)
			table_printf(&cltab, " DELETE");
		if(md & AUTH_CHACCPERM)
			table_printf(&cltab, " CHACCPERM");
		if(md & AUTH_CHMODPERM)
			table_printf(&cltab, " CHMODPERM");
		if(md & AUTH_KILL)
			table_printf(&cltab, " KILL");
		if(md & AUTH_BROADCAST)
			table_printf(&cltab, " BCAST");
		if(md & AUTH_ACL)
			table_printf(&cltab, " ACL");
		if(md & AUTH_QUERY)
			table_printf(&cltab, " QUERY");
		if(md & AUTH_SOPT)
			table_printf(&cltab, " SOPT");
		if(md & AUTH_FIXCHAN)
			table_printf(&cltab, " FIX");
		if(md & AUTH_STATUS)
			table_printf(&cltab, " STATUS");
		if(md & AUTH_EXCLUDE)
			table_printf(&cltab, " EXCLUDE");

		table_printf(&cltab, "%s", S_OFF);
		table_flush(&cltab);

		return;
	}

	/*
	 * Process nickname list.
	 */

	for(i = 0; i < xargc; i++)
	{
		arg = xargv[i];

		/*
		 * Deny modifications to own permissions.
		 */

		if(!strcmp(arg, ptr->nick))
		{
			table_xprintf(&cltab, "\n* You may not modify your own %s permissions - ignored",
				(which == 'p') ? "access" : "modify");
			continue;
		}

		/*
		 * Get user database record for current nickname.
		 */

		if(!(udat = userdb_findaddr(arg)))
		{
			if(errno == ENOENT)
			{
				table_xprintf(&cltab, "\n* Nickname '%s%s%s' "
					"not found in user database", S_FG_RED,
					arg, S_OFF);
			}
			else
			{
				table_xprintf(&cltab, "\n* Failed to get user "
					"database record for '%s%s%s' (%s%s%s)",
					S_FG_RED, arg, S_OFF,
					S_FG_RED, strerror(errno), S_OFF);
			}

			continue;
		}
		
		/*
		 * Deny modification of the VChat super-user's permissions.
		 */

		if(udat->accperm == AUTH_ROOT && udat->modperm == AUTH_ROOT)
		{
			table_xprintf(&cltab, "* The %s permissions of the "
				"'%s%s%s' user may not be changed - ignored",
				(which == 'p') ? "access" : "modify",
				S_FG_RED, arg, S_OFF);

			continue;
		}

		/*
		 * Modifiy the permissions.
		 */

		md = ((which == 'm') ? udat->modperm : udat->accperm);
		md |= givemask;
		md &= ~takemask;

		if(md != ((which == 'm') ? udat->modperm : udat->accperm))
		{
			uint32_t records = cldb_records();
			register CLDATA *cp;

			if(which == 'm')
				udat->modperm = md;
			else
				udat->accperm = md;

			modified++;

			/*
			 * Check if the nickname is currently signed on
			 * and change permissions accordingly.
			 */

			if(records && (cp = cldb_addr(0)))
			{
				register int y;

				for(y = 0; y < records; y++, cp++)
				{
					if(!strncmp(cp->nick, arg, VPNICKSIZE))
					{
						if(which == 'm')
						{
							cp->modperm = md;
						}
						else
						{
							cp->accperm = md;
						}
					}
				}
			}
		}
	}

	/*
	 * Update user database file if necessary.
	 */

	if(modified)
	{
		if(userdb_save() < 0)
		{
			table_xprintf(&cltab, "\n* Failure updating user database (%s%s%s) - inform system operator",
				S_FG_RED, strerror(errno), S_OFF);
		}
		else
		{
			table_xprintf(&cltab, "\n* User database updated successfully");
		}
	}
	else
	{
		table_xprintf(&cltab, "\n* User database unchanged");
	}

	/*
	 * Flush output.
	 */

	table_flush(&cltab);
}
