/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: cmd_chnick.c,v 1.11 2009/08/14 11:54:50 mickey Exp $ */

#include "vserver.h"

/*
 * Change to another nickname
 */

void cmd_chnick(CLDATA *ptr, void *data, size_t len)
{
	uint32_t	records;
	register CLDATA	*cp;
	USERDATA	*udat;
	char		*nick;
	char		*passwd;
	char		updateudb	= 0;
	uint32_t	oldflags	= 0;
	uint32_t	old_opts	= 0;
	time_t		last_in		= 0;

	/*
	 * Split password from nickname.
	 */

	nick = (char *)data;

	if((passwd = strchr(nick, ' ')) == NULL)
		passwd = "";
	else
		*passwd++ = '\0';

	/*
	 * Check against current nickname.
	 */

	if(!strcmp(nick, ptr->nick))
	{
		vcp_msg(ptr, "* Nickname unchanged");
		return;
	}

	/*
	 * Reject invalid nicknames.
	 */

	if(!nick_ok(nick))
	{
		vcp_msg(ptr, "* Nickname '%s%s%s' is invalid",
			S_FG_RED, nick, S_OFF);
		return;
	}

	/*
	 * Check if the specified nickname is in the user database.
	 */

	if((udat = userdb_findaddr(nick)) != NULL)
	{
		/*
		 * Yes - Check if the supplied password matches.
		 */

		if(*passwd == '\0')
		{
			vcp_msg(ptr, "* Permission denied");
			return;
		}

		if(strncmp(udat->passwd, passwd, VPPASSWDSIZE))
		{
			vcp_msg(ptr, "* Permission denied");
			return;
		}

		/*
		 * Switch to new access and modify permissions.
		 */

		ptr->accperm = udat->accperm;
		ptr->modperm = udat->modperm;

		/*
		 * Get saved client options from authority record.
		 */

		old_opts = ptr->cl_opts;
		ptr->cl_opts = udat->cl_opts;

		/*
		 * Save old client flags value for later use
		 * and set the client registered flag.
		 */

		oldflags = ptr->flags;
		ptr->flags |= CFLAG_REGISTERED;

		/*
		 * Save time of last signon to display it later and
		 * update the user database record with the current time.
		 */

		last_in = udat->last_signon;
		time(&udat->last_signon);
		udat->last_signoff = 0xffffffff;
		++udat->sessioncnt;

		/*
		 * User database file needs to be updated.
		 */

		updateudb = 1;
	}
	else if(errno == ENOENT)
	{
		/*
		 * New nickname is not in the user database.
		 * If the server is in closed mode then reject
		 * any unregistered nicknames.
		 */

		if(vsopt.flags & VO_REGONLY)
		{
			vcp_msg(ptr, "* Nickname '%s%s%s' is not registered",
				S_FG_RED, nick, S_OFF);
			return;
		}

		/*
		 * Check for duplicated nicknames.
		 * Unregistered nicknames may only be present
		 * once at every given time.
		 */

		if(cldb_chkpresent(nick))
		{
			vcp_msg(ptr, "* Nickname '%s%s%s' is already in use",
				S_FG_RED, nick, S_OFF);
			return;
		}

		/*
		 * Unregistered nicknames do not get any
		 * access or modify permissions
		 */

		ptr->accperm = ptr->modperm = 0;

		/*
		 * Reset client options to default.
		 */

		old_opts = ptr->cl_opts;
		ptr->cl_opts = DEF_CLOPTMASK;

		/*
		 * Clear client registered flag.
		 */

		oldflags = ptr->flags;
		ptr->flags &= ~CFLAG_REGISTERED;
	}
	else
	{
		vcp_msg(ptr, "* User database failure (%s%s%s) - nickname "
			"change temporarily rejected",
			S_FG_RED, strerror(errno), S_OFF);
		return;
	}

	/*
	 * Update user database record of previous nickname,
	 * when changing from a registered nickname.
	 */

	if(oldflags & CFLAG_REGISTERED)
	{
		if((udat = userdb_findaddr(ptr->nick)))
		{
			time(&udat->last_signoff);

			udat->pubmsgcnt += ptr->pubmsgcnt;
			udat->prvmsgcnt	+= ptr->prvmsgcnt;

			udat->cl_opts = old_opts;

			/*
			 * User database file needs to be updated.
			 */

			updateudb = 1;
		}
		else
		{
			vlog(VLOG_ERR, "Registered nickname '%s' without user "
				"database record ?! (%s)", ptr->nick,
				strerror(errno));
		}
	}

	/*
	 * Update user database file if necessary.
	 */

	if(updateudb)
	{
		if(userdb_save() < 0)
		{
			vcp_msg(ptr, "* Failed to update user database file "
				"(%s%s%s) - inform operator", S_FG_RED,
				strerror(errno), S_OFF);
		}
	}

	/*
	 * Add a record to the lastlog database.
	 */

	lastlog_add_record(ptr, "nickname changed");

	/*
	 * Make a new login and activity timestamp
	 * in the client data and reset all counters.
	 */

	time(&ptr->login);
	ptr->stamp = ptr->login;

	ptr->pubmsgcnt = 0;
	ptr->prvmsgcnt = 0;

	/*
	 * Acknowledge the new nickname.
	 */

	vcp_snd(ptr, NICK_ACK, NULL, 0);	

	/*
	 * Notify clients of name change.
	 */

	if((records = cldb_records()) && (cp = cldb_addr(0)))
	{
		register int i;
		size_t nicklen = strlen(nick);

		for(i = 0; i < records; i++, cp++)
		{
			/*
			 * Client with a protocol revision of 4-10 and above
			 * receive ULIST update messages to keep their userlist
			 * in sync with the server.
			 */

			if(cp->protocol >= 0x00040010)
			{
				vcp_snd(cp, ULIST_REMOVE, ptr->nick, ptr->nicklen);
				vcp_snd(cp, ULIST_ADD, nick, nicklen);
			}

			if(cp->channel == ptr->channel)
			{
				vcp_msg(cp, "* %s%s%s is now known as %s%s%s",
					S_FG_CYN, ptr->nick, S_OFF,
					S_FG_YEL, nick, S_OFF);
			}
		}
	}

	vlog(VLOG_INFO, "Rename client USER: %s OLD-NICK: %s NEW-NICK: %s",
		ptr->user, ptr->nick, nick);

	/*
	 * Actually change the nickname.
	 */

	ptr->nicklen = sncopy(nick, ptr->nick, sizeof(ptr->nick));

	/*
	 * If the last signon time is known, display it now.
	 */

	if(last_in)
	{
		char	dtbuf[DT_SIZE];

		strftime(dtbuf, sizeof(dtbuf), DT_FMT, localtime(&last_in));
		dtbuf[sizeof(dtbuf)-1] = 0;

		vcp_msg(ptr, "* Last login %s", dtbuf);
	}
}
