/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: cmd_chpasswd.c,v 1.10 2009/08/14 11:54:50 mickey Exp $ */

#include "vserver.h"

/*
 * Change the password of the calling user.
 */

void cmd_chpasswd(CLDATA *ptr, void *data, size_t len)
{
	USERDATA	*udat;
	char		*newpass;
	int		plen;
	char		ignore	= 0;

	char		curpwd[VPPASSWDSIZE+1];

	/*
	 * Check if nickname is registered.
	 */

	if(!(ptr->flags & CFLAG_REGISTERED))
	{
		vcp_msg(ptr, "* Your nickname is currently not registered\n"
			"* Apparently only registered users may change their "
			"password");
		return;
	}

	/*
	 * The current password and the new password are separated by
	 * a single space character. Since passwords may also contain
	 * space characters, all occurances of spaces within the first
	 * argument are escaped by a CSI character. So find the first
	 * non escaped space and replace it by a NUL character to split
	 * both strings. Then unescape the first argument to get both
	 * passwords separately.
	 */

	for(newpass = (char *)data; *newpass
	&& !(*newpass == ' ' && !ignore) ; newpass++)
	{
		ignore = (*(uint8_t *)newpass == CSI) ? 1 : 0;
	}

	if(!*newpass)
	{
		vcp_msg(ptr, "* Invalid argument");
		return;
	}

	*newpass++ = 0;

	unescapestr((uint8_t *)data, (uint8_t *)curpwd, sizeof(curpwd));

	/*
	 * Check for empty passwords.
	 */

	if(!curpwd[0] || !*newpass)
	{
		vcp_msg(ptr, "* Empty passwords are not allowed");
		return;
	}

	/*
	 * Check for valid password length of new password.
	 */

	plen = strlen(newpass);

	if(plen < MINPASSWDSIZE || plen > VPPASSWDSIZE)
	{
		vcp_msg(ptr, "* Please choose a password %d - %d characters"
			"in length", MINPASSWDSIZE, VPPASSWDSIZE);
		return;
	}

	/*
	 * Get user database record.
	 */

	if((udat = userdb_findaddr(ptr->nick)))
	{
		/*
		 * Check the supplied current password against database.
		 */

		if(!strncmp(udat->passwd, curpwd, VPPASSWDSIZE))
		{
			if(strncmp(udat->passwd, newpass, VPPASSWDSIZE))
			{
				/*
				 * Set new password, sync user database file.
				 */

				sncopy(newpass, udat->passwd, sizeof(udat->passwd));

				if(userdb_save() >= 0)
				{
					vcp_msg(ptr, "* Password successfully changed");
				}
				else
				{
					/*
					 * Failed to write user database to
					 * disk! Restore the old password
					 * and try saving user database again.
					 */

					vcp_msg(ptr, "* Failed to update "
						"user database (%s%s%s) - "
						"inform operator", S_FG_RED,
						strerror(errno), S_OFF);

					sncopy(curpwd, udat->passwd, sizeof(udat->passwd));
					userdb_save();

					vcp_msg(ptr, "* Password unchanged");
				}
			}
			else
			{
				vcp_msg(ptr, "* Please choose a password "
					"different from your old one");	
			}
		}
		else
		{
			vcp_msg(ptr, "* Permission denied");
		}
	}
	else if(errno == ENOENT)
	{
		vlog(VLOG_WARNING, "Client data for nickname '%s' says it's "
			"registered, but no record found in user database?",
			ptr->nick);

		vcp_msg(ptr, "* Your nickname is currently not registered\n"
			"* Apparently only registered users may change their "
			"password");
	}
	else
	{
		vcp_msg(ptr, "* User database failure (%s%s%s)\n"
			"* Changing your password is temporarily not possible",
			S_FG_RED, strerror(errno), S_OFF);
	}
}
