/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: cmd_clientinfo.c,v 1.15 2009/08/14 11:54:50 mickey Exp $ */

#include "vserver.h"

/*
 * Report information about a client
 */

void cmd_clientinfo(CLDATA *ptr, void *data, size_t len)
{
	if(ptr->flags & CFLAG_REGISTERED)
	{
		TABLE cltab;
		int xargc;
		char *xargv[VCMD_MAXARGS];

		table_init(&cltab, ptr);

		/*
		 * Parse nicknames into argument vector.
		 */

		if((xargc = nicksplit(&cltab, (char *)data, xargv, VCMD_MAXARGS, 1)))
		{
			uint32_t records = cldb_records();
			register CLDATA	*cp;

			if(records)
			{
				register int i;
				register int y;
				register int n;
				int cnt	= 0;

				/*
				 * Lookup matching client entries.
				 */

				for(y = 0; y < xargc; y++)
				{
					for(i = n = 0, cp = cldb_addr(0); cp && i < records; i++, cp++)
					{
						/*
						 * Skip the current entry if
						 * it is not on the list of
						 * requested nicknames.
						 */

						if(strcmp(cp->nick, xargv[y]))
							continue;

						/*
						 * Print header along with
						 * first listed entry.
						 */

						if(!cnt)
						{
							table_printf(&cltab, "\n*%s%s                                %s CLIENTINFO %s                                 %s",
								S_BG_BLU, S_FG_WHT, S_BG_RED, S_BG_BLU, S_OFF);
						}
						else
						{
							table_printf(&cltab, "\n*");
						}

						/*
						 * Show client info.
						 */

						clinfo(cp, &cltab, ptr);

						++n;
						++cnt;
					}

					if(!n)
					{
						table_xprintf(&cltab, "\n* Nickname '%s%s%s' not found",
							S_FG_RED, xargv[y], S_OFF);
					}
				}
			}
		}

		/*
		 * Flush output table.
		 */

		table_flush(&cltab);
	}
	else
	{
		vcp_msg(ptr, "* Only registered users may use the CLIENTINFO command");
	}
}

void clinfo(CLDATA *cp, TABLE *tab, CLDATA *ptr)
{
	time_t		online;
	char		buf1[128];
	char		buf2[128];
	char		dtbuf[DT_SIZE];

	/*
	 * Line 1: username, nickname, channel
	 */

	if(cp->type == CLIENT_GATEWAY
	|| !strcmp(cp->user, cp->nick))
	{
		snprintf(buf1, sizeof(buf1), "%s",
			cp->nick);
	}
	else
	{
		snprintf(buf1, sizeof(buf1), "%s (%s)",
			cp->user, cp->nick);
	}

	if(cp->channel >= 0 || cp->channel == ptr->channel)
	{
		snprintf(buf2, sizeof(buf2),
			" on channel %d", cp->channel);
	}
	else
	{
		buf2[0] = '\0';
	}

	table_printf(tab, "\n*     user: %s%s, %s",
		buf1, buf2, (cp->flags & CFLAG_REGISTERED) ?
		"registered" : "unregistered");

	/*
	 * Line 2: messages total, public, private, msgs/sec.
	 */

	time(&online);

	if((online -= cp->login) <= 0)
		online = 1;

	snprintf(buf1, sizeof(buf1), "%.4f",
		((double)(cp->pubmsgcnt + cp->prvmsgcnt) / online));

	table_printf(tab, "\n* messages: %d total, %d public, %d private, avg. %s msgs/sec",
		(cp->pubmsgcnt + cp->prvmsgcnt), cp->pubmsgcnt, cp->prvmsgcnt,
		buf1);

	/*
	 * Line 3: login@
	 */

	strftime(dtbuf, sizeof(dtbuf), DT_FMT, localtime(&cp->login));
	dtbuf[sizeof(dtbuf)-1] = 0;

	table_printf(tab, "\n*   login@: %s", dtbuf);

	/*
	 * Line 4: time online, idle.
	 */

	table_printf(tab, "\n*     time: %s online, %s idle",
		elapsed_now(cp->login), elapsed_now(cp->stamp));

	/*
	 * Line 5: client ident, type, proto, conn-id,
	 *         dnslvl.
	 */

	table_printf(tab, "\n*   client: %s, type %s, proto %x-%x, conn 0x%X, dnslvl %d",
		cp->ident, (cp->type == CLIENT_FRONTEND) ?
		"FRONTEND" : "GATEWAY",
		MAJOR_REV(cp->protocol),
		MINOR_REV(cp->protocol),
		cp->fd, cp->dnschklvl);

	/*
	 * Line 6: host ip-address, hostname.
	 */

	if(!(vsopt.flags & VO_ANONYMOUS))
	{
		table_printf(tab, "\n*     from: %s (%s)",
			iptostr(cp->ip_addr), cp->host);
	}
	else
	{
		table_printf(tab, "\n*     from: -concealed-");
	}

	/*
	 * Line 7: gateway ip-address, hostname.
	 */

	if(cp->type == CLIENT_GATEWAY)
	{
		table_printf(tab, "\n*      via: %s (%s)",
			iptostr(cp->gw_addr), cp->gw_name);
	}
}
