/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: cmd_exclude.c,v 1.11 2009/08/14 11:54:50 mickey Exp $ */

#include "vserver.h"

/*
 * Exclude users from chat by adding a dynamic ACL first and then kicking
 * them from the server.
 */

void cmd_exclude(CLDATA *ptr, void *data, size_t len)
{
	char		*dat = (char *)data;
	static uint16_t	dynacl_next	= DYNACL_FIRST_RULE;
	uint32_t	records;
	register CLDATA	*cp;
	register int	i;
	register int	y;
	register int	n;
	TABLE		cltab;
	char		*xargv[VCMD_MAXARGS];
	time_t		xttlv[VCMD_MAXARGS];
	time_t		ttl;
	char		**aptr;
	char		*s;
	int		xargc		= 0;

	/*
	 * Initialize output table.
	 */

	table_init(&cltab, ptr);

	/*
	 * Parse arguments into vector.
	 */

#if HAVE_STRSEP
	for(aptr = xargv; (*aptr = strsep(&dat, " \t")) != NULL && xargc < VCMD_MAXARGS;)
#else
	aptr = xargv;

	*aptr = strtok(dat, " \t");

	while(*aptr != NULL && xargc < VCMD_MAXARGS)
#endif
	{
		/*
		 * Check for optional TTL.
		 */

		if((s = strchr(*aptr, ':')) != NULL)
		{
			/*
			 * Parse TTL.
			 */

			*s++ = '\0';
			ttl = strtol(s, NULL, 0);
		}
		else
		{
			/*
			 * No TTL given, use default TTL.
			 */

			ttl = DYNACL_DEFAULT_TTL;
		}

		/*
		 * Check for invalid nicknames.
		 */

		if(!nick_ok(*aptr))
		{
			/*
			 * Nickname invalid.
			 */

			table_xprintf(&cltab, "\n* Invalid nickname '%s%s%s' ignored",
				S_FG_RED, *aptr, S_OFF);
			y = 1;
		}
		else
		{
			/*
			 * Check for duplicated nicknames.
			 */

			y = 0;

			for(i = 0; i < xargc; i++)
			{
				if(!strcmp(xargv[i], *aptr))
				{
					table_xprintf(&cltab, "\n* Duplicated nickname '%s%s%s' ignored",
						S_FG_RED, *aptr, S_OFF);

					y = 1;
					break;
				}
			}
		}

		if(**aptr != '\0' && !y)
		{
			++aptr;
			xttlv[xargc++] = ttl;
		}

#if !HAVE_STRSEP
		*aptr = strtok(NULL, " \t");
#endif
	}

	if(!xargc)
	{
		table_xprintf(&cltab, "\n* No nickname(s) to exclude");
		table_flush(&cltab);
		return;
	}

	/*
	 * Lookup client entries
	 */

	records = cldb_records();

	for(y = 0; y < xargc; y++)
	{
		/*
		 * Check if TTL specification is valid.
		 */

		if(xttlv[y] > DYNACL_MAX_TTL)
		{
			table_xprintf(&cltab, "\n* TTL for nickname '%s%s%s' (%d) invalid, using max TTL (%d)",
				S_FG_RED, xargv[y], S_OFF, xttlv[y],
				DYNACL_MAX_TTL);

			xttlv[y] = DYNACL_MAX_TTL;
		}

		/*
		 * Lookup matching clients.
		 */

		for(i = 0, n = 0, cp = cldb_addr(0); cp && i < records; i++, cp++)
		{
			if(!strcmp(cp->nick, xargv[y]))
			{
				ACLRULE new;

				bzero(&new, sizeof(new));

				/*
				 * Mark the client to be killed.
				 */

				cp->flags |= CFLAG_EXCLUDE;
				sncopy(ptr->nick, cp->killinfo, sizeof(cp->killinfo));

				/*
				 * Create a new dynamic ACL that excludes
				 * the user's IP address for the given TTL.
				 */

				if(dynacl_next >= 65534)
					dynacl_next = DYNACL_FIRST_RULE;

				new.rule = dynacl_next++;
				new.action = ACL_REJECT;
				bcopy(&cp->ip_addr, &new.address, sizeof(new.address));
				new.netmask.s_addr = 0xffffffff;
				time(&new.expires);
				new.expires += xttlv[y];

				if(!acldb_add(&new))
				{
					char buf[256];

					acldb_print(&new, buf, sizeof(buf));

					table_xprintf(&cltab, "\n* Excluded "
						"'%s%s%s' by ACL %s",
						S_FG_YEL, xargv[y], S_OFF, buf);
				}
				else
				{
					table_xprintf(&cltab, "\n* Failed to "
						"add ACL rule (%s%s%s)",
						S_FG_RED, strerror(errno),
						S_OFF);
				}

				++n;
			}
		}

		if(!n)
		{
			table_xprintf(&cltab, "\n* Nickname '%s%s%s' not found",
				S_FG_RED, xargv[y], S_OFF);
		}
	}

	/*
	 * Flush output table.
	 */

	table_flush(&cltab);
}
