/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: cmd_query.c,v 1.13 2009/08/14 11:54:51 mickey Exp $ */

#include "vserver.h"

/*
 * Query information about registered nicknames
 */

void cmd_query(CLDATA *ptr, void *data, size_t len)
{
	register int		i;
	register USERDATA	*udat;
	TABLE			cltab;
	char			dtbuf[DT_SIZE];
	char			*xargv[VCMD_MAXARGS];
	int			xargc;

	/*
	 * Default to own nickname, if no nicknames are specified.
	 */

	if(*((char *)data) == 0)
		data = ptr->nick;

	/*
	 * Initialize output table.
	 */

	table_init(&cltab, ptr);

	/*
	 * Parse nicknames into argument vector.
	 */

	if(!(xargc = nicksplit(&cltab, (char *)data, xargv, VCMD_MAXARGS, 1)))
		return;

	/*
	 * Lookup specified nickname in the user database.
	 */

	for(i = 0; i < xargc; i++)
	{
		if(!(udat = userdb_findaddr(xargv[i])))
		{
			if(errno == ENOENT)
			{
				table_xprintf(&cltab, "\n* Nickname '%s%s%s' not registered",
					S_FG_RED, xargv[i], S_OFF);
			}
			else
			{
				table_xprintf(&cltab, "\n* User database "
					"failed for '%s%s%s' (%s%s%s)",
					S_FG_RED, xargv[i], S_OFF,
					S_FG_RED, strerror(errno), S_OFF);
			}

			continue;
		}

		/*
		 * Display data from user database record.
		 */

		if(i)
			table_printf(&cltab, "\n");

		table_printf(&cltab, "\n*%s%s%s %s:%s%s%*s%s\n", S_BG_BLU,
			S_FG_WHT, S_BOLD, xargv[i], S_OFF, S_BG_BLU,
			(75 - strlen(xargv[i])), " ", S_OFF);

		/*
		 * Display the total number of sessions.
		 */

		table_printf(&cltab, "*          Sessions: %d\n",
			udat->sessioncnt);

		/*
		 * Display the total number of public and private messages.
		 */

		table_printf(&cltab, "*          Messages: %d public / %d private\n",
			udat->pubmsgcnt, udat->prvmsgcnt);

		/*
		 * Display the date of last signon.
		 */

		if(udat->last_signon)
		{
			strftime(dtbuf, sizeof(dtbuf), DT_FMT,
				localtime(&udat->last_signon));
			dtbuf[sizeof(dtbuf)-1] = 0;
		}
		else
		{
			snprintf(dtbuf, sizeof(dtbuf), "<never logged in>");
		}

		table_printf(&cltab, "*        Last login: %s\n", dtbuf);

		/*
		 * Display the date of last signoff.
		 */

		if(udat->last_signoff == 0xffffffff)
		{
			table_printf(&cltab, "*       Last logout: <still logged in>  (%s)\n",
				elapsed_now(udat->last_signon));
		}
		else if(udat->last_signoff)
		{
			strftime(dtbuf, sizeof(dtbuf), DT_FMT,
				localtime(&udat->last_signoff));
			dtbuf[sizeof(dtbuf)-1] = 0;

			table_printf(&cltab, "*       Last logout: %s  (%s)\n",
				dtbuf, elapsed(udat->last_signon,
				udat->last_signoff));
		}
		else
		{
			table_printf(&cltab, "*       Last logout: <never logged in>\n");
		}

		/*
		 * Display the date when the account was created.
		 */

		strftime(dtbuf, sizeof(dtbuf), DT_FMT,
			localtime(&udat->account_created));
		dtbuf[sizeof(dtbuf)-1] = 0;

		table_printf(&cltab, "*   Account created: %s\n", dtbuf);

		/*
		 * Display access privileges
		 */

		table_printf(&cltab, "* Access permission: ",
			S_UL, S_OFF);

		display_permission1(&cltab, udat->accperm);

		table_printf(&cltab, "\n*                    ");

		display_permission2(&cltab, udat->accperm);

		if((udat->accperm & AUTH_CHACCPERM)
		|| (udat->accperm & AUTH_CHMODPERM))
		{
			table_printf(&cltab, "\n* Modify permission: ",
				S_UL, S_OFF);

			display_permission1(&cltab, udat->modperm);

			table_printf(&cltab, "\n*                    ");

			display_permission2(&cltab, udat->modperm);
		}

	}

	table_flush(&cltab);
}

void display_permission1(TABLE *tab, uint32_t perm)
{
	if(perm & AUTH_ADD)
		table_printf(tab, "%s+%sADD", S_FG_GRN, S_OFF);
	else
		table_printf(tab, "%s-%sADD", S_FG_RED, S_OFF);	

	if(perm & AUTH_DELETE)
		table_printf(tab, " %s+%sDELETE", S_FG_GRN, S_OFF);
	else
		table_printf(tab, " %s-%sDELETE", S_FG_RED, S_OFF);

	if(perm & AUTH_CHACCPERM)
		table_printf(tab, " %s+%sCHACC", S_FG_GRN, S_OFF);
	else
		table_printf(tab, " %s-%sCHACC", S_FG_RED, S_OFF);

	if(perm & AUTH_CHMODPERM)
		table_printf(tab, " %s+%sCHMOD", S_FG_GRN, S_OFF);
	else
		table_printf(tab, " %s-%sCHMOD", S_FG_RED, S_OFF);

	if(perm & AUTH_KILL)
		table_printf(tab, " %s+%sKILL", S_FG_GRN, S_OFF);
	else
		table_printf(tab, " %s-%sKILL", S_FG_RED, S_OFF);

	if(perm & AUTH_BROADCAST)
		table_printf(tab, " %s+%sBCAST", S_FG_GRN, S_OFF);
	else
		table_printf(tab, " %s-%sBCAST", S_FG_RED, S_OFF);

	if((perm & AUTH_ACL) == AUTH_ACL)
		table_printf(tab, " %s+%sACL", S_FG_GRN, S_OFF);
	else
		table_printf(tab, " %s-%sACL", S_FG_RED, S_OFF);

	if((perm & AUTH_QUERY) == AUTH_QUERY)
		table_printf(tab, " %s+%sQUERY", S_FG_GRN, S_OFF);
	else
		table_printf(tab, " %s-%sQUERY", S_FG_RED, S_OFF);

	if((perm & AUTH_SOPT) == AUTH_SOPT)
		table_printf(tab, " %s+%sSOPT", S_FG_GRN, S_OFF);
	else
		table_printf(tab, " %s-%sSOPT", S_FG_RED, S_OFF);
}

void display_permission2(TABLE *tab, uint32_t perm)
{
	if(perm & AUTH_FIXCHAN)
		table_printf(tab, "%s+%sFIX", S_FG_GRN, S_OFF);
	else
		table_printf(tab, "%s-%sFIX", S_FG_RED, S_OFF);	

	if(perm & AUTH_STATUS)
		table_printf(tab, " %s+%sSTATUS", S_FG_GRN, S_OFF);
	else
		table_printf(tab, " %s-%sSTATUS", S_FG_RED, S_OFF);

	if(perm & AUTH_EXCLUDE)
		table_printf(tab, " %s+%sEXCLUDE", S_FG_GRN, S_OFF);
	else
		table_printf(tab, " %s-%sEXCLUDE", S_FG_RED, S_OFF);
}
