/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: cmd_signoff.c,v 1.13 2009/08/14 11:54:51 mickey Exp $ */

#include "vserver.h"

/*
 * Mark the calling client for regular signoff.
 */

void cmd_signoff(CLDATA *ptr, void *data, size_t len)
{
	if(data && len > 0)
		sncopy((char *)data, ptr->offmsg, sizeof(ptr->offmsg));
	else
		ptr->offmsg[0] = 0;

	ptr->flags |= CFLAG_SIGNOFF;
}

/*
 * Sigoff user from the conference.
 */

void signoff(CLDATA *ptr)
{
	uint32_t records = cldb_records();
	register CLDATA	*cp;
	char reason[64];
	char offmsg[128];

	/*
	 * Construct signoff reason string accordings to flags
	 * set on the CLDATA structure.
	 */

	if(ptr->flags & CFLAG_ERROR)
		snprintf(reason, sizeof(reason), "%s", strerror(ptr->error));
	else if(ptr->flags & CFLAG_KILL)
		snprintf(reason, sizeof(reason), "Killed by %s", ptr->killinfo);
	else if(ptr->flags & CFLAG_EXCLUDE)
		snprintf(reason, sizeof(reason), "Excluded by %s", ptr->killinfo);
	else if(ptr->flags & CFLAG_SHUTDOWN)
		snprintf(reason, sizeof(reason), "Server shutdown");
	else
		reason[0] = 0;

	/*
	 * Format the signoff message
	 */

	if(reason[0])
	{
		snprintf(offmsg, sizeof(offmsg), "* Signoff %s%s%s (%s%s%s)",
			S_FG_YEL, ptr->nick, S_OFF,
			S_FG_RED, reason, S_OFF);
	}
	else if(ptr->offmsg[0])
	{
		snprintf(offmsg, sizeof(offmsg), "* Signoff %s%s%s [%s]",
			S_FG_YEL, ptr->nick, S_OFF, ptr->offmsg);
	}
	else
	{
		snprintf(offmsg, sizeof(offmsg), "* Signoff %s%s%s", S_FG_YEL,
			ptr->nick, S_OFF);
	}

	/*
	 * Inform others about the user who is signing off
	 */

	if(records && (cp = cldb_addr(0)))
	{
		register int i;

		for(i = 0; i < records; i++, cp++)
		{
			/*
			 * Skip over the client for that signoff is being
			 * called.
			 */

			if(cp == ptr)
				continue;

			/*
			 * Clients with protocol revision of 4-10 and above
			 * receive ulist update messages to keep their userlist
			 * in sync with the server.
			 */

			if(cp->protocol >= 0x00040010)
			{
				vcp_snd(cp, ULIST_REMOVE, ptr->nick,
					ptr->nicklen);
			}

			/*
			 * All clients that have enabled signon/off messages
			 * receive the full signoff message. All others get
			 * a short signoff message, if they are on the current
			 * channel only.
			 */

			if(cp->cl_opts & CL_OPT_SGNMSG)
			{
				vcp_msg(cp, "%s", offmsg);
			}
			else if(cp->channel == ptr->channel)
			{
				vcp_msg(cp, "* %s%s%s signed off",
					S_FG_YEL, ptr->nick, S_OFF);
			}
		}
	}

	vlog(VLOG_INFO, "Signoff USER: %s NICK: %s HOST: %s MSG: %s REASON: %s",
		ptr->user, ptr->nick, ptr->host, ptr->offmsg, reason);

	/*
	 * Expire channel topic, if this user was the last one
	 * on that conference channel.
	 */

	topicdb_expire(ptr->channel);

	/*
	 * For registered nicknames, update the signoff
	 * timestamp and public- and private message counters.
	 */

	if(ptr->flags & CFLAG_REGISTERED)
	{
		USERDATA	*udat;

		if((udat = userdb_findaddr(ptr->nick)))
		{
			time(&udat->last_signoff);

			udat->pubmsgcnt += ptr->pubmsgcnt;
			udat->prvmsgcnt += ptr->prvmsgcnt;

			udat->cl_opts = ptr->cl_opts;

			userdb_save();
		}
		else if(errno == ENOENT)
		{
			/*
			 * Oops - registered client without a
			 * user database record?
			 */

			vlog(VLOG_NOTICE, "OOps: Registered nickname without user database record?!");
		}
		else
		{
			vlog(VLOG_ERR, "User database update failed (%s)",
				strerror(errno));
		}
	}

	/*
	 * Add a record to the lastlog database.
	 */

	lastlog_add_record(ptr, reason[0] ? reason : ptr->offmsg);

	/*
	 * Finally disconnect and delete the client.
	 */

#ifdef WIN32
	closesocket(ptr->fd);
#else
	close(ptr->fd);
#endif
	if(cldb_delete(ptr))
	{
		vlog(VLOG_ERR, "Failed to delete client database record (%s)",
			strerror(errno));
	}
}
