/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: cmd_signon.c,v 1.15 2009/08/14 11:54:51 mickey Exp $ */

#include "vserver.h"

/*
 * Signon a user to the conference.
 */

void cmd_signon(CLDATA *ptr)
{
	uint32_t records = cldb_records();
	register CLDATA	*cp;
	register int i;
	VTOPIC *vt;
	char onmsg[128];
	int32_t chan = htonl(ptr->channel);
	struct stat st;

	/*
	 * If the client that signs on has a protocol revision
	 * of 4-10 or above, then feed it with an initial username list.
	 */

	if(ptr->protocol >= 0x00040010)
	{
		vcp_snd(ptr, ULIST_CLEAR, NULL, 0);

		if(records && (cp = cldb_addr(0)))
		{
			for(i = 0; i < records; i++, cp++)
			{
				vcp_snd(ptr, ULIST_ADD, cp->nick, cp->nicklen);
			}
		}
	}

	/*
	 * Send a ulist update to all clients that have a protocol
	 * revision of 4-10 or above.
	 */

	if(records && (cp = cldb_addr(0)))
	{
		for(i = 0; i < records; i++, cp++)
		{
			if(cp != ptr && cp->protocol >= 0x00040010)
			{
				vcp_snd(cp, ULIST_ADD, ptr->nick, ptr->nicklen);
			}
		}
	}

	/*
	 * Inform the client of the current conference channel.
	 */

	vcp_snd(ptr, CHAN_ACK, &chan, sizeof(int32_t));

	/*
	 * Display welcome banner file (if present).
	 */

	if(!stat(fn_welcome, &st))
	{
		if(st.st_size < SVMSGBUF)
		{
			int fd;
			char buf[st.st_size];

			if((fd = open(fn_welcome, (O_RDONLY | O_BINARY | O_SEQUENTIAL))) != -1)
			{
				if(read(fd, buf, st.st_size) == st.st_size)
				{
					vcp_snd(ptr, SERV_MSG, buf, st.st_size);
				}
				else
				{
					vlog(VLOG_ERR, "Read error on welcome "
						"banner file '%s' (%s)",
						fn_welcome, strerror(errno));
				}

				close(fd);
			}
			else
			{
				vlog(VLOG_ERR, "Failed to open welcome banner "
					"file '%s' (%s)", fn_welcome,
					strerror(errno));
			}
		}
		else
		{
			vlog(VLOG_NOTICE, "Welcome banner file '%s' exceeds "
				"the maximum size of %d bytes",
				fn_welcome, SVMSGBUF-1);
		}
	}

	/*
	 * Spit out server info
	 */

	vcp_msg(ptr, "* Connected to %sV%sChat%s server at %s%s%s\n"
		"* %sV%sChat%s TCP/IP conference server %s%s%s / "
		"protocol revision %s%x-%x%s\n",
		S_FG_MAG, S_FG_YEL, S_OFF,
		S_FG_GRN, ptr->if_name, S_OFF,
		S_FG_MAG, S_FG_YEL, S_OFF,
		S_FG_CYN, PACKAGE_VERSION, S_OFF,
		S_FG_CYN, MAJOR_REV(PROTO_REV), MINOR_REV(PROTO_REV), S_OFF);

	/*
	 * Construct signon message
	 */

	if(!(vsopt.flags & VO_ANONYMOUS))
	{
		/*** Display client IP address ***/

		if(ptr->type == CLIENT_GATEWAY)
		{
			snprintf(onmsg, sizeof(onmsg),
				"* Signon %s%s%s from %s%s%s via %s%s%s",
				S_FG_YEL, ptr->nick, S_OFF,
				S_FG_GRN, ptr->host, S_OFF,
				S_FG_GRN, ptr->gw_name, S_OFF);
		}
		else
		{
			if(!strcmp(ptr->user, ptr->nick) || !ptr->user[0])
			{
				snprintf(onmsg, sizeof(onmsg),
					"* Signon %s%s%s from %s%s%s",
					S_FG_YEL, ptr->nick, S_OFF,
					S_FG_GRN, ptr->host, S_OFF);
			}
			else
			{
				snprintf(onmsg, sizeof(onmsg),
					"* Signon %s%s%s (%s%s%s) from %s%s%s",
					S_FG_MAG, ptr->user, S_OFF,
					S_FG_YEL, ptr->nick, S_OFF,
					S_FG_GRN, ptr->host, S_OFF);
			}
		}
	}
	else
	{
		/*** Suppress display of client IP addr ***/

		if(ptr->type == CLIENT_GATEWAY)
		{
			snprintf(onmsg, sizeof(onmsg),
				"* Signon %s%s%s via %s%s%s",
				S_FG_YEL, ptr->nick, S_OFF,
				S_FG_GRN, ptr->gw_name, S_OFF);
		}
		else
		{
			if(!strcmp(ptr->user, ptr->nick) || !ptr->user[0])
			{
				snprintf(onmsg, sizeof(onmsg),
					"* Signon %s%s%s",
					S_FG_YEL, ptr->nick, S_OFF);
			}
			else
			{
				snprintf(onmsg, sizeof(onmsg),
					"* Signon %s%s%s (%s%s%s)",
					S_FG_MAG, ptr->user, S_OFF,
					S_FG_YEL, ptr->nick, S_OFF);
			}
		}
	}

	/*
	 * Inform others about the user signing on.
	 */

	if(records && (cp = cldb_addr(0)))
	{
		for(i = 0; i < records; i++, cp++)
		{
			/*
			 * Send signon message to all clients that have
			 * enabled signon/off messages.
			 */

			if(cp->cl_opts & CL_OPT_SGNMSG)
			{
				vcp_msg(cp, "%s", onmsg);
			}

			/*
			 * Send channel join message to all clients on the
			 * channel the new user has just joined.
			 */

			if(cp != ptr && cp->channel == ptr->channel)
			{
				vcp_msg(cp, "* %s%s%s joined this channel",
					S_FG_YEL, ptr->nick, S_OFF);
			}
		}
	}

	/*
	 * Show date of last login if available
	 */

	if(ptr->lastlogin)
	{
		char	dtbuf[DT_SIZE];

		strftime(dtbuf, sizeof(dtbuf), DT_FMT, localtime(&ptr->lastlogin));
		dtbuf[sizeof(dtbuf)-1] = 0;

		vcp_msg(ptr, "* Last login %s", dtbuf);
	}

	/*
	 * Write log info
	 */

	vlog(VLOG_INFO, "Signon USER: %s NICK: %s CHANNEL: %d HOST: %s GATEWAY: %s DNSLVL: %d",
		ptr->user, ptr->nick, ptr->channel, ptr->host,
		(ptr->type == CLIENT_GATEWAY) ?
		ptr->gw_name : "<none>",
		ptr->dnschklvl);

	/*
	 * If there is currently no topic set for this channel
	 * then go ahead and request one from the client.
	 */

	vt = topicdb_findaddr(ptr->channel);

	if((!vt && errno == ENOENT) || (vt && !vt->topic[0]))
	{
		topicdb_set(ptr->channel, "<none>");
		vcp_snd(ptr, SUBMIT_TOPIC, NULL, 0);
	}
	else if(vt && vt->topic[0])
	{
		vcp_msg(ptr, "* Topic on this channel is '%s%s%s'",
			S_BOLD, vt->topic, S_OFF);
	}
	else
	{
		vcp_msg(ptr, "* Topic database failure (%s%s%s)",
			S_FG_RED, strerror(errno), S_OFF);
	}
}
