/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: cmd_status.c,v 1.14 2009/08/14 11:54:51 mickey Exp $ */

#include "vserver.h"

#if HAVE_SYS_RESOURCE_H
  #include <sys/resource.h>
#endif

/*
 * Send server status information
 */

void cmd_status(CLDATA *ptr, void *data, size_t len)
{
	uint32_t page;

	if(len == sizeof(uint32_t))
		page = ntohl(*(uint32_t *)data);
	else
		page = 0;

	/*
	 * Display status page
	 */

	switch(page)
	{
		case 0x00:	status_page0(ptr);
				break;
		case 0x01:	status_page1(ptr);
				break;
		case 0x02:	status_page2(ptr);
				break;
		case 0x03:	status_page3(ptr);
				break;
		default:	vcp_msg(ptr, "* Invalid server status page number %d", page);
				break;
	}
}

/*
 * STATUS PAGE 1:
 */

void status_page0(CLDATA *ptr)
{
	TABLE		tab;
	TMREPORT	*tmr;
	char		*osinfo;
	time_t		now;
	time_t		uptime;
	char		dtbuf[DT_SIZE];
	VIDENT		*sid = ident(0);

	table_init(&tab, ptr);

	/*
	 * Get server load average if possible
	 */

#if HAVE_GETLOADAVG
	getloadavg(vs.vs_loadavg, 3);
#endif

	/*
	 * Show status report
	 */

	table_printf(&tab, "\n*%s%s%s SERVER STATUS               %s GENERAL INFORMATION %s                    PAGE 0 %s\n*\n",
		S_FG_WHT, S_BG_BLU, S_BOLD, S_BG_RED, S_BG_BLU, S_OFF);

	time(&now);
	uptime = (now - vs.vs_start);

	strftime(dtbuf, sizeof(dtbuf), DT_FMT, localtime(&vs.vs_start));
	dtbuf[sizeof(dtbuf)-1] = 0;

	table_printf(&tab, "*   Server location: %s%s%s (%s%s%s)\n", S_BOLD,
		inet_ntoa(ptr->if_addr), S_OFF, S_UL, ptr->if_name, S_OFF);

	if((osinfo = get_os_info()) != NULL)
	{
		table_printf(&tab, "*  Operating system: %s%s%s\n",
			S_BOLD, osinfo, S_OFF);
	}

	table_printf(&tab, "*   Server up since: %s (%s)\n", dtbuf, elapsed_now(vs.vs_start));
#ifdef WIN32
	table_printf(&tab, "*        Process ID: %s%s%d%s\n", S_BOLD, S_BLINK, _getpid(), S_OFF);
#else
	table_printf(&tab, "*        Process ID: %s%s%d%s\n", S_BOLD, S_BLINK, getpid(), S_OFF);
#endif

#if HAVE_GETLOADAVG
	table_printf(&tab, "*     Load averages: %s%.2f%s / %s%.2f%s / %s%.2f%s (5/10/15 min)\n",
		S_FG_YEL, vs.vs_loadavg[0], S_OFF,
		S_FG_CYN, vs.vs_loadavg[1], S_OFF,
		S_FG_MAG, vs.vs_loadavg[2], S_OFF);
#endif
	tmr = mt_report();

	table_printf(&tab, "*  Command duration: %s%s%s / %s%s%s / %s%s%s (min/avg/max)\n",
		S_FG_YEL, mt_print(&tmr->_min), S_OFF,
		S_FG_CYN, mt_print(&tmr->_avg), S_OFF,
		S_FG_MAG, mt_print(&tmr->_max), S_OFF);

	table_printf(&tab, "*       Connections: %s%d%s / %s%d%s / %s%d%s (active/total/dropped)\n",
		S_FG_GRN, vs.vs_active_conn, S_OFF,
		S_FG_CYN, vs.vs_total_conn, S_OFF,
		S_FG_RED, vs.vs_err_conn, S_OFF);

	table_printf(&tab, "* Software revision: (%svserver%s %s%s%s - %s%s%s) %s[%s]%s",
		S_BOLD, S_OFF, S_FG_CYN, PACKAGE_VERSION, S_OFF,
		S_FG_RED, PACKAGE_RELDATE, S_OFF,
		S_BOLD, sid->ident, S_OFF);

	table_flush(&tab);
}

/*
 * STATUS PAGE 1:
 */

void status_page1(CLDATA *ptr)
{
	TABLE	tab;
	register int i;
	register int n;

	/*
	 * Status report of processed server requests
	 */

	table_init(&tab, ptr);

	table_printf(&tab, "\n*%s%s%s SERVER STATUS               %s REQUEST STATISTICS %s                     PAGE 1 %s\n*\n",
		S_FG_WHT, S_BG_BLU, S_BOLD, S_BG_RED, S_BG_BLU, S_OFF);

	table_printf(&tab, "*      %stotal%s: %s%-13d%s    %sunknown%s: %s%d%s\n",
		S_BOLD, S_OFF, S_FG_MAG, vs.vs_requests_total, S_OFF,
		S_BOLD, S_OFF, S_FG_RED, svcmd[0].requests, S_OFF);

	for(i = 1; i < svcmdlen; i += 3)
	{
		table_printf(&tab, "*");

		for(n = i; n < svcmdlen && n < i+3; n++)
		{
			table_printf(&tab, "%11s: %s%-13d%s", svcmd[n].name,
				S_FG_GRN, svcmd[n].requests, S_OFF);
		}

		if(n < (svcmdlen - 1))
			table_printf(&tab, "\n");
	}

	table_flush(&tab);
}

/*
 * STATUS PAGE 2:
 */

void status_page2(CLDATA *ptr)
{
	TABLE		tab;

	table_init(&tab, ptr);

	table_printf(&tab, "\n*%s%s%s SERVER STATUS                  %s MEMORY USAGE %s                        PAGE 2 %s\n*\n",
		S_FG_WHT, S_BG_BLU, S_BOLD, S_BG_RED, S_BG_BLU, S_OFF);

	rms_info(&tab);

	table_printf(&tab, "* %sLastlog db:%s\n", S_BOLD, S_OFF);

	table_printf(&tab, "*   %sstate%s=%s%s%s %sfile%s=%s%s%s\n",
		S_FG_CYN, S_OFF,
		S_FG_YEL, lastlog ? "open  " : "closed", S_OFF,
		S_FG_CYN, S_OFF,
		S_FG_YEL, fn_lastlog, S_OFF);

	if(lastlog)
	{
		table_printf(&tab, "*   %s in %s%d%s%s+%s%s1%s records of %s each",
			print_size(lastlog->size),
			S_FG_GRN, lastlog->header->total, S_OFF,
			S_BOLD, S_OFF, S_FG_RED, S_OFF,
			print_size(sizeof(LLOGDAT)));
#if HAVE_MMAP
		table_printf(&tab, " %s(%s%smemory mapped%s%s)%s",
			S_BOLD, S_OFF, S_FG_CYN, S_OFF, S_BOLD, S_OFF);
#endif
	}

	table_flush(&tab);
}

/*
 * STATUS PAGE 3:
 */

void status_page3(CLDATA *ptr)
{
	TABLE		tab;
#ifndef WIN32
	struct rusage	ru;
#endif
	table_init(&tab, ptr);

	table_printf(&tab, "\n*%s%s%s SERVER STATUS                 %s RESOURCE USAGE %s                       PAGE 3 %s\n*\n",
		S_FG_WHT, S_BG_BLU, S_BOLD, S_BG_RED, S_BG_BLU, S_OFF);
#ifndef WIN32
	if(!getrusage(RUSAGE_SELF, &ru))
	{
		table_printf(&tab, "*     %s%s%s real %s%d.%03d%s user %s%d.%03d%s sys\n*\n",
			S_FG_YEL, elapsed_now(vs.vs_start), S_OFF,
			S_FG_CYN, ru.ru_utime.tv_sec,
			ru.ru_utime.tv_usec / 10000, S_OFF,
			S_FG_MAG, ru.ru_stime.tv_sec,
			ru.ru_stime.tv_usec / 10000, S_OFF);

		table_printf(&tab, "*    Maximum resident set size: %s%-16d%s       Page faults: %s%d%s\n",
			S_FG_GRN, ru.ru_maxrss, S_OFF,
			S_FG_GRN, ru.ru_majflt, S_OFF);
		table_printf(&tab, "*     Average shared text size: %s%-16d%s     Page reclaims: %s%d%s\n",
			S_FG_GRN, ru.ru_ixrss, S_OFF,
			S_FG_GRN, ru.ru_minflt, S_OFF);
		table_printf(&tab, "*   Average unshared data size: %s%-16d%s   Swap operations: %s%d%s\n",
			S_FG_GRN, ru.ru_idrss, S_OFF,
			S_FG_GRN, ru.ru_nswap, S_OFF);
		table_printf(&tab, "*  Average unshared stack size: %s%-16d%s     Messages sent: %s%d%s\n",
			S_FG_GRN, ru.ru_isrss, S_OFF,
			S_FG_GRN, ru.ru_msgsnd, S_OFF);
		table_printf(&tab, "*    Blocking input operations: %s%-16d%s Messages received: %s%d%s\n",
			S_FG_GRN, ru.ru_inblock, S_OFF,
			S_FG_GRN, ru.ru_msgrcv, S_OFF);
		table_printf(&tab, "*   Blocking output operations: %s%-16d%s  Signals received: %s%d%s\n",
			S_FG_GRN, ru.ru_oublock, S_OFF,
			S_FG_GRN, ru.ru_nsignals, S_OFF);
		table_printf(&tab, "*   Voluntary context switches: %s%d%s\n", S_FG_GRN, ru.ru_nvcsw, S_OFF);
		table_printf(&tab, "* Involuntary context switches: %s%d%s", S_FG_GRN, ru.ru_nivcsw, S_OFF);
	}
	else
#endif
	{
		table_printf(&tab, "* No resource usage statistics available");
	}

	table_flush(&tab);
}
