/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: compat.h,v 1.5 2009/08/14 11:54:53 mickey Exp $ */

#ifndef __COMPAT_H__
#define __COMPAT_H__

/*
 * Suck in config.h generated by autoconf.
 */

#if HAVE_CONFIG_H
  #include "config.h"
#endif

/*
 * Define HAVE_SIGNALS if either sigaction() or sigvec() are present.
 */

#if HAVE_SIGACTION || HAVE_SIGVEC
  #define HAVE_SIGNALS 1
#else
  #define HAVE_SIGNALS 0
#endif


/*
 * Define EFTYPE and ENOATTR if not defined.
 */

#ifndef EFTYPE
  #define EFTYPE			79
#endif
#ifndef ENOATTR
  #define ENOATTR			87
#endif

/*
 * Define PATH_MAX if the system header files do not define it.
 */

#ifndef PATH_MAX
  #define	PATH_MAX		256
#endif

/*
 * Set up an appropriate directory separator character.
 */

#ifdef WIN32
  #define	DIRSEP			'\\'
#else
  #define	DIRSEP			'/'
#endif

/*
 * Define INADDR_NONE if the system header files do not define it.
 */

#ifndef INADDR_NONE
  #define	INADDR_NONE		0xffffffff
#endif

/*
 * Provide a macro mapping bcopy() to memcpy() for systems that
 * do not have a bcopy() call.
 */

#if !HAVE_BCOPY
  #if HAVE_MEMCPY
    #define	bcopy(a,b,c)		memcpy(b,a,c)
  #else
    #error "Neither bcopy nor memcpy are available"
  #endif
#endif

/*
 * Provide a macro mapping bzero() to memset() for systems that
 * do not have a bzero() call.
 */

#if !HAVE_BZERO
  #if HAVE_MEMSET
    #define	bzero(a,b)		memset(a,0,b)
  #else
    #error "Neither bzero nor memset are available"
  #endif
#endif

/*
 * Provide a macro mapping bcmp() to memcmp() for systems that
 * do not have a bcmp() call.
 */

#if !HAVE_BCMP
  #if HAVE_MEMCMP
    #define	bcmp(a,b,c)             memcmp(a,b,c)
  #else 
    #error "Neither bcmp nor memcmp are available"
  #endif
#endif

/*
 * Provide a macro mapping strchr() to index() for systems that
 * do not have a strchr() call.
 */

#if !HAVE_STRCHR
  #if HAVE_INDEX
    #define	strchr(a, b)		index(a, b)
  #else
    #error "Neither strchr nor index are available"
  #endif
#endif

/*
 * Provide a macro mapping strrchr() to rindex() for systems that
 * do not have a strrchr() call.
 */

#if !HAVE_STRRCHR
  #if HAVE_RINDEX
    #define	strrchr(a, b)		rindex(a, b)
  #else
    #error "Neither strrchr nor rindex are available"
  #endif
#endif

/*
 * Provide a macro mapping snprintf() to sprintf() for systems
 * that do not have an snprintf() call.
 */

#if !HAVE_SNPRINTF
  #define	snprintf(a, b, ...)	sprintf(a, __VA_ARGS__)
  #warning "sprintf() is being used for snprintf(). This is inherently unsafe"
#endif

/*
 * Provide a macro mapping vsnprintf() to vsprintf() for systems
 * that do not have a vsnprintf() call.
 */

#if !HAVE_VSNPRINTF
  #define	vsnprintf(a, b, c, d)	vsprintf(a, c, d);
  #warning "vsprintf() is being used for vsnprintf(). This is inherently unsafe"
#endif

/*
 * Define file modes O_BINARY, O_RANDOM and O_SEQUENTIAL
 * if they are not already defined.
 */

#ifndef O_BINARY
  #define O_BINARY	0
#endif
#ifndef O_RANDOM
  #define O_RANDOM	0
#endif
#ifndef O_SEQUENTIAL
  #define O_SEQUENTIAL	0
#endif

/*
 * Define the logging priorities. If syslog(3) is present, include
 * <syslog.h> and duplicate values, otherwise define them ourselves.
 */

#if HAVE_SYSLOG_H
  #include <syslog.h>
  #define	VLOG_EMERG	LOG_EMERG
  #define	VLOG_ALERT	LOG_ALERT
  #define	VLOG_CRIT	LOG_CRIT
  #define	VLOG_ERR	LOG_ERR
  #define	VLOG_WARNING	LOG_WARNING
  #define	VLOG_NOTICE	LOG_NOTICE
  #define	VLOG_INFO	LOG_INFO
  #define	VLOG_DEBUG	LOG_DEBUG
#else
  #define	VLOG_EMERG	0
  #define	VLOG_ALERT	1
  #define	VLOG_CRIT	2
  #define	VLOG_ERR	3
  #define	VLOG_WARNING	4
  #define	VLOG_NOTICE	5
  #define	VLOG_INFO	6
  #define	VLOG_DEBUG	7
#endif

/*
 * Define missing types.
 */

#if !HAVE_SSIZE_T
typedef ssize_t int;
#endif
#if !HAVE_OFF_T
typedef off_t int;
#endif
#if !HAVE_SOCKLEN_T
typedef socklen_t int;
#endif

#endif /* __COMPAT_H__ */
