/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: vcutil.h,v 1.4 2009/08/14 11:54:53 mickey Exp $ */

#ifndef __VCUTIL_H__
#define __VCUTIL_H__

/*
 * Number and size of static cyclic string buffers used by various
 * functions that format some kind of data and return a char pointer
 * to the result. These functions may be called STCB_NUM times
 * without overwriting the result of a previous call.
 */

#define STCB_NUM                8
#define STCB_SIZE               256

/*
 * TMREPORT:
 * =========
 *
 * The mt_report() function calculates the minimum, average and
 * maximum execution times from the values stored in the TMSTAT
 * structure and returns a structure of this type as the result.
 */

typedef struct
{
	struct	timeval	_min;
	struct	timeval _avg;
	struct	timeval _max;

} TMREPORT;

/*
 * TMSTAT:
 * =======
 *
 * While executing client commands, the server keeps track
 * of the execution time of each command and stores the last
 * NUMSTATVALS execution times within this structure for
 * statistic reports.
 */

typedef struct
{
#define NUMSTATVALS	512
	struct timeval	value[NUMSTATVALS];
	uint32_t	next;
	uint32_t	used;

} TMSTAT;

/*
 * Parameter/Qualifier definition structure.
 */

typedef struct
{
	uint32_t	id;
	char		*name;
	uint32_t	argtype;	/* Argument requirement & type	*/
#define	PQARG_NONE	0x00000000	/* - No argument		*/
#define	PQARG_OPTIONAL	0x00000001	/* - Optional argument		*/
#define	PQARG_MANDATORY	0x00000002	/* - Mandatory argument		*/

#define	PQARG_STRING	0x00000000	/* + String argument		*/
#define	PQARG_BOOL	0x00010000	/* + Boolean argument		*/
#define	PQARG_NUM	0x00020000	/* + Numerical argument		*/
#define	PQARG_IPADDR	0x00030000	/* + IP-address argument	*/
#define PQARG_NETADDR	0x00040000	/* + IP-address[/mask|/bits]	*/
#define	PQARG_PDATE	0x00050000	/* + Past date argument		*/
#define	PQARG_FDATE	0x00060000	/* + Future date argument	*/
	char		*argname;

} PQDEF;

struct pqlink;

typedef int (*PQFUNC)(int, char *[], struct pqlink *);

typedef struct pqlink
{
	PQDEF		*pqd;
	PQFUNC		pqf;
	struct pqlink	*link;
	size_t		nlinks;
	char		*desc;

} PQLINK;

#define	NPQL(x)		(sizeof(x) / sizeof(PQLINK))

/*
 * Macros for defining / querying parameters or qualifiers.
 */

#define PARAMETER(x)	(x & 0x0FFFFFFF)
#define	QUALIFIER(x)	((x & 0x0FFFFFFF) | 0x80000000)

/*
 * Parameter/Qualifier status indicators.
 */

#define PQS_INVALID	0x10000000
#define PQS_AMBIGIOUS	0x20000000
#define	PQS_NOARG	0x30000000
#define	PQS_REQARG	0x40000000
#define	PQS_NOPARSE	0x50000000
#define PQS_DATEERR	0x60000000

/*
 * Parameter/Qualifier argument type handling macros.
 */

#define PQARG_PRESENCE(x)	(x & 0x0000FFFF)
#define	PQARG_VALTYPE(x)	(x & 0xFFFF0000)

#define PQARG_FMT_PRESENCE(x)	(PQARG_PRESENCE(x) == PQARG_NONE ? "no" : PQARG_PRESENCE(x) == PQARG_OPTIONAL ? "an optional" : "a mandatory")
#define PQARG_FMT_VALTYPE(x)	(PQARG_VALTYPE(x) == PQARG_STRING ? "string" : PQARG_VALTYPE(x) == PQARG_BOOL ? "bool" : PQARG_VALTYPE(x) == PQARG_NUM ? "numeric" : PQARG_VALTYPE(x) == PQARG_IPADDR ? "ip-address" : PQARG_VALTYPE(x) == PQARG_NETADDR ? "ip-network" : PQARG_VALTYPE(x) == PQARG_PDATE ? "past date" : PQARG_VALTYPE(x) == PQARG_FDATE ? "future date" : "????")

/*
 * Parameter/Qualifier query macros.
 */

#define ISPARAMETER(x)	(!(x & 0x80000000))
#define ISQUALIFIER(x)	(x & 0x80000000)

#define PQ_INVALID(x)	((x & 0x70000000) == PQS_INVALID)
#define	PQ_AMBIGIOUS(x)	((x & 0x70000000) == PQS_AMBIGIOUS)
#define	PQ_NOARG(x)	((x & 0x70000000) == PQS_NOARG)
#define	PQ_REQARG(x)	((x & 0x70000000) == PQS_REQARG)
#define PQ_NOPARSE(x)	((x & 0x70000000) == PQS_NOPARSE)
#define PQ_DATEERR(x)	((x & 0x70000000) == PQS_DATEERR)

#define	PQ_EXHAUSTED	0xFFFFFFFF

/*
 * The DNS check level as determined by dns_chk().
 */

#define DNS_CHK_NONE	0x00	/* IP -> ???			*/
#define DNS_CHK_FWD	0x01	/* IP -> Name -> ???		*/
#define DNS_CHK_REV	0x02	/* IP -> Name -> another IP	*/
#define DNS_CHK_STRICT	0x03	/* IP -> Name -> same IP	*/

/*
 * Function prototypes.
 */

uint32_t	getnextpq(int argc, char *argv[], PQLINK *pql, PQLINK **result);
char		*pqfmt(PQDEF *pq, size_t *pqlen);
void		addpqpath(PQDEF *pqd);
void		resetpqpath(void);
void		pqsetinteractive(uint8_t onoff);
int		pqchkfail(uint32_t pqid);
int		pqcheckend(int argc, char *argv[]);
int		pqstdfunc(int argc, char *argv[], PQLINK *pql);
int		pqhelp(int argc, char *argv[], PQLINK *pql);
time_t		*parsedate(char *arg, int future);
size_t		sncopy(register char *src, register char *dst, size_t len);
char		*iptostr(struct in_addr ip);
char		*iptoname(struct in_addr ip);
char		*dnsname(struct in_addr ip);
char		*lowerstr(char *str);
char		*upperstr(char *str);
char		*trimstr(register char *str, size_t maxlen);
size_t		dspwidth(register char *str);
size_t		unescapestr(uint8_t *str, uint8_t *buf, size_t len);
void		tty_mode(int mode);
int		promptkey(char *validkeys, char *prompt, ...);
int		netmask_bits(struct in_addr netmask);
void		netmask_create(int bits, struct in_addr *netmask);
char		*elapsed(time_t clock1, time_t clock2);
char		*elapsed_now(time_t clock);
void		mt_stamp(struct timeval *tv);
struct		timeval *mt_elapsed(struct timeval *tv1, struct timeval *tv2);
struct		timeval *mt_elapsed_now(struct timeval *tv);
void		mt_store(struct timeval *tv);
char		*mt_print(struct timeval *tv);
TMREPORT	*mt_report(void);
void		mt_sum(struct timeval *tv, struct timeval *sum);
void		mt_avg(struct timeval *sum);
uint8_t		dns_check(struct in_addr *ip, char *namebuf, size_t maxsize);
char		*dnslvlname(uint8_t lvl);
uint8_t		dnslvlparse(char *arg);

/*
 * Extern declarations.
 */

  #ifndef __VCUTIL_C__
extern int		pqindex;
extern char		*pqname;
extern uint32_t		pqargtype;
extern char		*pqarg;
extern uint8_t		pqbool;
extern uint32_t		pqvalue;
extern struct in_addr	pqaddress;
extern struct in_addr	pqnetmask;
extern time_t		pqdate;

extern uint32_t		pqerrcount;

extern char		pqpath[PATH_MAX];
extern size_t		pqpathlen;
extern int		pqpathdepth;

extern TMSTAT		tmstat;
  #endif

#endif /* __VCUTIL_H__ */
