/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: vserver.h,v 1.26 2009/08/14 11:54:53 mickey Exp $ */

#ifndef __VSERVER_H__
#define __VSERVER_H__

/*
 * Suck in config.h generated by autoconf.
 */

#if HAVE_CONFIG_H
  #include "config.h"
#endif

/*
 * Include commonly used system header files.
 */

#include <stdio.h>
#if HAVE_SYS_TYPES_H
  #include <sys/types.h>
#endif
#if HAVE_SYS_STAT_H
  #include <sys/stat.h>
#endif
#if STDC_HEADERS
  #include <stdlib.h>
  #include <stddef.h>
#else
  #if HAVE_STDLIB_H
    #include <stdlib.h>
  #endif
#endif
#if HAVE_STRING_H
  #if !STDC_HEADERS && HAVE_MEMORY_H
    #include <memory.h>
  #endif
  #include <string.h>
#endif
#if HAVE_STRINGS_H
  #include <strings.h>
#endif
#if HAVE_INTTYPES_H
  #include <inttypes.h>
#else
  #if HAVE_STDINT_H
    #include <stdint.h>
  #endif
#endif
#if HAVE_UNISTD_H
  #include <unistd.h>
#endif

#if HAVE_STDARG_H
  #include <stdarg.h>
#else
  #if HAVE_VARARGS_H
    #include <varargs.h>
  #endif
#endif

#if HAVE_CTYPE_H
  #include <ctype.h>
#endif

#if HAVE_LIMITS_H
  #include <limits.h>
#endif

#if TIME_WITH_SYS_TIME
  #include <sys/time.h>
  #include <time.h>
#else
  #if HAVE_SYS_TIME_H
    #include <sys/time.h>
  #else
    #include <time.h>
  #endif
#endif

#if HAVE_ERRNO_H
  #include <errno.h>
#endif

#if HAVE_SETJMP_H
  #include <setjmp.h>
#endif

#if HAVE_SYS_UIO_H
  #include <sys/uio.h>
#endif

#if HAVE_SYS_SOCKET_H
  #include <sys/socket.h>
#endif
#if HAVE_NETINET_IN_H
  #include <netinet/in.h>
#endif
#if HAVE_ARPA_INET_H
  #include <arpa/inet.h>
#endif
#if HAVE_SYS_UN_H
  #include <sys/un.h>
#endif
#if HAVE_NETDB_H
  #include <netdb.h>
#endif
#if HAVE_FCNTL_H
  #include <fcntl.h>
#endif

/*
 * Windows only includes.
 */

#ifdef WIN32
  #include <windows.h>
  #include <process.h>
  #include <winsock.h>
#endif

/*
 * Include common VChat headers.
 */

#include "compat.h"
#include "vcutil.h"
#include "ascii.h"
#include "defaults.h"
#include "options.h"
#include "vproto/types.h"
#include "vctlproto.h"

/*
 * Program Qualifiers.
 */

#define		QR_ADDRESS	0x0101
#define		QN_CREATEUDB	0x0102
#define		QR_HOSTNAME	0x0103
#define		QR_LASTLOG	0x0104
#define		QO_LOGFILE	0x0105
#define		QR_MAXPENDING	0x0106
#define		QR_PORT         0x0107

#if HAVE_DAEMON
  #define	QN_DETACH	0x0201
#endif
#if HAVE_SYSLOG
  #define	QN_SYSLOG	0x0202
#endif

#ifdef WIN32
  #define	QR_CONFIG	0x0301
  #define	QN_INSTALL	0x0302
  #define	QN_UNINSTALL	0x0303
#else
  #define	QR_BASEDIR	0x0401
  #define	QR_ROTATELOG	0x0402	
#endif

/*
 * CFDATA:
 * =======
 *
 * Global server configuration data structure.
 *
 * This structure is initialized with compile-time defaults, which are then
 * overridden by any options specified on the commandline, loaded from the
 * windows registry or by state changes occurring at run-time.
 */

typedef struct
{
	char		basedir[PATH_MAX];
	char		logfile[PATH_MAX];
#ifdef WIN32
	char		svc_name[PATH_MAX];
	char		*svc_config;
#endif
	char		hostname[PATH_MAX];
	struct in_addr	bind_addr;	/* Interface address to bind to  */
	uint32_t	port;		/* TCP service port              */
	uint32_t	max_pending;	/* Max. # of pending requests    */
	uint32_t	rotate_log;	/* Number of logfile versions    */
	uint32_t	lastlog_recs;	/* Number of lastlog entries     */
	uint32_t	flags;		/* Configuration flags -v-v-v-   */
#define	CF_DETACH	0x00000001	/* Detach from terminal on start */
#define	CF_CREATE_UDB	0x00000002	/* Create new userdb and exit    */
#define CF_INSTALL	0x00000004	/* WIN32: Install service        */
#define	CF_UNINSTALL	0x00000008	/* WIN32: Uninstall service      */
#define	CF_WSA_INIT	0x00000010	/* WIN32: Winsock is initialized */
#define	CF_LOG_FILE	0x00000020	/* Enable logging to log file	 */
#define CF_LOG_SYSLOG	0x00000040	/* Enable logging to syslog(3)	 */
#define CF_SHUTDOWN	0x80000000	/* Server shutdown request flag  */

} CFDATA;

/*
 * SVOPT:
 * ======
 *
 * Server options permanently saved to disk.
 */

typedef struct
{
	uint32_t	flags;		/* Server option flags           */
#define	VO_ANONYMOUS	0x00000001	/* Anonymous server mode         */
#define	VO_ENABLE_MAIL	0x00000002	/* Enable E-Mail invitations     */
#define	VO_REGONLY	0x00000004	/* Registered users only         */
#define	VO_FLOODGUARD	0x00000008	/* Enable FloodGuard protection  */
	struct timeval	t_guard;	/* Flood guard threshold time    */
	uint16_t	dnschklvl;	/* Client DNS checking level     */

} VSOPT;

/*
 * VS_STAT:
 * ========
 *
 * This structure holds the server start time, the number
 * of connections served since startup, request counters
 * and other data for statistical purposes.
 */

typedef struct
{
	time_t		vs_start;		/* Server start time	*/
	uint32_t	vs_total_conn;		/* Total connections	*/
	uint32_t	vs_active_conn;		/* Active connections	*/
	uint32_t	vs_err_conn;		/* Erroneous conns	*/
	double		vs_loadavg[3];		/* Server load averages	*/
	uint32_t	vs_requests_total;	/* Total requests 	*/

} VS_STAT;

/*
 * ACLRULE:
 * ========
 *
 * The server maintains an access control list that is
 * composed of a number of rules of this type. Each rule
 * either accepts or rejects addresses matched by the
 * internet address and netmask values. A non zero
 * expires field specifies the time when the corresponding
 * rule automatically expires.
 */

typedef struct
{
	struct in_addr	address;		/* Internet address	*/
	struct in_addr	netmask;		/* Netmask		*/
	time_t		expires;		/* Time of expiration	*/
	uint16_t	action;			/* Rule action -v-v-v-  */
#define	ACL_REJECT	0x00			/* Reject if matched	*/
#define	ACL_ACCEPT	0x01			/* Accept if matched	*/
	uint16_t	rule;			/* Rule #0...65534	*/

} ACLRULE;

/*
 * USERDATA:
 * =========
 * 
 * The server keeps a record of this type within the
 * user database file for every registered user.
 */

typedef struct
{
	uint32_t	accperm;		/* Access permission mask    */
	uint32_t	modperm;		/* Modify permission mask    */
	time_t		account_created;	/* Time account was created  */
	time_t		last_signon;		/* Time of last signon	     */
	time_t		last_signoff;		/* Time of last signoff	     */
	uint32_t	sessioncnt;		/* Session counter	     */
	uint32_t	pubmsgcnt;		/* Public message counter    */
	uint32_t	prvmsgcnt;		/* Private message counter   */
	uint32_t	cl_opts;		/* Client option flags	     */
	char		nick[VPNICKSIZE+1];	/* Nickname		     */
	char		passwd[VPPASSWDSIZE+1];	/* Password		     */

} USERDATA;

/*
 * CLDATA:
 * =======
 *
 * Client data structure kept in memory
 * for every connected client.
 */

typedef struct
{
	int		fd;			/* Descriptor of I/O socket  */
	int		error;			/* Last errno if CFLAG_ERROR */
	uint32_t	type;			/* Type of client software   */
	uint32_t	protocol;		/* Client protocol revision  */
	uint32_t	flags;			/* Client flags  -v-v-v-     */
#define	CFLAG_REGISTERED	0x10000000	/* Nickname is registered    */
#define	CFLAG_SIGNOFF		0x20000000	/* Client is about signoff   */
#define	CFLAG_ERROR		0x40000000	/* Client had error          */
#define	CFLAG_KILL		0x80000000	/* Client has been killed    */
#define	CFLAG_EXCLUDE		0x01000000	/* Client has been excluded  */
#define	CFLAG_SHUTDOWN		0x02000000	/* Signoff due to shutdown   */
	uint32_t	cl_opts;		/* Client option flags       */
	uint32_t	accperm;		/* Access permission bits    */
	uint32_t	modperm;		/* CHMOD permission bits     */
	uint32_t	pubmsgcnt;		/* Public message counter    */
	uint32_t	prvmsgcnt;		/* Private message counter   */
	int32_t		channel;		/* Current channel number    */
	time_t		login;			/* Login timestamp           */
	time_t		lastlogin;		/* Last login timestamp      */
	time_t		stamp;			/* Last activity timestamp   */
	struct in_addr	ip_addr;		/* Client IP address         */
	struct in_addr	gw_addr;		/* Gateway address (if any)  */
	struct in_addr	if_addr;		/* Server interface address  */
	char		host[128];		/* The client's hostname     */
	char		gw_name[128];		/* Gateway name (if any)     */
	char		if_name[128];		/* Server interface hostname */
	char		user[VPUSERSIZE+1];	/* Username if FRONTEND      */
	char		nick[VPNICKSIZE+1];	/* Current nickname          */
	size_t		nicklen;		/* Length of nickname        */
	char		ident[VPIDENTSIZE+1];	/* Client identification     */
	char		killinfo[VPNICKSIZE+1]; /* Nickname of killing user  */
	char		offmsg[64];		/* Signoff message if any    */
	struct timeval	lastsend;		/* Time of last packet sent  */
	uint8_t		dnschklvl;		/* DNS check level           */

} CLDATA;

/*
 * Lastlog database structures:
 *
 * The server maintains a lastlog database file that
 * consists of a LLOGHEAD header and a number of
 * LLOGDAT records. When the server starts, the file
 * is memory-mapped and information about the mapped
 * database is entered into a global LASTLOG structure.
 * LLOGHEAD and LLOGDAT are both padded to 256 bytes
 * for improved performance.
 */

typedef struct
{
	int32_t		first;
	int32_t		next;
	uint32_t	used;
	uint32_t	total;
	uint8_t		padding[240];	/* Pad up to 256 bytes total	*/

} LLOGHEAD;

typedef struct
{
	time_t		signon_stamp;
	time_t		signoff_stamp;
	char		username[VPUSERSIZE+1];
	char		nickname[VPNICKSIZE+1];
	char		info[40];
	struct in_addr	ipaddr;
	char		hostname[64];
	char		padding[108];	/* Pad up to 256 bytes total	*/

} LLOGDAT;

typedef struct
{
	int		fd;		/* Lastlog database descriptor	*/
	size_t		size;		/* Lastlog database size	*/
	LLOGHEAD	*header;	/* Lastlog header pointer	*/
#if HAVE_MMAP
	LLOGDAT		*data;		/* Lastlog data pointer		*/
	void		*base;		/* Mapped region base address	*/
#endif
} LASTLOG;

/*
 * VTOPIC:
 * =======
 *
 * A data structure of this type is kept for every conference
 * channel, that has a topic assigned to it.
 */

typedef struct
{
	int32_t		channel;
	uint32_t	flags;
#define	VT_FIXED	0x00000001	/* Permanently stored until deleted */
	time_t		stamp;
	char		topic[VPTOPICSIZE+1];

} VTOPIC;

/*
 * TABLE:
 * ======
 *
 * Tables are used to collate multiple lines of text to be sent
 * to a client into a single SERV_MSG protocol message.
 */

typedef struct
{
	CLDATA		*client;
	uint32_t	used;

	VMSG		svcmd;
	char		buf[8184];

} TABLE;

/*
 * SVCMDFUNC:
 * ==========
 *
 * Server command type. All server command handler functions
 * adhere to this type.
 */

typedef void (*SVCMDFUNC)(CLDATA *, void *, size_t);

/*
 * SVCMD:
 * ======
 *
 * Server command structure.
 */

typedef struct
{
	uint32_t	requests;	/* Command request counter	*/
	char		*name;		/* Command name			*/
	SVCMDFUNC	func;		/* Command handler function	*/
	uint32_t	perm;		/* Required access permissions	*/

} SVCMD;

/*
 * RMS database numbering.
 */

#define RMS_USER	0
#define	RMS_ACL		1
#define	RMS_TOPIC	2
#define	RMS_CLIENT	3

/*
 * Flags for rms_open()
 */

#define RMS_O_CREATE   0x00000001      /* - Create file if necessary   */

/*
 * Win32 I/O operation structure
 */

#ifdef WIN32
typedef struct
{
	int	fd;
	void	*buf;
	size_t	len;

} IO_OP;
#endif

/*****************************************************************************
 * FUNCTION PROTOTYPES:                                                      *
 *****************************************************************************/

/*----------*
 | Commands |
 *----------*/

void		cmd_acl_add(CLDATA *ptr, void *data, size_t len);
void		cmd_acl_delete(CLDATA *ptr, void *data, size_t len);
void		cmd_acl_flush(CLDATA *ptr, void *data, size_t len);
void		cmd_acl_list(CLDATA *ptr, void *data, size_t len);
void		cmd_action(CLDATA *ptr, void *data, size_t len);
void		cmd_addnick(CLDATA *ptr, void *data, size_t len);
void		cmd_asstopic(CLDATA *ptr, void *data, size_t len);
void		cmd_broadcast(CLDATA *ptr, void *data, size_t len);
void		cmd_chanlist(CLDATA *ptr, void *data, size_t len);
void		cmd_chchan(CLDATA *ptr, void *data, size_t len);
void		cmd_chaccperm(CLDATA *ptr, void *data, size_t len);
void		change_perm(CLDATA *ptr, char which, void *data, size_t len);
void		cmd_chmodperm(CLDATA *ptr, void *data, size_t len);
void		cmd_chnick(CLDATA *ptr, void *data, size_t len);
void		cmd_chpasswd(CLDATA *ptr, void *data, size_t len);
void		cmd_chtopic(CLDATA *ptr, void *data, size_t len);
void		cmd_clientinfo(CLDATA *ptr, void *data, size_t len);
void		clinfo(CLDATA *cp, TABLE *tab, CLDATA *ptr);
void		cmd_clopt(CLDATA *ptr, void *data, size_t len);
void		display_clopt(TABLE *cltab, uint32_t opts);
void		cmd_delnick(CLDATA *ptr, void *data, size_t len);
void		cmd_exclude(CLDATA *ptr, void *data, size_t len);
void		cmd_fixchan(CLDATA *ptr, void *data, size_t len);
void		cmd_invite(CLDATA *ptr, void *data, size_t len);
void		cmd_kill(CLDATA *ptr, void *data, size_t len);
void		cmd_lastlog(CLDATA *ptr, void *data, size_t len);
void		cmd_page(CLDATA *ptr, void *data, size_t len);
void		cmd_prvmsg(CLDATA *ptr, void *data, size_t len);
void		cmd_pubmsg(CLDATA *ptr, void *data, size_t len);
void		cmd_query(CLDATA *ptr, void *data, size_t len);
void		display_permission1(TABLE *tab, uint32_t perm);
void		display_permission2(TABLE *tab, uint32_t perm);
void		cmd_rtcmp(CLDATA *ptr, void *data, size_t len);
void		cmd_setpasswd(CLDATA *ptr, void *data, size_t len);
void		cmd_signmsg(CLDATA *ptr, void *data, size_t len);
void		cmd_signoff(CLDATA *ptr, void *data, size_t len);
void		signoff(CLDATA *ptr);
void		cmd_signon(CLDATA *ptr);
void		cmd_status(CLDATA *ptr, void *data, size_t len);
void		status_page0(CLDATA *ptr);
void		status_page1(CLDATA *ptr);
void		status_page2(CLDATA *ptr);
void		status_page3(CLDATA *ptr);
void		cmd_svdnsopt(CLDATA *ptr, void *data, size_t len);
void		cmd_svopt(CLDATA *ptr, void *data, size_t len);
void		display_svopt(TABLE *cltab);
void		cmd_ulist(CLDATA *ptr, void *data, size_t len);
void		cmd_version(CLDATA *ptr, void *data, size_t len);
void		cmd_wakeup(CLDATA *ptr, void *data, size_t len);

/*-----------*
 | vserver.c |
 *-----------*/

int		main(int argc, char *argv[]);
int		Service(void);

/*-------------*
 | win32util.c |
 *-------------*/

#ifdef WIN32
int		GetRegistryData(void);
VOID WINAPI	ServiceMain(DWORD dwArgc, LPTSTR *lpszArgv);
VOID WINAPI	ServiceControlHandler(DWORD dwCtlCode);
int		InstallService(void);
int		UninstallService(void);
void		LogMessage(int pri, char *fmt, ...);
char		*winerror(DWORD code);
#endif

/*-------------*
 | serv_subr.c |
 *-------------*/

void		mkpaths(void);
void		mkdefaults(void);
void		default_options(void);
void		load_options(void);
void		save_options(void);
int		global_init(void);
void		global_exit(void);
int		db_init();
void		db_exit();
int		sock_init();
void		sock_exit();
void		pidfile(int on);
uint32_t	chatters(int32_t chan);
int		nick_ok(register char *nick);
VIDENT		*ident(uint32_t type);
int		nicksplit(TABLE *tab, char *list, char *argv[], size_t nmax, int wempty);
char		*print_size(size_t bytes);
char		*perm_str(uint32_t perm);
#if HAVE_SIGNALS
int		install_sighandlers(void);
void		uninstall_sighandlers(void);
char		*signame(int sig);
void		signal_handler(int sig);
#endif
int		process_cmdline(int argc, char *argv[]);

/*-------*
 | rms.c |
 *-------*/

void		rms_status(void);
void		rms_init(void);
void		rms_exit(void);
int		rms_open(uint32_t dbno, size_t recsize, const char *file, uint32_t flags, mode_t mode);
int		rms_close(uint32_t dbno);
void		*rms_add_rec(uint32_t dbno, void *buf);
int		rms_del_rec(uint32_t dbno, int recno);
ssize_t		rms_save(uint32_t dbno);
uint32_t	rms_records(uint32_t dbno);
void		*rms_addr(uint32_t dbno, int recno);
int		rms_index(uint32_t dbno, void *rec);
int		rms_reorg(void);
void		rms_discard(uint32_t dbno);
void		rms_info(TABLE *tab);

/*--------*
 | cldb.c |
 *--------*/

int		cldb_open(void);
void		cldb_close(void);
CLDATA		*cldb_add(CLDATA *rec);
int		cldb_delete(CLDATA *rec);
uint32_t	cldb_records(void);
CLDATA		*cldb_addr(int idx);
int		cldb_find(int fd);
CLDATA		*cldb_findaddr(int fd);
uint32_t	cldb_ison(char *nick, int32_t channel);
uint32_t	cldb_chkpresent(char *nick);
int		cldb_kill(char *nick, char *traitor);
int		cldb_killchannel(int32_t channel, char *traitor);
int		cldb_killall(char *traitor);
uint32_t	cldb_dnschklvl(uint8_t lvl);

/*----------*
 | vtopic.c |
 *----------*/

int		topicdb_open(void);
int		topicdb_close(void);
int		topicdb_find(int32_t channel);
int		topicdb_records(void);
VTOPIC		*topicdb_addr(int idx);
VTOPIC		*topicdb_findaddr(int32_t channel);
int		topicdb_fix(int32_t channel, char *topic);
int		topicdb_unfix(int32_t channel);
int		topicdb_set(int32_t channel, char *topic);
int		topicdb_expire(int32_t channel);
void		touch_channel(int32_t channel);

/*-------*
 | acl.c |
 *-------*/

int		acldb_open(void);
int		acldb_close(void);
ssize_t		acldb_save(void);
void		acldb_sort(void);
int		acldb_add(ACLRULE *rule);
int		acldb_delete(uint16_t rule);
int		acldb_records(void);
ACLRULE		*acldb_addr(int32_t idx);
int		acldb_find(uint16_t rule);
ACLRULE		*acldb_findaddr(uint16_t rule);
int		acldb_authorize(struct in_addr address, char *buf, size_t len);
void		acldb_delete_expired(void);
void		acldb_flush(void);
void		acldb_log_status(void);
void		acldb_print(ACLRULE *rule, char *buf, size_t len);

/*----------*
 | userdb.c |
 *----------*/

int		userdb_open(void);
void		userdb_close(void);
ssize_t		userdb_save(void);
int		userdb_add(char *nick, char *passwd);
int		userdb_delete(char *nick);
uint32_t	userdb_records(void);
USERDATA	*userdb_addr(int32_t idx);
int		userdb_find(char *nick);
USERDATA	*userdb_findaddr(char *nick);
int		userdb_create(void);

/*-----------*
 | lastlog.c |
 *-----------*/

void		lastlog_open(void);
int		lastlog_create(void);
void		lastlog_close(void);
#if HAVE_MMAP
void		lastlog_sync(void *addr, size_t len);
#endif
LLOGDAT		*lastlog_get_record(int num);
void		lastlog_add_record(CLDATA *cl, char *info);
int		lastlog_resize(void);

/*--------*
 | vlog.c |
 *--------*/

void		vlog_open(void);
void		vlog_close(void);
void		vlog(int pri, char *fmt, ...);
void		rotate_logfile(void);

/*---------*
 | table.c |
 *---------*/

void		table_init(TABLE *table, CLDATA *cl);
void		table_printf(TABLE *table, char *fmt, ...);
void		table_xprintf(TABLE *table, char *fmt, ...);
void		table_flush(TABLE *table);

/*------------*
 | vchatctl.c |
 *------------*/

#if HAVE_UNIX_SOCKETS
int		vchatctl_open(void);
void		vchatctl_close(void);
void		vchatctl_rcvcmd(void);
#endif

/*-------*
 | vcp.c |
 *-------*/

int		vcp_setupconn(void);
int		vcp_handshake(CLDATA *new);
int		vcp_login(CLDATA *new);
uint8_t		vcp_gwcheck(CLDATA *new, VLOGINDAT *vl);
int		vcp_auth(CLDATA *new, USERDATA *ud);
int		vcp_finish(CLDATA *new);
void		vcp_msg(CLDATA *cp, char *fmt, ...);
uint32_t	vcp_msgnick(char *nick, char *fmt, ...);
uint32_t	vcp_msgchan(int32_t channel, CLDATA *notme, char *fmt, ...);
uint32_t	vcp_broadcast(char *from, char *tty, char *fmt, ...);
void		vcp_snd(CLDATA *cp, uint32_t packtype, void *data, size_t len);

/*-----------*
 | os_info.c |
 *-----------*/

char		*get_os_info(void);

/*-------*
 | vio.c |
 *-------*/

ssize_t		vio_read(int fd, void *buf, size_t nbytes, time_t tout);
ssize_t		vio_write(int fd, void *buf, size_t nbytes, time_t tout);
#if HAVE_SYS_UIO_H
ssize_t		vio_writev(int fd, struct iovec *iov, int iovcnt, time_t tout);
#endif

/*---------*
 | usage.c |
 *---------*/

void		usage(char *fmt, ...);

/*
 * Include vserver external declarations.
 */

#ifndef __GLOBAL_C__
  #include "extern.h"
#endif

#endif /* __VSERVER_H__ */
