/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: lastlog.c,v 1.14 2009/08/14 11:54:51 mickey Exp $ */

#include "vserver.h"

#if HAVE_SYS_MMAN_H
  #include <sys/mman.h>
#endif

/*
 * Open the lastlog database file and map it to memory.
 * If the lastlog file does not already exist, it is created.
 */

void lastlog_open(void)
{
	int		fd;
	struct stat	st;
	static LASTLOG	desc;
#if HAVE_MMAP
	void		*addr;
#else
	static LLOGHEAD	llhead;
#endif

	/*
	 * Nothing to do if lastlog is disabled.
	 */

	if(cfdata.lastlog_recs <= 0)
		return;

	/*
	 * Open the lastlog database file.
	 */

	if((fd = open(fn_lastlog, (O_RDWR | O_BINARY | O_RANDOM))) == -1)
	{
		/*
		 * If the file could not be opened cause
		 * it does not exist, then create it.
		 */

		if(errno == ENOENT)
		{
			if((fd = lastlog_create()) == -1)
			{
				vlog(VLOG_CRIT, "Failed to create lastlog "
					"database '%s' - lastlog disabled",
					fn_lastlog);

				return;
			}
		}
		else
		{
			vlog(VLOG_CRIT, "Failed to open lastlog database '%s' "
				"(%s) - lastlog disabled",
				fn_lastlog, strerror(errno));

			return;
		}
	}

	/*
	 * Get the file size.
	 */

	if(fstat(fd, &st) == -1)
	{
		close(fd);

		vlog(VLOG_CRIT, "Failed to stat file '%s' (%s) - lastlog "
			"disabled", fn_lastlog, strerror(errno));

		return;
	}

#if HAVE_MMAP
	/*
	 * Map the lastlog database file into memory.
	 */

	if((addr = mmap(NULL, st.st_size, (PROT_READ | PROT_WRITE), MAP_SHARED,
		fd, 0)) == NULL)
	{
		close(fd);

		vlog(VLOG_ERR, "Failed to map lastlog database '%s' to memory "
			"(%s) - lastlog disabled",
			fn_lastlog, strerror(errno));

		return;
	}

	/*
	 * Give the system a hint about how we intend to use
	 * the mapped memory region.
	 */

  #if HAVE_MADVISE
	if(madvise(addr, st.st_size, MADV_SEQUENTIAL) == -1)
	{
		vlog(VLOG_WARNING, "Failed to madvise() memory at $%08x (%s)",
			addr, strerror(errno));
	}
  #endif
#else /* !HAVE_MMAP */
	/*
	 * Load the lastlog database header.
	 */

	if(read(fd, &llhead, sizeof(llhead)) != sizeof(llhead))
	{
		close(fd);

		vlog(VLOG_ERR, "Failed to read lastlog database header (%s) "
			"- lastlog disabled", strerror(errno));

		return;
	}
#endif
	/*
	 * All went fine - set up lastlog descriptor pointer.
	 */

	bzero(&desc, sizeof(desc));

	desc.fd		= fd;
	desc.size	= st.st_size;
#if HAVE_MMAP
	desc.base	= addr;
	desc.header	= (LLOGHEAD *)addr;
	desc.data	= (LLOGDAT *)(addr + sizeof(LLOGHEAD));

	lastlog = &desc;

	vlog(VLOG_INFO, "Lastlog database mapped (%d records)",
		lastlog->header->total);

#else
	desc.header	= &llhead;

	lastlog = &desc;

	vlog(VLOG_INFO, "Lastlog database opened (%d records)",
		lastlog->header->total);
#endif
}

/*
 * Unmap the lastlog database from memory and close
 * the underlying file descriptor.
 */

void lastlog_close(void)
{
	/*
	 * Nothing to do if no lastlog database mapped.
	 */

	if(!lastlog)
		return;

#if HAVE_MMAP
	/*
	 * Unmap the lastlog database from memory.
	 */

	if(munmap(lastlog->base, lastlog->size) == -1)
	{
		vlog(VLOG_ERR, "Failed to unmap lastlog database (%s)",
			strerror(errno));
	}
#endif

	/*
	 * Close the underlaying file descriptor.
	 */

	close(lastlog->fd);

	/*
	 * Clear lastlog descriptor pointer.
	 */

	lastlog = (LASTLOG *)0;

	/*
	 * Transaction done.
	 */

#if HAVE_MMAP
	vlog(VLOG_INFO, "Lastlog database unmapped");
#else
	vlog(VLOG_INFO, "Lastlog database closed");
#endif
}

/*
 * Create a new lastlog database file.
 * Returns the file descriptor of the newly created lastlog
 * database file, or -1 in case of failure.
 */

int lastlog_create(void)
{
	LLOGHEAD	llhead;
	LLOGDAT		lldat;
	int		fd;
	register int	i;

	/*
	 * Initialize lastlog header and data structures.
	 */

	bzero(&llhead, sizeof(llhead));
	llhead.total = cfdata.lastlog_recs;
	bzero(&lldat, sizeof(lldat));

	/*
	 * Open lastlog database file for output.
	 */

	if((fd = open(fn_lastlog, (O_RDWR | O_CREAT | O_BINARY | O_SEQUENTIAL),
		0640)) == -1)
	{
		vlog(VLOG_ERR, "Failed to create lastlog database '%s' (%s)",
			fn_lastlog, strerror(errno));
		return -1;
	}

	/*
	 * Write header structure to file.
	 */

	if(write(fd, &llhead, sizeof(llhead)) != sizeof(llhead))
	{
		close(fd);
		unlink(fn_lastlog);

		vlog(VLOG_ERR, "Write failed while creatng lastlog database "
			"'%s' (%s)", fn_lastlog, strerror(errno));

		return -1;
	}

	/*
	 * Write empty data entries to file.
	 */

	for(i = 0; i < cfdata.lastlog_recs; i++)
	{
		if(write(fd, &lldat, sizeof(lldat)) != sizeof(lldat))
		{
			close(fd);
			unlink(fn_lastlog);

			vlog(VLOG_ERR, "Write failed while creatng lastlog "
				"database '%s' (%s)", fn_lastlog, strerror(errno));

			return -1;
		}
	}

	/*
	 * Seek to offset 0.
	 */

	if(lseek(fd, 0, SEEK_SET) == -1)
	{
		close(fd);
		unlink(fn_lastlog);

		vlog(VLOG_ERR, "Failed to seek within lastlog file '%s' (%s)",
			fn_lastlog, strerror(errno));

		return -1;
	}

	/*
	 * Transaction done, return file descriptor.
	 */

	return fd;
}

#if HAVE_MMAP
/*
 * Sync the region of the lastlog database that is specified
 * by <addr> and <len> to the underlying file.
 */

void lastlog_sync(void *addr, size_t len)
{
	/*
	 * Nothing to sync if no lastlog database mapped.
	 */

	if(!lastlog)
		return;

	/*
	 * Sync specified region of lastlog database.
	 */

	if(msync(addr, len, MS_SYNC) == -1)
	{
		vlog(VLOG_ERR, "Failed to sync %d bytes of lastlog database "
			"memory at address $%08x (%s)",
			len, addr, strerror(errno));
	}

	/*
	 * Sync lastlog database header.
	 */

	if(msync(lastlog->base, sizeof(LLOGHEAD), MS_SYNC) == -1)
	{
		vlog(VLOG_ERR, "Failed to sync lastlog database header (%s)",
			strerror(errno));
	}
}
#endif	/* HAVE_MMAP */

/*
 * Return the address of the database record specified by <num>,
 * or NULL in case the specified record number is out of bounds
 * or no lastlog database is present.
 */

LLOGDAT *lastlog_get_record(int num)
{
#if !HAVE_MMAP
  #define	CACHE_RECS	32
	static	LLOGDAT		lldat[CACHE_RECS];
	static	uint32_t	cache_start	= 0;
	static	uint32_t	cache_len	= 0;
	off_t			offset;
#endif
	/*
	 * No lastlog - no record.
	 */

	if(!lastlog)
		return (LLOGDAT *)0;

	/*
	 * Check for record number out of bounds.
	 */

	if(num < 0 || num >= lastlog->header->used)
		return (LLOGDAT *)0;

	/*
	 * Translate logical to physical record number.
	 */

	if((lastlog->header->first + num) >= lastlog->header->total)
		num -= lastlog->header->first;

#if HAVE_MMAP
	/*
	 * Return address of record.
	 */

	return lastlog->data + num;
#else
	/*
	 * Check if the requested record is in cache.
	 */

	if(num >= cache_start && num < (cache_start + cache_len))
	{
		/*
		 * Record is in cache - return it.
		 */

		return &lldat[num - cache_start];
	}

	/*
	 * Record is not in cache.
	 * Determine the cache start record.
	 */

	cache_start = num >= CACHE_RECS ? num - (CACHE_RECS-1) : 0;

	/*
	 * Calculate file offset from cache starting record.
	 */

	offset = sizeof(LLOGHEAD) + (cache_start * sizeof(LLOGDAT));

	/*
	 * Determine the maximum number of records we may load into the cache.
	 */

	cache_len = (offset + sizeof(lldat)) > lastlog->size ?
		(lastlog->size - offset) / sizeof(LLOGDAT) : CACHE_RECS;

	/*
	 * Fill cache.
	 */

#if 0
	vlog(VLOG_INFO, "Cache miss rec=%d start=%d num=%d off=%d",
		num, cache_start, cache_len, offset);
#endif

	if(lseek(lastlog->fd, offset, SEEK_SET) != offset)
	{
		close(lastlog->fd);
		cache_start = cache_len = 0;
		lastlog = (LASTLOG *)0;

		vlog(VLOG_ERR, "Seek error on lastlog database (%s) "
			"- lastlog disabled", strerror(errno));

		return (LLOGDAT *)0;
	}

	if(read(lastlog->fd, &lldat[0], cache_len * sizeof(LLOGDAT))
		!= cache_len * sizeof(LLOGDAT))
	{
		close(lastlog->fd);
		lastlog = (LASTLOG *)0;

		vlog(VLOG_ERR, "Read error on lastlog database (%s)"
			"- lastlog disabled", strerror(errno));

		cache_start = cache_len = 0;

		return (LLOGDAT *)0;
	}

	return &lldat[num - cache_start];
#endif
}

/*
 * Add a new entry to the lastlog database for the client specified
 * by <cl> with the information string <info>.
 */

void lastlog_add_record(CLDATA *cl, char *info)
{
	LLOGDAT	ldat;
#if HAVE_MMAP
	LLOGDAT	*lptr;
#else
	off_t	offset;
#endif

	/*
	 * Cannot add entry without database mapped.
	 */

	if(!lastlog)
		return;

	/*
	 * Construct the new lastlog entry
	 */

	bzero(&ldat, sizeof(ldat));

	ldat.signon_stamp = cl->login;
	time(&ldat.signoff_stamp);

	ldat.ipaddr.s_addr = cl->ip_addr.s_addr;

	sncopy(cl->user, ldat.username, sizeof(ldat.username));
	sncopy(cl->nick, ldat.nickname, sizeof(ldat.nickname));
	sncopy(cl->host, ldat.hostname, sizeof(ldat.hostname));

	if(info)
		sncopy(info, ldat.info, sizeof(ldat.info));

	/*
	 * Insert the new entry into the database.
	 */

	if(lastlog->header->used < lastlog->header->total)
	{
		++lastlog->header->used;
	}
	else
	{
		/*
		 * Cut off one entry
		 */

		if(++lastlog->header->first >= lastlog->header->total)
			lastlog->header->first = 0;
	}

#if HAVE_MMAP
	lptr = lastlog->data + lastlog->header->next;
#else
	offset = sizeof(LLOGHEAD) + (lastlog->header->next * sizeof(LLOGDAT));
#endif

	if(++lastlog->header->next >= lastlog->header->total)
		lastlog->header->next = 0;

#if HAVE_MMAP
	bcopy(&ldat, (void *)lptr, sizeof(ldat));

	/*
	 * Sync the modified region of the database.
	 */

	lastlog_sync((void *)lptr, sizeof(ldat));

#else /* !HAVE_MMAP */

	/*
	 * Write new record to database.
	 */

	if(lseek(lastlog->fd, offset, SEEK_SET) != offset)
	{
		close(lastlog->fd);
		lastlog = (LASTLOG *)0;

		vlog(VLOG_ERR, "Seek error on lastlog database (%s) "
			"- lastlog disabled", strerror(errno));
		return;
	}

	if(write(lastlog->fd, &ldat, sizeof(ldat)) != sizeof(ldat))
	{
		close(lastlog->fd);
		lastlog = (LASTLOG *)0;

		vlog(VLOG_ERR, "Write error on lastlog database (%s) "
			"- lastlog disabled", strerror(errno));

		return;
	}

	/*
	 * Write back lastlog database header.
	 */

	if(lseek(lastlog->fd, 0, SEEK_SET) != 0)
	{
		close(lastlog->fd);
		lastlog = (LASTLOG *)0;

		vlog(VLOG_ERR, "Seek error on lastlog database (%s) "
			"- lastlog disabled", strerror(errno));
		return;
	}

	if(write(lastlog->fd, lastlog->header, sizeof(LLOGHEAD)) != sizeof(LLOGHEAD))
	{
		close(lastlog->fd);
		lastlog = (LASTLOG *)0;

		vlog(VLOG_ERR, "Write error on lastlog database (%s) "
			"- lastlog disabled", strerror(errno));

		return;
	}
#endif
}

/*
 * Resize the currently mapped lastlog database if necessary.
 */

int lastlog_resize(void)
{
	register int	i;
	int		fd;
	int		oldsize;
	char		backup[PATH_MAX];
	char		newfile[PATH_MAX];
	LLOGHEAD	newhead;
	LLOGDAT		empty;
	LLOGDAT		*recptr;
	struct stat	st;

	/*
	 * No lastlog database - no action.
	 */

	if(!lastlog)
		return 0;

	/*
	 * Check if the lastlog database that is currently
	 * mapped needs to be resized to match the configured
	 * number of records.
	 */

	if((oldsize = lastlog->header->total) == cfdata.lastlog_recs)
	{
		/*
		 * Nope.
		 */

		return 0;
	}

	vlog(VLOG_INFO, "Lastlog database needs to be resized");

	/*
	 * Create the new file we will be writing to.
	 */

	snprintf(newfile, sizeof(newfile), "%s.new", fn_lastlog);

	if((fd = open(newfile, (O_WRONLY | O_CREAT | O_BINARY | O_SEQUENTIAL),
		0640)) == -1)
	{
		vlog(VLOG_ERR, "Failed to create file '%s' (%s)",
			newfile, strerror(errno));
		return -1;
	}

	/*
	 * Write an appropriate lastlog database header to the new file.
	 */

	bzero(&newhead, sizeof(newhead));

	newhead.total	= cfdata.lastlog_recs;
	newhead.used	= lastlog->header->used > cfdata.lastlog_recs ?
				cfdata.lastlog_recs : lastlog->header->used;
	newhead.first	= 0;
	newhead.next	= newhead.used < newhead.total ? newhead.used : 0;

	if(write(fd, &newhead, sizeof(newhead)) != sizeof(newhead))
	{
		close(fd);
		unlink(newfile);

		vlog(VLOG_ERR, "Write failed on file '%s' (%s)",
			newfile, strerror(errno));

		return -1;
	}

	/*
	 * Write records from the current lastlog database to the new
	 * file, filling it up with empty records if necessary.
	 */

	bzero(&empty, sizeof(empty));

	for(i = 0; i < cfdata.lastlog_recs; i++)
	{
		if(i < lastlog->header->used)
		{
			/*
			 * Use existing record if possible.
			 */

			if((recptr = lastlog_get_record(i)) == NULL)
				recptr = &empty;
		}
		else
		{
			/*
			 * Fill with empty record.
			 */

			recptr = &empty;
		}

		/*
		 * Write record to new file.
		 */

		if(write(fd, recptr, sizeof(LLOGDAT)) != sizeof(LLOGDAT))
		{
			close(fd);
			unlink(newfile);

			vlog(VLOG_ERR, "Write failed on file '%s' (%s)",
				newfile, strerror(errno));

			return -1;
		}
	}

	/*
	 * Close the new file.
	 */

	close(fd);

	/*
	 * Unmap the current lastlog database.
	 */

	lastlog_close();

	/*
	 * Find next available backup file version.
	 */

	bzero(&backup, sizeof(backup));

	i = 0;

	do
	{
		++i;
		snprintf(backup, sizeof(backup), "%s;%d", fn_lastlog, i);
	} while (stat(backup, &st) == 0);

	/*
	 * Rename current lastlog database file to backup version.
	 */

	if(rename(fn_lastlog, backup))
	{
		/*
		 * Rename failed, revert to using the old lastlog
		 * database file.
		 */

		vlog(VLOG_ERR, "Failed to rename '%s' to '%s' (%s)",
			fn_lastlog, backup, strerror(errno));

		lastlog_open();
		unlink(newfile);

		return -1;
	}

	vlog(VLOG_INFO, "Created lastlog database backup '%s'", backup);

	/*
	 * Move newly created lastlog database file in place.
	 */

	if(rename(newfile, fn_lastlog))
	{
		vlog(VLOG_CRIT, "Failed to rename '%s' to '%s' (%s) "
			"- lastlog disabled - please move the file manually "
			"and restart vserver afterwards",
			newfile, fn_lastlog, strerror(errno));

		return -1;
	}

	vlog(VLOG_INFO, "Resized lastlog database (%d -> %d records)",
		oldsize, cfdata.lastlog_recs);

	/*
	 * Map the new lastlog database file.
	 */

	lastlog_open();

	/*
	 * Transaction done.
	 */

	return 0;
}
