/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: os_info.c,v 1.9 2009/08/14 11:54:52 mickey Exp $ */

#include "vserver.h"

#if HAVE_SYS_PARAM_H
  #include <sys/param.h>
#endif

#if HAVE_SYS_SYSCTL_H
  #include <sys/sysctl.h>
#else
  #if HAVE_SYS_SYSTEMINFO_H
    #include <sys/systeminfo.h>
  #else
    #if HAVE_SYS_UTSNAME_H
      #include <sys/utsname.h>
    #endif
  #endif
#endif

/*
 * Retrieve various information about the host operating system
 * and return a descriptive string. If no information could be
 * retrieved, NULL is returned.
 */

char *get_os_info(void)
{
	static char	osinfo[96];

#if HAVE_SYSCTL && HAVE_SYS_SYSCTL_H

	char	os_type[24];
	char	os_release[24];
	char	hw_model[24];
	int	ncpu;
	size_t	len;
	int	mib[2];

	bzero(os_type, sizeof(os_type));
	bzero(os_release, sizeof(os_release));
	bzero(hw_model, sizeof(hw_model));
	ncpu = -1;

	/*** Get OS type ***/

	mib[0] = CTL_KERN;
	mib[1] = KERN_OSTYPE;

	len = sizeof(os_type);

	sysctl(mib, 2, os_type, &len, NULL, 0);

	/*** Get OS release ***/

	mib[0] = CTL_KERN;
	mib[1] = KERN_OSRELEASE;

	len = sizeof(os_release);

	sysctl(mib, 2, &os_release, &len, NULL, 0);

  #if defined(CTL_HW)
    #if defined(HW_MODEL)

	/*** Get hardware model ***/

	mib[0] = CTL_HW;
	mib[1] = HW_MODEL;

	len = sizeof(hw_model);;

	sysctl(mib, 2, &hw_model, &len, NULL, 0);

    #endif

    #if defined(HW_NCPU)

	/**** Get number of CPUs installed ***/

	mib[0] = CTL_HW;
	mib[1] = HW_NCPU;

	len = sizeof(ncpu);

	sysctl(mib, 2, &ncpu, &len, NULL, 0);

    #endif
  #endif

	/*
	 * Format descriptive text.
	 */

	if(ncpu > 0)
	{
		snprintf(osinfo, sizeof(osinfo), "%s %s%s%s %d CPU%s", os_type,
			os_release, hw_model[0] ? " " : "", hw_model,
			ncpu, ncpu > 1 ? "s" : "");
	}
	else
	{
		snprintf(osinfo, sizeof(osinfo), "%s %s%s%s", os_type,
			os_release, hw_model[0] ? " " : "", hw_model);
	}

	return osinfo;

#else
  #if HAVE_SYSTEMINFO && HAVE_SYS_SYSTEMINFO_H

	char sysname[24];
	char hostname[24];
	char release[24];
	char version[24];
	char machine[24];
	char architecture[24];

	if(sysinfo(SI_SYSNAME, sysname, sizeof(sysname)) == -1
	|| sysinfo(SI_HOSTNAME, hostname, sizeof(hostname)) == -1
	|| sysinfo(SI_RELEASE, release, sizeof(release)) == -1
	|| sysinfo(SI_VERSION, version, sizeof(version)) == -1
	|| sysinfo(SI_MACHINE, machine, sizeof(machine)) == -1
	|| sysinfo(SI_ARCHITECTURE, architecture, sizeof(architecture)) == -1)
	{
		return NULL;
	}
	else
	{
		snprintf(osinfo, sizeof(osinfo), "%s %s %s %s %s %s",
			sysname, hostname, release, version,
			machine, architecture);

		return osinfo;
	}
  #else
    #if HAVE_UNAME && HAVE_SYS_UTSNAME_H

	struct utsname uts;

	if(uname(&uts))
	{
		return NULL;
	}
	else
	{
		snprintf(osinfo, sizeof(osinfo), "%s %s %s", uts.sysname,
			uts.release, uts.machine);

		return osinfo;
	}
    #else
      #ifdef WIN32
	HKEY	hKey;
	DWORD	dwValIdx	= 0;
	TCHAR	valname[MAX_PATH];
	DWORD	valnamesize	= MAX_PATH;
	BYTE	valbuf[MAX_PATH];
	DWORD	valbufsize	= MAX_PATH;
	DWORD	valtype;

	char	osname[32]	= "WindowsNT";
	char	osrev[6]	= "???";
	char	svcpack[32]	= "\0";
	char	buildnum[10]	= "???";

#define	INFO_KEY1	"SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion"
#define	INFO_KEY2	"SYSTEM\\CurrentControlSet\\Control\\ProductOptions"

	/*
	 * Open the first registry key that contains
	 * some of the infos we want to know.
	 */

	if(!RegOpenKeyEx(HKEY_LOCAL_MACHINE,INFO_KEY1,0,KEY_QUERY_VALUE,&hKey))
	{
		/*
		 * Step through all values for this key and
		 * watch out for the ones we want to know.
		 */

		while(!RegEnumValue(hKey, dwValIdx, valname, &valnamesize,
			NULL, &valtype, valbuf, &valbufsize))
		{
			if(!strcmp(valname, "CSDVersion")
			&& (valtype == REG_EXPAND_SZ || valtype == REG_SZ))
			{
				sncopy((char *)valbuf, svcpack,
					sizeof(svcpack));
			}
			else if(!strcmp(valname, "CurrentBuildNumber")
			&& (valtype == REG_EXPAND_SZ || valtype == REG_SZ))
			{
				sncopy((char *)valbuf, buildnum,
					sizeof(buildnum));
			}
			else if(!strcmp(valname, "CurrentVersion")
			&& (valtype == REG_EXPAND_SZ || valtype == REG_SZ))
			{
				sncopy((char *)valbuf, osrev,
					sizeof(osrev));
			}

			valnamesize = valbufsize = MAX_PATH;
			++dwValIdx;
		}

		/*
		 * Close the first registry key.
		 */

		RegCloseKey(hKey);
	}
	else
	{
		vlog(VLOG_ERR, "Failed to open registry key '%s'", INFO_KEY1);
	}

	/*
	 * Open the second registry key that contains
	 * some other infos we want to know.
	 */

	if(!RegOpenKeyEx(HKEY_LOCAL_MACHINE,INFO_KEY2,0,KEY_QUERY_VALUE,&hKey))
	{
		/*
		 * Step through all values for this key and
		 * watch out for the ones we want to know.
		 */

		valnamesize = valbufsize = MAX_PATH;
		dwValIdx = 0;

		while(!RegEnumValue(hKey, dwValIdx, valname, &valnamesize,
			NULL, &valtype, valbuf, &valbufsize))
		{
			if(!strcmp(valname, "ProductType")
			&& (valtype == REG_EXPAND_SZ || valtype == REG_SZ))
			{
				sncopy((char *)valbuf, osname,
					sizeof(osname));
			}

			valnamesize = valbufsize = MAX_PATH;
			++dwValIdx;
		}

		/*
		 * Close the key.
		 */

		RegCloseKey(hKey);
	}
	else
	{
		vlog(VLOG_ERR, "Failed to open registry key '%s'", INFO_KEY2);
	}

	snprintf(osinfo, sizeof(osinfo), "%s %s Build: %s %s",
		osname, osrev, buildnum, svcpack);

	return osinfo;
      #else
	return NULL;
        #warning os-info not supported on this system
      #endif
    #endif
  #endif
#endif
}
