/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: rms.c,v 1.5 2009/08/14 11:54:52 mickey Exp $ */

#include "vserver.h"

/*
 * Static RMS configuration parameters.
 * These may possibly be configurable
 * per database in later RMS revisions.
 */

#define	RMS_MAX_DB	4		/* Max. number of databases	*/
#define RMS_SEGSIZE	8		/* Records per segment		*/
#define	RMS_RETAIN	10		/* Records retained before	*/
					/* segment release		*/
/*
 * The RMS database descriptor structure.
 */

typedef struct
{
	uint32_t	flags;		/* Database flags		*/
	void		*base;		/* Memory base address		*/
	size_t		size;		/* Total memory allocated	*/

	const char	*file;		/* File associated with DB	*/

	uint32_t	s_recs;		/* Records per segment		*/
	uint32_t	s_size;		/* Segment size in bytes	*/
	uint32_t	s_total;	/* Total number of segments	*/
	uint32_t	s_retain;	/* Segment retain threshold	*/

	uint32_t	r_size;		/* Record size			*/
	uint32_t	r_total;	/* Total records		*/
	uint32_t	r_used;		/* Used records			*/
	uint32_t	r_free;		/* Free records			*/

	time_t		reorg;		/* Reorganization timestamp	*/

} RMSDB;

/*
 * RMS internal per database flags.
 */

#define RMS_DB_OPEN	0x80000000	/* Database is in open state	*/

/*
 * Mask of per database flags that is accepted by the <flags>
 * parameter of the rms_open() function and transferred into
 * the database descriptor flags. All bits of the database
 * descriptor flags not contained in this mask, are reserved
 * for RMS internal use.
 */

#define RMS_OPEN_MASK	(RMS_O_CREATE)

/*
 * Global RMS state.
 */

uint32_t	rms_state	= 0x00000000;
#define	RMS_S_INIT		0x00000001

/*
 * RMS database descriptors.
 */

RMSDB		rmsdb[RMS_MAX_DB];

/*
 * Log status information about all databases.
 */

void rms_status(void)
{
	register int dbno;
	register RMSDB *db;

	for(dbno = 0; dbno < RMS_MAX_DB; dbno++)
	{
		db = &rmsdb[dbno];

		vlog(VLOG_DEBUG, "RMS database %d: flags=$%08X base=$%08X size=%d "
			"segments: %d segsize=%d recsize=%d records total=%d used=%d "
			"free=%d file='%s'", dbno, db->flags, db->base, db->size,
			db->s_total, db->s_size, db->r_size, db->r_total, db->r_used,
			db->r_free, db->file ? db->file : "<none>");
	}
}

/*
 * Initialize RMS.
 */

void rms_init(void)
{
	if(rms_state & RMS_S_INIT)
	{
		vlog(VLOG_WARNING, "RMS already initialized");
		return;
	}

	bzero(&rmsdb, sizeof(rmsdb));
	rms_state |= RMS_S_INIT;

	vlog(VLOG_INFO, "RMS initialized");

	return;
}

/*
 * De-Initialize RMS and close all possibly open databases.
 */

void rms_exit(void)
{
	if(rms_state & RMS_S_INIT)
	{
		register int i;

		for(i = 0; i < RMS_MAX_DB; i++)
		{
			if(rmsdb[i].flags & RMS_DB_OPEN)
			{
				vlog(VLOG_WARNING, "RMS database %d still "
					"open on exit - closing it", i);
				rms_close(i);
			}
		}

		vlog(VLOG_INFO, "RMS de-initialized");
	}
	else
	{
		vlog(VLOG_WARNING, "RMS is currently not initialized");
	}

	rms_state &= ~RMS_S_INIT;
}

/*
 * Open the RMS database specified by the <dbno> parameter and set
 * the record size to be used for it to <recsize> bytes. If the
 * <file> parameter is a non NULL pointer, the database is primed
 * with the contents of the specified file, and sucessive calls
 * of the rms_save() function upon this database will save it back
 * to the specified file. If the specified <flags> parameter contains
 * RMS_O_CREATE, the specified <file> will be created with <mode> if
 * it does not exist, otherwise it is considered an error if the specified
 * file does not already exist, and -1 is returned. Upon successful
 * completion, a value of 0 is returned. Otherwise a value of -1
 * is returned and the global variable errno is set to indicate
 * the error.
 *
 * rms_open() will succeed unless:
 *
 * EINVAL    - The supplied recordsize is invalid.
 * EBUSY     - The database specified by <dbno> is already in open state.
 * EFTYPE    - The size of the specified <file> does not correspond with the
 *             the specified <recsize>.
 *
 * Furthermore errno may contain any error codes set by the open(2), stat(2),
 * read(2) and malloc(3) system calls.
 */

int rms_open(uint32_t dbno, size_t recsize, const char *file, uint32_t flags, mode_t mode)
{
	int	ret = -1;
	RMSDB	*db = &rmsdb[dbno];

	if(!(db->flags & RMS_DB_OPEN))
	{
		if(recsize)
		{
			/* So far, so good */

			ret = 0;
		}
		else
		{
			vlog(VLOG_ERR, "Invalid record size '%d' "
				"specified for opeing RMS database "
				"'%d'", dbno);

			errno = EINVAL;
		}
	}
	else
	{
		vlog(VLOG_ERR, "RMS database '%d' is already open", dbno);

		errno = EBUSY;
	}

	if(ret)
		return ret;

	/*
	 * Set the database state to OPEN and set up basic
	 * database parameters to be used on this database.
	 */

	flags		&= RMS_OPEN_MASK;

	db->flags	= (RMS_DB_OPEN | flags);

	db->base	= 0;
	db->size	= 0;

	db->s_recs	= RMS_SEGSIZE;
	db->s_size	= recsize * RMS_SEGSIZE;
	db->s_total	= 0;
	db->s_retain	= RMS_RETAIN;

	if(db->s_retain <= db->s_recs)
		db->s_retain = db->s_recs + 1;

	if(!(db->s_retain % db->s_recs))
		db->s_retain++;

	db->r_size	= recsize;
	db->r_total	= 0;
	db->r_used	= 0;
	db->r_free	= 0;

	time(&db->reorg);

	/*
	 * If no file is specified, then there is nothing left to do.
	 */

	if((db->file = file))
	{
		int fd;
		int oflags = (O_BINARY | O_SEQUENTIAL);

		if(db->flags & RMS_O_CREATE)
			oflags |= (O_RDWR | O_CREAT);
		else
			oflags |= O_RDONLY;

		ret = -1; /* Assume failure until proven otherwise */

		if((fd = open(file, oflags, mode)) >= 0)
		{
			struct stat st;

			if(!fstat(fd, &st))
			{
				if(!(st.st_size % recsize))
				{
					/*
					 * Calculate number of segments needed.
					 */

					db->s_total = (st.st_size
						/ db->s_size);

					if(st.st_size % db->s_size)
						db->s_total++;

					/*
					 * Calculate total memory needed.
					 */

					db->size = db->s_total * db->s_size;

					/*
					 * Allocate memory if necessary.
					 */

					if(db->size)
					{
						if((db->base = malloc(db->size)))
						{
							if(read(fd, db->base, st.st_size) == st.st_size)
							{
								db->r_total = db->size
									/ recsize;

								db->r_used = st.st_size
									/ recsize;

								db->r_free = db->r_total
									- db->r_used;

								/* OK */

								errno = 0;
								ret = 0;
							}
							else
							{
								free(db->base);
								vlog(VLOG_ERR, "Read failed on RMS database file '%s' (%s)",
									file, strerror(errno));
							}
						}
						else
						{
							vlog(VLOG_CRIT, "RMS memory "
								"allocation failure "
								"(%s)",	strerror(errno));
						}
					}
					else
					{
						/* Bueno! */

						errno = 0;
						ret = 0;
					}
				}
				else
				{
					vlog(VLOG_ERR, "Filesize of RMS "
						"database file '%s' (%d) does "
						"not correspond to specified "
						"record size (%d)", file,
						st.st_size, recsize);

					errno = EFTYPE;
				}
			}
			else
			{
				vlog(VLOG_ERR, "Failed to stat RMS database "
					"file '%s' (%s)", file,
					strerror(errno));
			}

			close(fd);
		}
		else
		{
			vlog(VLOG_ERR, "Failed to open RMS database file '%s' "
				"(%s)", file, strerror(errno));
		}
	}

	/*
	 * Check status and reset database descriptor on error.
	 */

	if(ret)
		bzero(db, sizeof(RMSDB));

	return ret;
}

/*
 * Close the RMS database specified by the <dbno> parameter, free any memory
 * associated with the database and reset the database to it's initial closed
 * state. Upon successful operation a value of 0 is returned. Otherwise a
 * value of -1 is returned and the global variable errno is set to indicate
 * the error.
 *
 * The rms_close() function will succeed unless:
 *
 * EBADF     - The specified database is currently not in open state.
 */

int rms_close(uint32_t dbno)
{
	int ret = -1;
	RMSDB *db = &rmsdb[dbno];

	if(db->flags & RMS_DB_OPEN)
	{
		/*
		 * Free associated memory (if any).
		 */

		if(db->base)
			free(db->base);

		/*
		 * Clear database descriptor.
		 */

		bzero(db, sizeof(RMSDB));

		/* OK */

		errno = 0;
		ret = 0;
	}
	else
	{
		vlog(VLOG_WARNING, "Called rms_close() on database '%d' which "
			"is not open", dbno);

		errno = EBADF;
	}

	return ret;
}

/*
 * Add the new record pointed to by <buf> to the database specified by
 * the <dbno> parameter. Upon successful completion, a pointer to the
 * newly created database record is returned. In case of failure a value
 * NULL is returned and the global variable errno is set to indicate the error.
 */

void *rms_add_rec(uint32_t dbno, void *buf)
{
	void *ret = (void *)0;
	RMSDB *db = &rmsdb[dbno];
	void *p = 0;
	int new = 0;

	/*
	 * Find a place for the new record.
	 */

	if(db->base)
	{
		if(db->r_free)
		{
			/*
			 * Free records available -> Get one.
			 */

			p = (db->base +	(db->r_used * db->r_size));
		}
		else
		{
			/*
			 * No free records -> Add new segment.
			 */

			size_t newsize = (db->s_total + 1) * db->s_size;

			if((p = realloc(db->base, newsize)))
			{
				db->base = p;
				db->size = newsize;
				db->s_total++;
				db->r_total += db->s_recs;
				db->r_free = db->s_recs;

				p += (db->r_used * db->r_size);

				new = 1;
			}
			else
			{
				vlog(VLOG_CRIT, "RMS memory reallocation "
					"failure on database '%d' (%s)",
					strerror(errno));
			}
		}
	}
	else
	{
		/*
		 * DB has no mem at all -> Add first segment.
		 */

		if((p = malloc(db->s_size)))
		{
			db->base = p;
			db->size = db->s_size;
			db->s_total = 1;
			db->r_total = db->r_free = db->s_recs;
			db->r_used = 0;

			new = 1;
		}
		else
		{
			vlog(VLOG_CRIT, "RMS memory allocation failure on "
				"database '%d' (%s)", dbno, strerror(errno));
		}
	}

	/*
	 * Add the new record if we've found a place for it.
	 */

	if(p)
	{
		db->r_free--;
		db->r_used++;

		bcopy(buf, p, db->r_size);

		/* OK */

		if(new)
		{
			vlog(VLOG_DEBUG, "Added %d segment of %d records to "
				"RMS database '%d'", new, db->s_recs, dbno);
		}

		errno = 0;
		ret = p;
	}

	return ret;
}

/*
 * Delete the record that is indexed by the <recno> parameter
 * from the database specified by the <dbno> parameter. If the
 * operation is completed successfully, a value of 0 is returned,
 * otherwise a value of -1 will be returned and the global variable
 * errno is set to indicate the error.
 *
 * The rms_del_rec() function succeeds unless:
 *
 * ENOENT   - No record with an index of <recno> exists in the database.
 */

int rms_del_rec(uint32_t dbno, int recno)
{
	int ret = -1;	/* Assume failure until proven otherwise */
	RMSDB *db = &rmsdb[dbno];

	if(recno >= 0 && recno < db->r_used)
	{
		/*
		 * Shift remainder as necessary.
		 */

		if(recno < (db->r_used - 1))
		{
			void *p1 = (db->base + (recno * db->r_size));
			void *p2 = (p1 + db->r_size);
			size_t cnt = (db->r_used - (recno+1));

			bcopy(p2, p1, cnt * db->r_size);
		}

		db->r_used--;
		db->r_free++;

		/* OK */

		errno = 0;
		ret = 0;
	}
	else
	{
		vlog(VLOG_ERR, "No record %d in RMS database %d", recno, dbno);

		errno = ENOENT;
	}

	return ret;
}

/*
 * Save the contents of the RMS database specified by the <dbno> parameter
 * back to the associated file. Upon successful operation, the number of bytes
 * written to the file is returned, otherwise a value of -1 is returned and
 * the global variable errno is set to indicate the error.
 */

ssize_t rms_save(uint32_t dbno)
{
	ssize_t ret = -1;	/* Assume failure until proven otherwise */
	RMSDB *db = &rmsdb[dbno];

	/*
	 * Write database back to file if there is any.
	 */

	if(db->file)
	{
		int fd;
		int flags = (O_WRONLY | O_TRUNC | O_BINARY | O_SEQUENTIAL);

		if((fd = open(db->file, flags)) >= 0)
		{
			size_t nbytes = db->r_used * db->r_size;

			if((ret = write(fd, db->base, nbytes)) == nbytes)
			{
				/* OK */

				errno = 0;

				vlog(VLOG_DEBUG, "RMS database %d dumped to "
					"'%s' (%d bytes)", dbno, db->file,
					nbytes);
			}
			else
			{
				vlog(VLOG_ERR, "RMS database %d write failure "
					"on file '%s' %d/%d bytes (%s)",
					dbno, db->file, ret, nbytes,
					strerror(errno));

				ret = -1;
			}

			if(close(fd))
				ret = -1;
		}
		else
		{
			vlog(VLOG_ERR, "RMS database %d open failed on '%s' (%s)",
				db->file, strerror(errno));
		}
	}
	else
	{
		vlog(VLOG_WARNING, "RMS database '%d' has no file associated",
			dbno);

		/* OK */

		errno = 0;
		ret = 0;
	}

	return ret;
}

/*
 * Get the number of records that is currently contained within
 * the database specified by the <dbno> parameter.
 */

uint32_t rms_records(uint32_t dbno)
{
	return rmsdb[dbno].r_used;
}

/*
 * Get the memory address where the record indexed by <recno> is stored
 * within the database specified by the <dbno> parameter. Upon successful
 * operation the memory address of the record is returned. Otherwise a value
 * of NULL is returned and the global variable errno is set to indicate the
 * error.
 * 
 * The rms_addr() function succeeds unless:
 *
 * ENOENT   - No record with an index of <recno> exists in the database.
 */

void *rms_addr(uint32_t dbno, int recno)
{
	void *ret = 0;
	RMSDB *db = &rmsdb[dbno];

	if(recno >= 0 && recno < db->r_used)
	{
		/* OK */

		errno = 0;
		ret = (db->base + (recno * db->r_size));
	}
	else
	{
		vlog(VLOG_ERR, "No record '%d' in RMS database %d", recno);

		errno = ENOENT;
	}

	return ret;
}

/*
 * Get the record index of the record located withing database <dbno> at
 * address <rec>. If <rec> specifies a valid address within <dbno>, the
 * non-negative record index is returned. Otherwise a value of -1 is
 * returned and the global variable errno is set to indicate the error.
 *
 * The rms_index() function succeeds unless:
 *
 * EFAULT  - The specified record address <rec> is not valid.
 */

int rms_index(uint32_t dbno, void *rec)
{
	RMSDB *db = &rmsdb[dbno];

	if(db->base && rec >= db->base)
	{
		int recno = ((rec - db->base) / db->r_size);

		if(recno >= 0 && recno < db->r_used)
		{
			errno = 0;
			return recno;
		}
	}

	errno = EFAULT;

	return -1;
}

/*
 * Check the number of free records of all RMS databases and release unneeded
 * memory segments. Upon successful operation a value of 0 is returned.
 * In case of a memory reallocation error, -1 is returned and the global
 * variable errno contains the specific error code set by realloc(3).
 */

int rms_reorg(void)
{
	register uint32_t dbno;
	register RMSDB *db;
	int ret = 0;

	for(dbno = 0; dbno < RMS_MAX_DB; dbno++)
	{
		db = &rmsdb[dbno];

		if(db->flags & RMS_DB_OPEN)
		{
			/*
			 * Free memory segment(s) if needed.
			 */

			if(db->r_free >= db->s_retain)
			{
				void *p;

				uint32_t segs = ((db->r_used / db->s_recs)
					+ ((db->r_used % db->s_recs) ? 1 : 0));

				size_t newsize = (segs * db->s_size);

				if((p = realloc(db->base, newsize)))
				{
					uint32_t rel = db->s_total - segs;

					db->base = p;
					db->size = newsize;
					db->s_total = segs;
					db->r_total = segs * db->s_recs;
					db->r_free = db->r_total - db->r_used;

					time(&db->reorg);

					vlog(VLOG_DEBUG, "Released %d segments of %d records from database '%d'",
						rel, db->s_recs, dbno);
				}
				else
				{
					vlog(VLOG_CRIT, "RMS memory "
					     "reallocation failure on "
					     "database '%d' (%s)",
					     dbno, strerror(errno));

					ret = -1;
				}
			}
			else if(!db->r_used && db->base)
			{
				free(db->base);

				db->base = 0;
				db->size = 0;

				db->s_total = 0;
				db->r_total = 0;
				db->r_used = 0;
				db->r_free = 0;

				time(&db->reorg);

				vlog(VLOG_DEBUG, "Database '%d' released",
					dbno);
			}
		}
	}

	if(!ret)
		errno = 0;

	return ret;
}

/*
 * Discard the current contents of the RMS database specified by the
 * <dbno> parameter.
 */

void rms_discard(uint32_t dbno)
{
	RMSDB *db = &rmsdb[dbno];

	db->r_used = 0;
	db->r_free = db->r_total;
}

/*
 * Print information about the current status of all RMS databases
 * to the table supplied by the <tab> parameter.
 */

void rms_info(TABLE *tab)
{
	register uint32_t dbno;
	register RMSDB *db;

	for(dbno = 0; dbno < RMS_MAX_DB; dbno++)
	{
		db = &rmsdb[dbno];

		table_printf(tab, "* %sRMS db #%d:%s\n", S_BOLD, dbno, S_OFF);

		table_printf(tab, "*   %sstate%s=%s%s%s %sfile%s=%s%s%s\n",
			S_FG_CYN, S_OFF,
			S_FG_YEL, (db->flags & RMS_DB_OPEN) ?
			"open  " : "closed", S_OFF, S_FG_CYN, S_OFF, S_FG_YEL,
			db->file ? db->file : "<none>", S_OFF);

		table_printf(tab, "*   %s in %s%d%s segment%s of %s%d%s records of %s each\n",
			print_size(db->size), S_FG_GRN, db->s_total, S_OFF,
			db->s_total == 1 ? "" : "s",
			S_FG_GRN, db->s_recs, S_OFF,
			print_size(db->r_size));

		table_printf(tab, "*   %s%d%s record%s total, %s%d%s record%s used, %s%d%s record%s free\n",
			S_FG_GRN, db->r_total, S_OFF,
			db->r_total == 1 ? "" : "s",
			S_FG_GRN, db->r_used, S_OFF,
			db->r_used == 1 ? "" : "s",
			S_FG_GRN, db->r_free, S_OFF,
			db->r_free == 1 ? "" : "s");

		if(db->reorg)
		{
			table_printf(tab, "*   last reorganized %s%s%s ago\n",
				S_FG_RED, elapsed_now(db->reorg), S_OFF);
		}

		table_printf(tab, "*\n");
	}
}
