/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: serv_subr.c,v 1.25 2009/09/03 11:30:08 mickey Exp $ */

#include "vserver.h"

#if HAVE_SIGNAL_H
  #include <signal.h>
#endif

#if HAVE_TERMIOS_H
  #include <termios.h>
#else
  #if HAVE_SGTTY_H
    #include <sgtty.h>
  #else
    #if HAVE_SYS_IOCTL_H
      #include <sys/ioctl.h>
    #endif
  #endif
#endif
#ifdef WIN32
  #include <conio.h>
#endif

/*
 * Qualifier definitions.
 */

PQDEF qr_address =
{ QUALIFIER(QR_ADDRESS), "ADDRESS", PQARG_MANDATORY | PQARG_IPADDR, "inet_addr" };

PQDEF qr_hostname =
{ QUALIFIER(QR_HOSTNAME), "HOSTNAME", PQARG_MANDATORY | PQARG_STRING, "fqdn" };

PQDEF qr_lastlog =
{ QUALIFIER(QR_LASTLOG), "LASTLOG", PQARG_MANDATORY | PQARG_NUM, "num" };

PQDEF qo_logfile =
{ QUALIFIER(QO_LOGFILE), "LOGFILE", PQARG_OPTIONAL | PQARG_STRING, "file" };

PQDEF qr_maxpending =
{ QUALIFIER(QR_MAXPENDING), "MAXPENDING", PQARG_MANDATORY | PQARG_NUM, "num" };

PQDEF qr_port =
{ QUALIFIER(QR_PORT), "PORT", PQARG_MANDATORY | PQARG_NUM, "num" };

#if HAVE_DAEMON
PQDEF qn_detach =
{ QUALIFIER(QN_DETACH), "DETACH", PQARG_NONE, "" };
#endif
#if HAVE_SYSLOG
PQDEF qn_syslog =
{ QUALIFIER(QN_SYSLOG), "SYSLOG", PQARG_NONE, "" };
#endif

#ifdef WIN32 /* ------------- Windows only qualifiers ------------------ */

PQDEF qr_config =
{ QUALIFIER(QR_CONFIG), "CONFIG", PQARG_MANDATORY | PQARG_STRING, "cfgname" };

PQDEF qn_install =
{ QUALIFIER(QN_INSTALL), "INSTALL", PQARG_NONE, "" };

PQDEF qn_uninstall =
{ QUALIFIER(QN_UNINSTALL), "UNINSTALL", PQARG_NONE, "" };

#else /* ------------------- UNIX only qualifiers ---------------------- */

PQDEF qr_basedir =
{ QUALIFIER(QR_BASEDIR), "BASEDIR", PQARG_MANDATORY | PQARG_STRING, "path" };

PQDEF qn_createudb =
{ QUALIFIER(QN_CREATEUDB), "CREATEUDB", PQARG_NONE, "" };

PQDEF qr_rotatelog =
{ QUALIFIER(QR_ROTATELOG), "ROTATELOG", PQARG_MANDATORY | PQARG_NUM, "num" };

#endif

/*
 * Process all arguments specified on the commandline
 * and setup program defaults accordingly.
 */

int pqf_vserver(int argc, char *argv[], PQLINK *pql)
{
	uint32_t pqid;
	char *what = "";

	pqindex = 0;

	while((pqid = getnextpq(argc, argv, pql, 0)) != PQ_EXHAUSTED)
	{
		what = ISQUALIFIER(pqid) ? "Qualifier '/" : "Parameter '";

		if(PQ_INVALID(pqid))
		{
			usage("%s: %s%s' is not valid.", pqpath, what, pqname);
			return -1;
		}
		else if(PQ_AMBIGIOUS(pqid))
		{
			usage("%s: %s%s' is ambigious.", pqpath, what, pqname);
			return -1;
		}
		else if(PQ_REQARG(pqid))
		{
			usage("%s: %s%s' requires an argument of type %s.",
				pqpath, what, pqname,
				PQARG_FMT_VALTYPE(pqargtype));
			return -1;
		}
		else if(PQ_NOARG(pqid))
		{
			usage("%s: %s%s' expects no argument.", pqpath, what,
				pqname);
			return -1;
		}
		else if(PQ_NOPARSE(pqid))
		{
			usage("%s: %s%s' expects an argument of type %s.",
				pqpath, what, PQARG_FMT_VALTYPE(pqargtype));
			return -1;
		}
		else if(PQ_DATEERR(pqid))
		{
			printf("%s: %s%s' expects a date in the %s.", pqpath,
				what, pqname,
				PQARG_VALTYPE(pqargtype) == PQARG_PDATE ?
				"past" : "future");
			return -1;
		}
		else if(pqid == QUALIFIER(QR_ADDRESS))
		{
			bcopy(&pqaddress, &cfdata.bind_addr, sizeof(cfdata.bind_addr));
		}
#if HAVE_DAEMON
		else if(pqid == QUALIFIER(QN_DETACH))
		{
			cfdata.flags |= CF_DETACH;
		}
#endif
		else if(pqid == QUALIFIER(QR_HOSTNAME))
		{
			sncopy(pqarg, cfdata.hostname, sizeof(cfdata.hostname));
		}
		else if(pqid == QUALIFIER(QR_LASTLOG))
		{
			if(pqvalue != 0 && pqvalue < LASTLOG_MIN)
				pqvalue = LASTLOG_MIN;
			else if(pqvalue > LASTLOG_MAX)
				pqvalue = LASTLOG_MAX;

			cfdata.lastlog_recs = pqvalue;
		}
		else if(pqid == QUALIFIER(QO_LOGFILE))
		{
			cfdata.flags |= CF_LOG_FILE;

			if(pqarg && *pqarg)
			{
				sncopy(pqarg, cfdata.logfile, sizeof(cfdata.logfile));
			}
		}
		else if(pqid == QUALIFIER(QR_MAXPENDING))
		{
			if(pqvalue < 1 || pqvalue > 64)
			{
				usage("Specified maximum number of pending "
					"connections out of range "
					"(1 <= n <= 64)");
				return -1;
			}

			cfdata.max_pending = pqvalue;
		}
		else if(pqid == QUALIFIER(QR_PORT))
		{
			if(pqvalue < 1 || pqvalue > 65535)
			{
				usage("Specified TCP port number out of range "
					"(1 <= p <= 65535)");

				return -1;
			}

			cfdata.port = pqvalue;
		}
#if HAVE_SYSLOG
		else if(pqid == QUALIFIER(QN_SYSLOG))
		{
			cfdata.flags |= CF_LOG_SYSLOG;
		}
#endif
#ifdef WIN32
		else if(pqid == QUALIFIER(QR_CONFIG))
		{
			snprintf(cfdata.svc_config, sizeof(cfdata.svc_name) -
				(cfdata.svc_config - &cfdata.svc_name[0]),
				":%s", pqarg);

			++cfdata.svc_config;
		}
		else if(pqid == QUALIFIER(QN_INSTALL))
		{
			cfdata.flags |= CF_INSTALL;
		}
		else if(pqid == QUALIFIER(QN_UNINSTALL))
		{
			cfdata.flags |= CF_UNINSTALL;
		}
#else
		else if(pqid == QUALIFIER(QR_BASEDIR))
		{
			size_t len = sncopy(pqarg, cfdata.basedir, sizeof(cfdata.basedir));

			if(len && cfdata.basedir[len-1] != DIRSEP)
			{
				snprintf(&cfdata.basedir[len],
					sizeof(cfdata.basedir) - len,
					"%c", DIRSEP);
			}
		}
		else if(pqid == QUALIFIER(QN_CREATEUDB))
		{
			cfdata.flags |= CF_CREATE_UDB;
		}
		else if(pqid == QUALIFIER(QR_ROTATELOG))
		{
			cfdata.rotate_log = pqvalue;
		}
#endif
	}

	argv += pqindex;
	argc -= pqindex;

	return 0;
}

/*
 * Set up global configuration defaults.
 */

void mkdefaults(void)
{
	bzero(&cfdata, sizeof(cfdata));

	sncopy(DEFAULT_BASEDIR, cfdata.basedir, sizeof(cfdata.basedir));

	snprintf(cfdata.logfile, sizeof(cfdata.logfile), "vserver.log");

#ifdef WIN32
	cfdata.svc_config = cfdata.svc_name +
		sncopy(SVC_NAME, cfdata.svc_name, sizeof(cfdata.svc_name));
#endif
	cfdata.bind_addr.s_addr	= INADDR_ANY;
	cfdata.port		= DEFAULT_SERVER_PORT;
	cfdata.max_pending	= DEFAULT_MAX_PENDING;
	cfdata.rotate_log	= DEFAULT_NUMLOGFILES;
	cfdata.lastlog_recs	= DEFAULT_LASTLOG_ENTRIES;

	default_options();
}

/*
 * Default server options.
 */

void default_options(void)
{
	bzero(&vsopt, sizeof(vsopt));

	vsopt.t_guard.tv_sec	= DEFAULT_FLOODGUARD_SEC;
	vsopt.t_guard.tv_usec	= DEFAULT_FLOODGUARD_USEC;

	vsopt.dnschklvl		= DNS_CHK_NONE;
}

/*
 * Load server options.
 */

void load_options(void)
{
	int fd;

	if((fd = open(fn_options, (O_RDONLY | O_BINARY | O_SEQUENTIAL))) >= 0)
	{
		if(read(fd, &vsopt, sizeof(vsopt)) != sizeof(vsopt))
		{
			vlog(VLOG_ERR, "Failed to load options from '%s' (%s)",
				fn_options, strerror(errno));
			default_options();
		}

		close(fd);
	}
	else if(errno == ENOENT)
	{
		default_options();
		save_options();
	}
	else
	{
		vlog(VLOG_ERR, "Failed to load options from '%s' (%s)",
			fn_options, strerror(errno));
		default_options();
	}
}

/*
 * Save server options.
 */

void save_options(void)
{
	int fd;

	if((fd = open(fn_options, (O_WRONLY | O_TRUNC | O_CREAT | O_BINARY
		| O_SEQUENTIAL), 0600)) >= 0)
	{
		if(write(fd, &vsopt, sizeof(vsopt)) != sizeof(vsopt))
		{
			vlog(VLOG_ERR, "Failed to save options to '%s' (%s)",
				fn_options, strerror(errno));
		}

		close(fd);
	}
	else
	{
		vlog(VLOG_ERR, "Failed to save options to '%s' (%s)",
			fn_options, strerror(errno));
	}
}

/*
 * Perform global initialization procedure and return status.
 */

int global_init(void)
{
	/*
	 * Clear all counters and store server start time.
	 */

	bzero(&vs, sizeof(VS_STAT));
	time(&vs.vs_start);

	/*
	 * Check base directory accessability and chdir into it.
	 */

	if(access(cfdata.basedir, (R_OK | W_OK | X_OK)) || chdir(cfdata.basedir))
	{
		vlog(VLOG_CRIT, "Base directory '%s' is not accessible (%s)",
			cfdata.basedir, strerror(errno));

		return -1;
	}

	/*
	 * Create PID file.
	 */

	pidfile(1);

	/*
	 * Initialize vserver logging.
	 */

	vlog_open();

	/*
	 * Load server options from file.
	 */

	load_options();

	/*
	 * Install signal handlers.
	 */
#if HAVE_SIGNALS
	if(!install_sighandlers())
#endif
	{
		/*
		 * Initialize databases.
		 */

		if(!db_init())
		{
			/*
			 * Initialize control socket.
			 */

#if HAVE_UNIX_SOCKETS
			if(!vchatctl_open())
#endif
			{
				/*
				 * Initialize server socket.
				 */

				if(!sock_init())
				{
					if(vsopt.flags & VO_FLOODGUARD)
					{
						vlog(VLOG_NOTICE, "FloodGuard enabled, treshold=%s",
							mt_print(&vsopt.t_guard));
					}

					vlog(VLOG_INFO, "Listening for connections on interface %s:%d",
						inet_ntoa(cfdata.bind_addr), (uint16_t)cfdata.port);

					/*
					 * Global initialization succeeded.
					 */

					vlog(VLOG_INFO, "VChat conference "
						"server %s protocol %x-%x "
						"pid %d -- ready",
						PACKAGE_VERSION,
						MAJOR_REV(PROTO_REV),
						MINOR_REV(PROTO_REV),
						getpid());

					return 0;
				}
#if HAVE_UNIX_SOCKETS
				vchatctl_close();
#endif				
			}

			db_exit();
		}
#if HAVE_SIGNALS
		uninstall_sighandlers();
#endif
	}

	/*
	 * Close logging.
	 */

	vlog_close();

	/*
	 * Remove PID file.
	 */

	pidfile(0);

	/*
	 * Return error status.
	 */

	return -1;
}

/*
 * Main server cleanup function
 */

void global_exit(void)
{
	/*
	 * Close the server socket.
	 */

	sock_exit();

	/*
	 * Dispose of server control socket.
	 */
#if HAVE_UNIX_SOCKETS
	vchatctl_close();
#endif
	/*
	 * De-initialize databases.
	 */

	db_exit();

	/*
	 * Uninstall signal handlers.
	 */
#if HAVE_SIGNALS
	uninstall_sighandlers();
#endif
	/*
	 * Log server uptime.
	 */

	vlog(VLOG_INFO, "Server has been up for %s", elapsed_now(vs.vs_start));

	/*
	 * Close logging.
	 */

	vlog_close();

	/*
	 * Unlink PID file (if any).
	 */

	pidfile(0);
}

/*
 * Initialize RMS and all databases then return status.
 */

int db_init(void)
{
	rms_init();

	if(!userdb_open())
	{
		if(!acldb_open())
		{
			if(!topicdb_open())
			{
				if(!cldb_open())
				{
					rms_status();

					lastlog_open();
#ifdef WIN32
					ssStatus.dwCheckPoint++;
#endif
					lastlog_resize();
#ifdef WIN32
					ssStatus.dwCheckPoint++;
#endif
					return 0;
				}

				topicdb_close();
			}

			acldb_close();
		}

		userdb_close();
	}

	rms_exit();

	return -1;
}

/*
 * De-initialize all databases and RMS.
 */

void db_exit(void)
{
	lastlog_close();
	cldb_close();
	topicdb_close();
	acldb_close();
	userdb_close();

	rms_exit();
}

/*
 * Initialize the server socket and return status.
 */

int sock_init(void)
{
#ifdef WIN32
	WSADATA			wsaData;
	WORD			version_requested;
	const char		on		= 1;
#else
	int			on		= 1;
#endif
#ifdef WIN32
	/*
	 * Initialize WINSOCK.
	 */

	version_requested = MAKEWORD(2, 0);

	while(WSAStartup(version_requested, &wsaData) != 0)
	{
		vlog(VLOG_CRIT, "Unable to find a suitable winsock version");
		return -1;
	}

	cfdata.flags |= CF_WSA_INIT;
#endif
	/*
	 * Open the server socket.
	 */
#ifdef WIN32
	if((sock = socket(AF_INET, SOCK_STREAM, 0)) != INVALID_SOCKET)
#else
	if((sock = socket(AF_INET, SOCK_STREAM, 0)) != -1)
#endif
	{
		/*
		 * Set socket options.
		 */

		if(!setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)))
		{
			/*
			 * Bind socket
			 */

			bzero(&sock_in, sizeof(sock_in));

			sock_in.sin_family = AF_INET;
			sock_in.sin_port = htons((uint16_t)cfdata.port);
			sock_in.sin_addr = cfdata.bind_addr;

			if(!bind(sock, (struct sockaddr *)&sock_in,
				sizeof(sock_in)))
			{
				/*
				 * Listen for incoming connections.
				 */

				if(!listen(sock, cfdata.max_pending))
				{
					/* OK */

					return 0;
				}
				else
				{
					vlog(VLOG_CRIT, "Failed to listen(2) "
						"on server socket, status=%d "
						"(%s)",	errno,
						strerror(errno));
				}
			}
			else
			{
				vlog(VLOG_CRIT, "Failed to bind server socket,"
					" status=%d (%s)", errno,
					strerror(errno));
			}
		}
		else
		{
			vlog(VLOG_CRIT, "Failed to set server socket options, "
				"status=%d (%s)", errno, strerror(errno));
		}
#ifdef WIN32
		closesocket(sock);
#else
		close(sock);
#endif
	}
	else
	{
#ifdef WIN32
		int errcod = WSAGetLastError();
#endif
		vlog(VLOG_CRIT, "Failed to open server socket, status=%d (%s)",
#ifdef WIN32
			errcod, strerror(errcod));
#else
			errno, strerror(errno));
#endif
	}

#ifdef WIN32
	WSACleanup();
	cfdata.flags &= ~CF_WSA_INIT;
#endif
	return -1;
}

/*
 * Close the server socket and perform necessary cleanup tasks.
 */

void sock_exit(void)
{
#ifdef WIN32
	if(sock != -1)
	{
		closesocket(sock);
		sock = -1;
	}

	if(cfdata.flags & CF_WSA_INIT)
	{
		WSACleanup();
		cfdata.flags &= ~CF_WSA_INIT;
	}

	ssStatus.dwCheckPoint++;
#else
	close(sock);
#endif
}

/*
 * Create or remove the PID file. If <on> is non-null,
 * the PID file is created, otherwise it is removed.
 */

void pidfile(int on)
{
	if(on)
	{
		FILE *pidf;

		if((pidf = fopen(fn_pid, "w")) != NULL)
		{
#ifdef WIN32
			fprintf(pidf, "%d\n", _getpid());
#else
			fprintf(pidf, "%d\n", getpid());
#endif
			fclose(pidf);
		}
	}
	else
	{
		unlink(fn_pid);
	}
}

/*
 * Return the number of users on the conference channel <chan>.
 */

uint32_t chatters(int32_t chan)
{
	uint32_t records = cldb_records();
	register uint32_t n = 0;

	if(records)
	{
		register CLDATA *cp = cldb_addr(0);

		if(cp)
		{
			register int i;

			for(i = 0; i < records; i++, cp++)
			{
				if(cp->channel == chan)
				{
					n++;
				}
			}
		}
	}

	return n;
}

/*
 * Check if the given nickname contains any invalid characters and if
 * it's length is valid. Return 0 if the specified nickname is invalid,
 * or 1 if the nickname is ok to use.
 */

int nick_ok(register char *nick)
{
	register size_t len;

	/*
	 * Check for empty nickname.
	 */

	if(!nick || !*nick)
		return 0;

	/*
	 * Check nickname character validity
	 */

	for(len = 0; *nick; nick++, len++)
	{
		if(!strchr(NICKVALIDCHARS, *nick) || len > VPNICKSIZE)
		{
			return 0;
		}
	}

	return 1;
}

/*
 * Create the server's VIDENT packet with the type field set to <type>.
 */

VIDENT *ident(uint32_t type)
{
	static VIDENT vid;

	vid.magic = htonl(VSERVER_MAGIC);
	vid.proto = htonl(PROTO_REV);
	vid.type  = htonl(type);
	snprintf(vid.ident, sizeof(vid.ident), "VCS-%s-%s", OS_TYPE,
		PACKAGE_VERSION);

	return &vid;
}

/*
 * Split the list of nicknames supplied in <list> into the argument
 * vector <argv> not exceeding a total of <nmax> elements. Invalid
 * or duplicated nicknames are ignored and corresponding error
 * messages will be written to the supplied TABLE <tab>. If the <wempty>
 * parameter is non-zero and the resulting nickname list is empty, a warning
 * message will be written to the TABLE <tab>. The return value is the number
 * of nicknames parsed into the argument vector.
 */

int nicksplit(TABLE *tab, char *list, char *argv[], size_t nmax, int wempty)
{
	char **ap;
	int argc = 0;
	int invalid;

#if HAVE_STRSEP
	for(ap = argv; (*ap = strsep(&list, " \t")) != NULL && argc < nmax; )
#else
	ap = argv;

	*ap = strtok(list, " \t");

	while(*ap != NULL && argc < nmax)
#endif
	{
		invalid = 0;

		if(!**ap)
		{
			continue;
		}
		else if(!nick_ok(*ap))
		{
			invalid = 1;

			table_xprintf(tab, "\n* Ignoring invalid nickname '%s%s%s'",
				S_FG_RED, *ap, S_OFF);
		}
		else
		{
			register int i;

			for(i = 0; i < argc; i++)
			{
				if(!strcmp(*ap, argv[i]))
				{
					invalid = 1;

					table_xprintf(tab, "\n* Ignoring duplicated nickname '%s%s%s'",
						S_FG_RED, *ap, S_OFF);
				}
			}
		}

		if(**ap && !invalid)
		{
			++ap;
			++argc;
		}

#if !HAVE_STRSEP
		*ap = strtok(NULL, " \t");
#endif
	}

	if(!argc && wempty)
	{
		table_xprintf(tab, "\n* Final nickname list is empty");
		table_flush(tab);
	}

	return argc;
}

/*
 * Print a convenient representation of the given size.
 */

char *print_size(size_t bytes)
{
	static char buf[STCB_NUM][STCB_SIZE];
	static int index = 0;

	if(++index >= STCB_NUM)
		index = 0;

	if(bytes >= 0x40000000)
	{
		snprintf(&buf[index][0], STCB_SIZE, "%s%.2f%s Gigabyte", S_FG_GRN,
			(double)bytes / 0x40000000, S_OFF);
	}
	else if(bytes >= 0x100000)
	{
		snprintf(&buf[index][0], STCB_SIZE, "%s%.2f%s Megabyte", S_FG_GRN,
			(double)bytes / 0x100000, S_OFF);
	}
	else if(bytes >= 0x400)
	{
		snprintf(&buf[index][0], STCB_SIZE, "%s%.2f%s kilobyte", S_FG_GRN,
			(double)bytes / 0x400, S_OFF);
	}
	else
	{
		snprintf(&buf[index][0], STCB_SIZE, "%s%d%s bytes", S_FG_GRN,
			bytes, S_OFF);
	}

	return &buf[index][0];
}

/*
 * Return a textual representation of the specified permission bit mask.
 */

char *perm_str(uint32_t perm)
{
	static char buf[PATH_MAX];

	if(perm != AUTH_ROOT)
	{
		register int i,j,r,s;
		register char *bp;

		for(i = j = 0, r = PATH_MAX, bp = buf; i < 32; i++)
		{
			if((perm & (1 << i)) != 0)
			{
				s = snprintf(bp, r, "%s%s", !j++ ? "" : "+",
					permnames[i]);

				r -= s;
				bp += s;
			}
		}
	}
	else
	{
		snprintf(buf, sizeof(buf), "%s", "root");
	}

	return buf;
}

#if HAVE_SIGNALS
/*
 * Install a handler for various signals.
 */

int install_sighandlers(void)
{
	register int i;
	int sigs[] = {SIGINT, SIGQUIT, SIGTERM, SIGHUP,	SIGUSR1, SIGPIPE, SIGALRM};
	int numsigs = sizeof(sigs) / sizeof(int);
  #if HAVE_SIGACTION
	struct sigaction sa;

	sa.sa_handler = signal_handler;
	sigemptyset(&sa.sa_mask);
  #else	/* SIGVEC */
	struct sigvec sv;

	sv.sv_handler = signal_handler;
	sv.sv_mask = 0;
  #endif
	for(i = 0; i < numsigs; i++)
	{
		if(sigs[i] == SIGALRM)
		{
  #if HAVE_SIGACTION
    #ifdef SA_INTERRUPT
			sa.sa_flags = SA_INTERRUPT;
    #else
			sa.sa_flags = 0;
    #endif
  #else
			sv.sv_flags = SV_INTERRUPT;
  #endif
		}
		else
		{
  #if HAVE_SIGACTION
    #ifdef SA_RESTART
			sa.sa_flags = SA_RESTART;
    #else
			sa.sa_flags = 0;
    #endif
  #else
			sv.sv_flags = 0;
  #endif
		}

  #if HAVE_SIGACTION
		if(sigaction(sigs[i], &sa, NULL))
  #else
		if(sigvec(sigs[i], &sv, NULL))
  #endif
		{
			vlog(VLOG_ERR, "Failed to set up signal handler for "
				"signal %s (%s)", signame(sigs[i]),
				strerror(errno));

			return -1;
		}
	}

	return 0;
}

/*
 * Uninstall the signal handlers, and set all signals to SIG_IGN,
 * so that no further signals will be delivered to the signal handler.
 */

void uninstall_sighandlers(void)
{
	register int i;
	int sigs[] = {SIGINT, SIGQUIT, SIGTERM, SIGHUP, SIGUSR1, SIGPIPE, SIGALRM};
	int numsigs = sizeof(sigs) / sizeof(int);
  #if HAVE_SIGACTION
	struct sigaction sa;

	sa.sa_handler = SIG_IGN;
	sa.sa_flags = 0;
	sigemptyset(&sa.sa_mask);
  #else
	struct sigvec sv;
	sv.sv_handler = SIG_IGN;
	sv.sv_flags = 0;
	sv.sv_mask = 0;
  #endif

	for(i = 0; i < numsigs; i++)
	{
  #if HAVE_SIGACTION
		sigaction(sigs[i], &sa, NULL);
  #else
		sigvec(sigs[i], &sv, NULL);
  #endif
	}
}

/*
 * Return a name for a given signal.
 */

char *signame(int sig)
{
	static char	buf[128];

  #if HAVE_DECL_SYS_SIGNAME
	snprintf(buf, sizeof(buf), "SIG%s", sys_signame[sig]);
	return upperstr(buf);
  #elif HAVE_DECL_SYS_SIGLIST
	snprintf(buf, sizeof(buf), "%s", sys_siglist[sig]);
	return upperstr(buf);
  #elif HAVE_STRSIGNAL
	snprintf(buf, sizeof(buf), "%s", strsignal(sig));
	return upperstr(buf);
  #else
	snprintf(buf, sizeof(buf), "%d", sig);
	return buf;
  #endif
}

/*
 * Signal handler function.
 */

void signal_handler(int sig)
{
	vlog(VLOG_NOTICE, "Signal %s received", signame(sig));

	switch(sig)
	{
		case SIGALRM:	if(vio_tovalid)
					longjmp(vio_jmpbuf, 1);
				else
					vlog(VLOG_ERR, "No timeout active");
				break;

		case SIGINT:
		case SIGQUIT:
		case SIGTERM:	vlog(VLOG_NOTICE, "Requesting server shutdown");
				cfdata.flags |= CF_SHUTDOWN;
				break;

		case SIGHUP:	/* No files to reload */
				break;

		case SIGUSR1:	vlog(VLOG_NOTICE, "Rotating logfile");
				rotate_logfile();
				break;

		case SIGPIPE:	vlog(VLOG_WARNING, "Signal ignored");
				break;

		default:	vlog(VLOG_WARNING, "No action assigned for signal");
				break;
	}
}
#endif /* HAVE_SIGNALS */

/*****************************************************************************
 * PQLINK setup:                                                             *
 *****************************************************************************/

PQLINK pql_vserver[] =
{
	{
		&qr_address,		0,
		0,			0,
		"Service internet address."
	},
#ifndef WIN32
	{
		&qr_basedir,		0,
		0,			0,
		"Base directory where files are located."
	},
#endif
#ifdef WIN32
	{
		&qr_config,		0,
		0,			0,
		"Use the named configuration instead of the default."
	},
#endif
#ifndef WIN32
	{
		&qn_createudb,		0,
		0,			0,
		"Create an initial user database file."
	},
#endif
#ifdef HAVE_DAEMON
	{
		&qn_detach,		0,
		0,			0,
		"Detach from terminal and run as a daemon process."
	},
#endif
	{
		&qr_hostname, 		0,
		0,			0,
		"Hostname as announced to clients."
	},
#ifdef WIN32
	{
		&qn_install,		0,
		0,			0,
		"Install service to Service Control Manager database."
	},
#endif
	{
		&qr_lastlog,		0,
		0,			0,
		"Number of lastlog records to store."
	},
	{
		&qo_logfile, 		0,
		0,			0,
		"Enable logging to file."
	},
	{
		&qr_maxpending,		0,
		0,			0,
		"Maximum number of pending setup requests on socket."
	},
	{
		&qr_port,		0,
		0,			0,
		"Service TCP port number."
	}
#ifndef WIN32
	,
	{
		&qr_rotatelog,		0,
		0,			0,
		"Number of versions to retain when rotating the logfile."
	}
#endif
#if HAVE_SYSLOG
	,
	{
		&qn_syslog,		0,
		0,			0,
		"Enable logging via LOG_LOCAL1 of syslog(3)."
	}
#endif
#ifdef WIN32
	,
	{
		&qn_uninstall,		0,
		0,			0,
		"Uninstall service from Service Control Manager database."
	}
#endif
};

PQLINK pql_master =
{
	0,			pqf_vserver,
	pql_vserver,		NPQL(pql_vserver),
	"VChat conference server."
};

/*****************************************************************************
 * Commandline processor                                                     *
 *****************************************************************************/

int process_cmdline(int argc, char *argv[])
{
	pqpathlen = sncopy(argv[0], pqpath, sizeof(pqpath));
	pqpathdepth = 1;

	argv++;
	argc--;

	return pqf_vserver(argc, argv, &pql_master);
}

/*****************************************************************************
 * PROGRAM USAGE:                                                            *
 *****************************************************************************/

void usage(char *fmt, ...)
{
	va_list ap;

	register int i;
	register PQLINK *lnk;
	char *s;
	size_t pl;
	size_t ll;
	size_t ql;
	size_t ml;

	va_start(ap, fmt);

	fprintf(stderr, "\n");
	fprintf(stderr, " ==============================================================================\n");
	fprintf(stderr, "                      VChat - Interactive conference system\n");
	fprintf(stderr, "               (vserver %s - %s) / Protocol revision: %x-%x\n", PACKAGE_VERSION, PACKAGE_RELDATE, MAJOR_REV(PROTO_REV), MINOR_REV(PROTO_REV));
	fprintf(stderr, "                       (c) 1993-2009 by Andreas S. Wetzel\n");
	fprintf(stderr, "                           *** All rights reserved ***\n");
	fprintf(stderr, " ==============================================================================\n\n");
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n\n");
	fprintf(stderr, "USAGE:\n\n");

	fprintf(stderr, "  %s", prog_name);

	for(i = ml = 0, pl = ll = (strlen(prog_name)+2), lnk = pql_master.link;
	i < pql_master.nlinks; i++, lnk++)
	{
		s = pqfmt(lnk->pqd, &ql);

		if(ql > ml)
			ml = ql;

		ql += 3;

		if(ll + ql > 79)
		{
			fprintf(stderr, "\n%*s", pl, " ");
			ll = pl;
		}

		ll += fprintf(stderr, " [%s]", s);
	}
	fprintf(stderr, "\n\n");

	fprintf(stderr, "OPTIONS:\n\n");

	ml += 2;

	for(i = 0, lnk = pql_master.link; i < pql_master.nlinks; i++, lnk++)
	{
		s = pqfmt(lnk->pqd, &ql);

		fprintf(stderr, "  %s%*s%s\n", s, ml - ql, " ", lnk->desc);
	}
#ifdef WIN32
	fprintf(stderr, "\nEXAMPLES:\n\n");
	fprintf(stderr, "  %s /CONFIG=example.com /ADDRESS=10.1.2.3 /PORT=1169 /INSTALL\n\n", prog_name);
	fprintf(stderr, "  %s /CONFIG=example.com /UNINSTALL\n", prog_name);
#endif
	fprintf(stderr, "\n");

	va_end(ap);
}
