/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: table.c,v 1.10 2009/08/14 11:54:52 mickey Exp $ */

#include "vserver.h"

/*
 * Initialize the table pointed to by <table> for outputting to
 * the client specified by <cl>. Any previous contents of the
 * table are discarded.
 */

void table_init(TABLE *table, CLDATA *cl)
{
	table->client		= cl;
	table->used		= 0;
	table->buf[0]		= '\0';	
}

/*
 * Print the string <fmt> and it's arguments to the table pointed to
 * by <table>.
 */

void table_printf(TABLE *table, char *fmt, ...)
{
	va_list	ap;

	va_start(ap, fmt);

	table->used += vsnprintf(&table->buf[table->used], sizeof(table->buf) -
			table->used, fmt, ap);

	/*
	 * Flush the table if there is less than 256 bytes
	 * of buffer space remaining in it.
	 */

	if((sizeof(table->buf) - table->used) < 256)
	{
		table_flush(table);
		table_init(table, table->client);
	}

	va_end(ap);
}

/*
 * Print the string <fmt> and it's arguments to the table pointed to
 * by <table>. If the <fmt> string starts with a newline, and the table
 * is currently empty, then ignore the leading newline in the format string.
 */

void table_xprintf(TABLE *table, char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);

	if(!table->used && *fmt == '\n')
		++fmt;

	table->used += vsnprintf(&table->buf[table->used], sizeof(table->buf) -
			table->used, fmt, ap);

	/*
	 * Flush the table if there is less than 256 bytes
	 * of buffer space remaining in it.
	 */

	if((sizeof(table->buf) - table->used) < 256)
	{
		table_flush(table);
		table_init(table, table->client);
	}

	va_end(ap);
}

/*
 * Flush the table pointed to by <table>, thus writing and pending output
 * to the client.
 */

void table_flush(TABLE *table)
{
	size_t	len;

	/*
	 * If the error flag is set upon the client, then
	 * we would do no good trying to output data to it.
	 */

	if(table->client->flags & CFLAG_ERROR)
	{
		table_init(table, table->client);
		return;
	}

	/*
	 * If the table is empty, do nothing.
	 */

	if(!table->used)
		return;

	/*
	 * Send the table data to the client.
	 */

	table->svcmd.cmd = htonl(SERV_MSG);
	table->svcmd.len = htonl(table->used);

	len = table->used + sizeof(table->svcmd);

	if(vio_write(table->client->fd, &table->svcmd, len, CONN_TIMEOUT) != len)
	{
		vlog(VLOG_ERR, "Write failed on client fd %d (%s)",
			table->client->fd, strerror(errno));

		table->client->flags |= CFLAG_ERROR;
		table->client->error = errno;
	}

	/*
	 * Reset the table.
	 */

	table_init(table, table->client);
}
