/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: userdb.c,v 1.16 2009/08/14 11:54:52 mickey Exp $ */

#include "vserver.h"

#ifdef WIN32
  #include <io.h>
#endif

/*
 * Load the user database file into memory
 * discarding the previous in-core copy (if any).
 */

int userdb_open(void)
{
	/*
	 * Open the user database.
	 */

	if(rms_open(RMS_USER, sizeof(USERDATA), fn_userdb, 0, 0))
	{
		vlog(VLOG_CRIT, "Failed to open user database '%s' (%s)",
			fn_userdb, strerror(errno));

		if(errno == ENOENT)
		{
			vlog(VLOG_CRIT, "If the above pathname is correct, "
				"then it is likely that you have not yet "
				"created the user database. You may create "
				"it by using '%s /CREATEUDB'. See vserver(8) "
				"for details.", prog_name);
		}

		return -1;
	}

	vlog(VLOG_INFO, "User database loaded (%d records)",
		rms_records(RMS_USER));

	return 0;
}

/*
 * Free any resources used by user database.
 */

void userdb_close(void)
{
	rms_close(RMS_USER);
}

/*
 * Save the in-core copy of the user database back to file.
 * If successful, the number of bytes written to the file
 * is returned. In case of failure a value of -1 is returned.
 */

ssize_t userdb_save(void)
{
	return rms_save(RMS_USER);
}

/*
 * Add user <nick> with <password> to the in-core copy of the user database.
 * Upon successful operation a value of 0 is returned. Otherwise a value of
 * -1 is returned and the global variable errno is set to indicate the error.
 *
 * The userdb_add() function succeeds unless:
 *
 * EEXIST    - A user with the specified <nick> already exists.
 * EBUSY     - The specified <nick> is currently being used.
 *
 * Furthermore errno may contain any specific error code set by the
 * rms_add_rec() function provided that it fails.
 */

int userdb_add(char *nick, char *passwd)
{
	int ret = -1;

	if(!cldb_chkpresent(nick))
	{
		int idx = userdb_find(nick);

		if(idx < 0 && errno == ENOENT)
		{
			USERDATA udat;

			bzero(&udat, sizeof(udat));

			time(&udat.account_created);
			udat.last_signon = udat.last_signoff = 0;
			udat.accperm = udat.modperm = udat.sessioncnt = 0;
			sncopy(nick, udat.nick, sizeof(udat.nick));
			sncopy(passwd, udat.passwd, sizeof(udat.passwd));
			udat.cl_opts = DEF_CLOPTMASK;

			ret = (rms_add_rec(RMS_USER, &udat) ? 0 : -1);
		}
		else if(idx >= 0)
		{
			/*
			 * A user with that name already exists.
			 */

			errno = EEXIST;
		}
	}
	else
	{
		/*
		 * The specified nickname is currently in use.
		 */

		errno = EBUSY;
	}

	return ret;
}

/*
 * Delete the user specified by <nick> from the in-core copy of the user
 * database. Upon successful operation a value of 0 is returned. Otherwise
 * a value of -1 is returned and the global variable errno is set to indicate
 * the error.
 *
 * The userdb_delete() function succeeds unless:
 *
 * ENOENT    - The specified <nick> was not found in the user database.
 * EBUSY     - The specified <nick> is currently being used.
 * EPERM     - The VChat super user may not be deleted.
 *
 * Furthermore errno may contain any specific error code set by the
 * rms_del_rec() function provided that it fails.
 */

int userdb_delete(char *nick)
{
	int ret = -1;

	if(!cldb_chkpresent(nick))
	{
		int idx = userdb_find(nick);

		if(idx >= 0)
		{
			USERDATA *udat = (USERDATA *)rms_addr(RMS_USER, idx);

			if(udat)
			{
				if(udat->accperm != AUTH_ROOT
				|| udat->modperm != AUTH_ROOT)
				{
					ret = rms_del_rec(RMS_USER, idx);
				}
				else
				{
					/*
					 * The VChat super-user may not
					 * be deleted.
					 */

					errno = EPERM;
				}
			}
		}
	}
	else
	{
		/*
		 * The specified nickname is currently in use.
		 */

		errno = EBUSY;
	}

	return ret;
}

/*
 * Get the number of user records currently contained within the user database.
 */

uint32_t userdb_records(void)
{
	return rms_records(RMS_USER);
}

/*
 * Get the address of the user database record at index <idx>.
 * If the specified index is invalid, NULL is returned.
 */

USERDATA *userdb_addr(int idx)
{
	return (USERDATA *)rms_addr(RMS_USER, idx);
}

/*
 * Find the user specified by <nick> in the user database. Upon successful
 * operation the non-negative index of the user database record is returned.
 * Otherwise a value of -1 is returned and the global variable errno is set
 * to indicate the error.
 *
 * The userdb_find() function succeeds unless:
 *
 * ENOENT   - The specified <nick> was not found in the database.
 */

int userdb_find(char *nick)
{
	uint32_t records = rms_records(RMS_USER);

	if(records)
	{
		register USERDATA *udat = (USERDATA *)rms_addr(RMS_USER, 0);

		if(udat)
		{
			register int i;

			for(i = 0; i < records; i++, udat++)
			{
				if(!strncmp(udat->nick, nick, VPNICKSIZE))
				{
					/* FOUND */

					errno = 0;

					return i;
				}
			}
		}
	}

	errno = ENOENT;

	return -1;
}

/*
 * Find the user specified by <nick> in the user database. Upon successful
 * operation the memory address of the user database record is returned.
 * Otherwise a value of NULL is returned and the global variable errno is set
 * to indicate the error.
 *
 * The userdb_findaddr() function succeeds unless:
 *
 * ENOENT   - The specified <nick> was not found in the database.
 */

USERDATA *userdb_findaddr(char *nick)
{
	uint32_t records = rms_records(RMS_USER);

	if(records)
	{
		register USERDATA *udat	= (USERDATA *)rms_addr(RMS_USER, 0);

		if(udat)
		{
			register int i;

			for(i = 0; i < records; i++, udat++)
			{
				if(!strncmp(udat->nick, nick, VPNICKSIZE))
				{
					/* FOUND */

					errno = 0;

					return udat;
				}
			}
		}
	}

	errno = ENOENT;

	return (USERDATA *)0;
}

/*
 * Interactively create a new user database file.
 */

int userdb_create()
{
	int		fd;
	struct stat	st;
	USERDATA	udat;
	char		udbfile[PATH_MAX];
	char		backup[PATH_MAX];
	char		input[64];
	char		c;
	char		*s;
	int		i;

	snprintf(udbfile, sizeof(udbfile), "%s%s", cfdata.basedir, fn_userdb);

	/*
	 * Initialize user record.
	 */

	bzero(&udat, sizeof(udat));

	udat.accperm	= AUTH_ROOT;
	udat.modperm	= AUTH_ROOT;
	udat.sessioncnt	= 0;
	udat.cl_opts	= DEF_CLOPTMASK;

	/*
	 * Set up stdout to unbuffered mode.
	 */

#if SETVBUF_REVERSED
	setvbuf(stdout, _IONBF, NULL, 0);
#else
	setvbuf(stdout, NULL, _IONBF, 0);
#endif

	/*
	 * Display info text.
	 */

	printf
	(
	/*       [...:....1....:....2....:....3....:....4....:....5....:....6....:....7....:....]  */
		"\n"
		"This procedure will create an initial user database file for use by vserver(8),\n"
		"which will only contain one user account that will receive full administrative\n"
		"permissions. It is necessary to create a user database file before vserver can\n"
		"start into normal operations. The created user database file will be saved to:\n"
		"\n%s\n\n",

		udbfile
	);

	/*
	 * Check if the user database file already exists
	 * and find the next free backup file name if it does.
	 */

	bzero(&backup, sizeof(backup));

	if(stat(udbfile, &st) == 0)
	{
		i = 0;

		do
		{
			++i;
			snprintf(backup, sizeof(backup), "%s;%d",
				udbfile, i);

		} while (stat(backup, &st) == 0);

		printf
		(
		/*       [...:....1....:....2....:....3....:....4....:....5....:....6....:....7....:....]  */
			"WARNING: The file '%s' already exists!\n"
			"         Usually it should not be necessary to create a new user database file\n"
			"         once you already have one, execept when discarding the old one and\n"
			"         starting over with a new user database file is exactly what you want.\n"
			"         You must ascertain that vserver is not currently running, using the\n"
			"         user database file you are about to replace. Inconsistent operation\n"
			"         would otherwise result.\n\n"
			"If you go ahead, then the existing user database file will be backed up to:\n\n%s\n\n",
		       udbfile, backup
		);
	}

	/*
	 * Ask user if he wants to continue the operation.
	 */

	c = promptkey("yn", "Continue?");

	if(c == 'n')
	{
		printf("No\n");
		return -1;
	}

	printf("Yes\n");

	/*
	 * Get the user name for the initial account.
	 */

	printf
	(
	/*       [...:....1....:....2....:....3....:....4....:....5....:....6....:....7....:....]  */
		"\n"
		"The administrative user account will receive full permissions and is initially\n"
		"the only user that may create new user accounts, modifiy user's permissions and\n"
		"delete user accounts from within VChat. It can neither be deleted nor be killed.\n"
		"The username should be %d to %d characters in length and may only contain\n"
		"alphanumeric characters, underscores and dots.\n",

		4, VPNICKSIZE
	);

	for(;;)
	{
		printf("\nEnter administrative account user name (ctrl-d aborts): ");
		fflush(stdout);

		bzero(input, sizeof(input));

		if(fgets(input, sizeof(input)-1, stdin) == NULL)
		{
			printf("\nProcedure aborted.\n");
			return -1;
		}

		/*
		 * Trim off any newline and/or carriage return.
		 */

		while((s = strrchr(input, '\n')) != NULL)
			*s = '\0';
		while((s = strrchr(input, '\r')) != NULL)
			*s = '\0';

		/*
		 * Check for correct size.
		 */

		if(strlen(input) < 4 || strlen(input) > VPNICKSIZE)
		{
			printf("\nPlease choose a name %d to %d characters in length.\n",
				4, VPNICKSIZE);
			continue;
		}

		/*
		 * Check if the supplied username is valid.
		 */

		if(nick_ok(input))
		{
			sncopy(input, udat.nick, sizeof(udat.nick));
			break;
		}
		else
		{
			printf
			(
				"\n"
				"The name you have entered contains invalid characters. Please compose the name\n"
				"of alphanumeric characters, the underscore character and the dot character only.\n"
			);
		}
	}

	/*
	 * Get password for the initial account.
	 */

	for(;;)
	{
		printf("\nEnter administrative account password (ctrl-d aborts): ");

		bzero(input, sizeof(input));

		if(fgets(input, sizeof(input)-1, stdin) == NULL)
		{
			printf("\nProcedure aborted.\n");
			return -1;
		}

		/*
		 * Trim off any newline and/or carriage return.
		 */

		while((s = strrchr(input, '\n')) != NULL)
			*s = '\0';
		while((s = strrchr(input, '\r')) != NULL)
			*s = '\0';

		/*
		 * Check for correct password size.
		 */

		if(!strlen(input))
		{
			printf("\nEmpty passwords are not allowed.\n");
			continue;
		}
		else if(strlen(input) < 6)
		{
			printf("\nPlease choose a password at least 6 characters in length.\n");
			continue;
		}
		else if(strlen(input) > VPPASSWDSIZE)
		{
			printf("\nPlease choose a password at most %d characters in length.\n",
				VPPASSWDSIZE);
			continue;
		}
		else
		{
			/* Password OK */

			sncopy(input, udat.passwd, sizeof(udat.passwd));

			break;
		}
	}

	/*
	 * Print summary and ask for confirmation.
	 */

	printf("\nSummary:\n========\n\nThe following operations will be performed:\n\n");

	i = 1;

	if(backup[0])
	{
		printf("%d) Back up existing user database file\n   '%s' to '%s'\n\n",
			i++, udbfile, backup);
	}

	printf("%d) Create new user database file '%s' with:\n        user = '%s'\n    password = '%s'\n\n",
		i++, udbfile, udat.nick, udat.passwd);

	c = promptkey("yn", "Continue?");

	if(c == 'n')
	{
		printf("No\n");
		return -1;
	}

	printf("Yes\n");

	/*
	 * Back up existing user database file (if any).
	 */

	if(backup[0])
	{
		if(rename(udbfile, backup))
		{
			printf("\nFailed to rename '%s' to '%s' (%s)\nOperation aborted.\n",
				udbfile, backup, strerror(errno));

			return -1;
		}

		printf("\nSuccessfully backed up '%s' to '%s'\n",
			udbfile, backup);
	}

	/*
	 * Create new user database file.
	 */

	if((fd = open(udbfile,
		(O_WRONLY | O_TRUNC | O_CREAT | O_BINARY | O_SEQUENTIAL),
		0600)) == -1)
	{
		printf("\nFailed to open file '%s' for output (%s)\n",
			udbfile, strerror(errno));

		return -1;
	}

	/*
	 * Timestamp the user record.
	 */

	time(&udat.account_created);
	time(&udat.last_signon);

	/*
	 * Write the user record to the database file.
	 */

	if(write(fd, &udat, sizeof(udat)) != sizeof(udat))
	{
		printf("\nWrite failed on file '%s' (%s)\n",
			udbfile, strerror(errno));

		close(fd);

		return -1;
	}

	close(fd);

	bzero(&input, sizeof(input));
	bzero(&udat, sizeof(udat));

	printf("\nSuccessfully created user database file '%s'\n\n",
		udbfile);

	return 0;
}
