/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: vchatctl.c,v 1.17 2009/08/14 11:54:52 mickey Exp $ */

#include "vserver.h"
#include "vctlproto.h"

#if HAVE_UNIX_SOCKETS

  #if HAVE_SYS_SELECT_H
    #include <sys/select.h>
  #endif

/*** Defines ***/

  #ifndef AF_LOCAL
    #define AF_LOCAL AF_UNIX
  #endif

  #ifndef PF_LOCAL
    #define PF_LOCAL PF_UNIX
  #endif

  #ifndef SUN_LEN
    #define SUN_LEN(x) (strlen((x)->sun_path) + sizeof((x)->sun_family))
  #endif

/*****************************************************************************
 * VCHATCTL I/O FUNCTIONS:                                                   *
 *****************************************************************************/

/*
 * Send the control message <vctl> from the control socket to <dst>.
 * Returns 0 on success, -1 otherwise.
 */

int _sendctlmsg(struct sockaddr_un *dst, VCTL *vctl)
{
	int ret = -1;

	if(vctl)
	{
		struct timeval to;
		fd_set fds;
		int sr;

		FD_ZERO(&fds);
		FD_SET(ctlsock, &fds);

		to.tv_sec = 2;
		to.tv_usec = 0;

		/*
		 * Wait until the control socket is ready for writing
		 * or the timeout limit is hit.
		 */

		if(!(sr = select(FD_SETSIZE, 0, &fds, 0, &to)))
			errno = ETIMEDOUT;

		if(sr == 1 && FD_ISSET(ctlsock, &fds))
		{
			/*
			 * Send the control message.
			 */

			if(sendto(ctlsock, vctl, vctl->len, 0,
			(struct sockaddr *)dst, SUN_LEN(dst)) == vctl->len)
			{
				/* OK */

				ret = 0;
			}
			else
			{
				vlog(VLOG_ERR, "Failed to send control message"
					" to %s (%s)", dst->sun_path,
					strerror(errno));
			}
		}
		else
		{
			vlog(VLOG_ERR, "Unable to send control message to '%s'"
					" (%s)", dst->sun_path,
					strerror(errno));
		}
	}
	else
	{
		/*
		 * No message - No send.
		 */

		ret = 0;
	}

	return ret;
}

/*
 * Send the data pointed to by <data> of <len> bytes as a control message
 * of type <type> with flags <flags> to the address specified by <dst>.
 * Returns 0 on success, -1 otherwise.
 */

int vchatctl_send(struct sockaddr_un *dst, uint16_t type, uint16_t flags, void *data, size_t len)
{
	VCTL vctl;

	/*
	 * Prepare the control message for sending.
	 */

	bzero(&vctl, sizeof(VCTL));

	if(data)
	{
		if(len > (sizeof(vctl.data) - 1))
			len = (sizeof(vctl.data) - 1);

		if(len > 0)
			bcopy(data, vctl.data, len);
	}
	else
	{
		len = 0;
	}

	vctl.len	= VCTLSIZE(len);
	vctl.flags	= flags;
	vctl.type	= type;
	vctl.version	= VCTL_PROTOCOL;

	return _sendctlmsg(dst, &vctl);
}

/*
 * Send the string <fmt> and it's arguments as a control message of type <type>
 * with flags <flags> to the address specified by <dst>. Returns 0 on success,
 * -1 otherwise.
 */

int vchatctl_sendmsg(struct sockaddr_un *dst, uint16_t type, uint16_t flags, char *fmt, ...)
{
	va_list ap;
	VCTL vctl;
	size_t len = 0;

	va_start(ap, fmt);

	/*
	 * Prepare the control message for sending.
	 */

	bzero(&vctl, sizeof(VCTL));

	if(fmt)
		len = vsnprintf(vctl.data, sizeof(vctl.data), fmt, ap);

	vctl.len	= VCTLSIZE(len);
	vctl.flags	= flags;
	vctl.type	= type;
	vctl.version	= VCTL_PROTOCOL;

	va_end(ap);

	return _sendctlmsg(dst, &vctl);
}

/*****************************************************************************
 * VCHATCTL CONTROL COMMAND HANDLER FUNCTIONS:                               *
 *****************************************************************************/

/*
 * Kill unwanted visitors.
 */

void vchatctl_kill(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	if(len == sizeof(VCTL_KILL))
	{
		VCTL_KILL *vck = (VCTL_KILL *)buf;
		int nkill = 0;

		if(vck->target == VCTL_KILLTARG_USER)
		{
			nkill = cldb_kill(vck->nick, vck->traitor);

			if(nkill == -1)
			{
				vchatctl_sendmsg(sun, VCTL_MSG_TEXT,
					VCTL_FLG_MSGSA, "The VChat super-user "
					"'%s' may not be killed.", vck->nick);
			}
			else if(!nkill)
			{
				vchatctl_sendmsg(sun, VCTL_MSG_TEXT,
					VCTL_FLG_MSGSA, "Nickname '%s' not "
					"found.", vck->nick);
			}
			else
			{
				vchatctl_sendmsg(sun, VCTL_MSG_TEXT,
					VCTL_FLG_MSGSA, "Killed %d instance%s "
					"of '%s'.", nkill, nkill == 1 ?
					"" : "s", vck->nick);
			}
		}
		else if(vck->target == VCTL_KILLTARG_CHANNEL)
		{
			nkill = cldb_killchannel(vck->channel, vck->traitor);

			vchatctl_sendmsg(sun, VCTL_MSG_TEXT,
				VCTL_FLG_MSGSA, "Killed %d user%s from channel"
				" %d", nkill, nkill == 1 ? "" : "s",
				vck->channel);
		}
		else if(vck->target == VCTL_KILLTARG_ALL)
		{
			nkill = cldb_killall(vck->traitor);

			vchatctl_sendmsg(sun, VCTL_MSG_TEXT,
				VCTL_FLG_MSGSA, "Killed %d user%s",
				nkill, nkill == 1 ? "" : "s");
		}
		else
		{
			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"Invalid target specified.");
			return;
		}
	}
	else
	{
		vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
			"Invalid argument.");
	}
}

/*
 * Retrieve a list of all logged on users or all users on a specified
 * conference channel.
 */

void vchatctl_ulist(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	uint32_t records = cldb_records();
	register CLDATA *cp;
	int32_t *pchannel = (int32_t *)0;
	register int n = 0;

	if(len == sizeof(int32_t))
	{
		pchannel = (int32_t *)buf;

		if(pchannel && *pchannel < 0)
		{
			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"You may not request information on negative "
				"channels.");
			return;
		}
	}

	if(records && (cp = cldb_addr(0)))
	{
		register int i;
		VCTL_UINFO vu;

		for(i = n = 0; i < records; i++, cp++)
		{
			if(pchannel && cp->channel != *pchannel)
				continue;

			bzero(&vu, sizeof(vu));
			sncopy(cp->nick, vu.nick, sizeof(vu.nick));
			sncopy(cp->host, vu.host, sizeof(vu.host));
			bcopy(&cp->ip_addr, &vu.ip, sizeof(vu.ip));
			vu.channel = cp->channel;
			vu.stamp = cp->stamp;

			vchatctl_send(sun, VCTL_MSG_UINFO,
				(!n ? VCTL_FLG_MSGIT : 0), &vu, sizeof(vu));

			n++;
		}
	}

	vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGCT,
		"%d user%s listed.", n, n == 1 ? "" : "s");
}

/*
 * Get or set server option flags.
 */

void vchatctl_svopt(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	uint32_t flag = 0;
	uint8_t onoff = 0;

	switch(type)
	{
		case VCTL_CMD_ANONYMOUS:	flag = VO_ANONYMOUS;
						break;
		case VCTL_CMD_ENABLEMAIL:	flag = VO_ENABLE_MAIL;
						break;
		case VCTL_CMD_REGONLY:		flag = VO_REGONLY;
						break;
		case VCTL_CMD_FLOODGUARD:	flag = VO_FLOODGUARD;
						break;
	}

	/*
	 * Without argument -> Show server option setup only.
	 * With argument    -> Set server option and display setup.
	 */

	if(len)
	{
		uint8_t arg = *(uint8_t *)buf;

		if(arg)
			vsopt.flags |= flag;
		else
			vsopt.flags &= ~flag;

		vlog(VLOG_NOTICE, "Server options have changed: "
			"%cANONYMOUS %cMAIL_INV %cREG_ONLY %cFLOOD_GUARD",
			(vsopt.flags & VO_ANONYMOUS) ? '+' : '-',
			(vsopt.flags & VO_ENABLE_MAIL) ? '+' : '-',
			(vsopt.flags & VO_REGONLY) ? '+' : '-',
			(vsopt.flags & VO_FLOODGUARD) ? '+' : '-');

		save_options();
	}

	/*
	 * Show current server option setup.
	 */

	onoff = (vsopt.flags & flag) ? 1 : 0;

	vchatctl_send(sun, VCTL_MSG_ONOFF, VCTL_FLG_MSGSA, &onoff, 1);
}

/*
 * Send broadcast messages to all logged on users.
 */

void vchatctl_broadcast(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	char *opname = (char *)buf;
	char *optty;
	char *message;

	/*
	 * Get operator name, terminal and the broadcast message
	 * from the supplied VCTL data
	 */

	if((optty = strchr(opname, ' ')))
	{
		*optty++ = 0;

		if((message = strchr(optty, ' ')))
		{
			*message++ = 0;

			if(*message)
			{
				char *lc = (char *)(buf + (len - 1));
				int n;

				if(*lc == '\n' || *lc == '\r')
					*lc = 0;

				n = vcp_broadcast(opname, optty, "%s", message);

				vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
					"%d user%s notified.", n,
					n == 1 ? "" : "s");

				return;
			}
			else
			{
				vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
					"no message - no broadcast");
				return;
			}
		}


	}

	vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA, "ERROR: Invalid message "
		"data for broadcast command.");
}

/*
 * Shut down the VChat conference server.
 */

void vchatctl_shutdown(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
		"Shutting down VChat conference server");

	vlog(VLOG_NOTICE, "Server shutdown requested on control socket");
	cfdata.flags |= CF_SHUTDOWN;
}

/*
 * Get or set the server FloodGuard threshold time.
 */

void vchatctl_fgthreshold(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	uint32_t val;

	if(len == sizeof(uint32_t))
	{
		val = *(uint32_t *)buf;

		if(val < FG_MIN || val > FG_MAX)
		{
			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"FloodGuard threshold out of range "
				"(%d <= n <= %d)", FG_MIN, FG_MAX);
			return;
		}
		else
		{
			vsopt.t_guard.tv_sec = val / 1000000;
			vsopt.t_guard.tv_usec = val % 1000000;

			vlog(VLOG_NOTICE, "New FloodGuard threshold set: %s",
				mt_print(&vsopt.t_guard));

			save_options();
		}
	}

	/*
	 * Send current threshold setting.
	 */

	val = (vsopt.t_guard.tv_sec * 1000000) + vsopt.t_guard.tv_usec;

	vchatctl_send(sun, VCTL_MSG_FGTHRESHOLD, VCTL_FLG_MSGSA, &val, sizeof(val));
}

/*
 * Get or set the server hostname.
 */

void vchatctl_servername(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	/*
	 * Without argument -> Display current servername.
	 * With argument    -> Set new servername.
	 */

	if(len)
	{
		sncopy((char *)buf, cfdata.hostname, sizeof(cfdata.hostname));
		vlog(VLOG_NOTICE, "Hostname has been set to: %s",
			cfdata.hostname[0] ? cfdata.hostname
			: "<use-name-of-interface-address>");
	}

	vchatctl_sendmsg(sun, VCTL_MSG_SERVERNAME, VCTL_FLG_MSGSA, "%s",
		cfdata.hostname[0] ? cfdata.hostname
		: "<use-name-of-interface-address>");
}

/*
 * Get or set the server DNS checking level.
 */

void vchatctl_dnschklvl(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	uint8_t lvl;

	/*
	 * Without argument -> Display current DNS check level.
	 * With argument    -> Set new DNS check level.
	 */

	if(len)
	{
		lvl = *(uint8_t *)buf;

		if(lvl > DNS_CHK_STRICT)
		{
			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"Invalid DNS check level %d specified", lvl);
			return;
		}
		else
		{
			/*
			 * If new DNS check level is more secure than the
			 * current level, kick off all clients that do not
			 * fulfill the new requirement. Then set new DNS level.
			 */

			if(lvl > vsopt.dnschklvl)
				cldb_dnschklvl(lvl);

			vsopt.dnschklvl = lvl;

			vlog(VLOG_NOTICE, "New DNS check level: %s",
				dnslvlname(vsopt.dnschklvl));

			save_options();
		}
	}

	/*
	 * Send current DNS check level.
	 */

	lvl = vsopt.dnschklvl;

	vchatctl_send(sun, VCTL_MSG_DNSCHKLVL, VCTL_FLG_MSGSA,
		&lvl, 1);
}

/*
 * Retrieve the time since when the server is operating.
 */

void vchatctl_uptime(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	vchatctl_send(sun, VCTL_MSG_UPTIME, VCTL_FLG_MSGSA, &vs.vs_start,
		sizeof(vs.vs_start));
}

/*
 * Add an ACL rule.
 */

void vchatctl_acl_add(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	VCTL_ACL *acl = (VCTL_ACL *)buf;
	ACLRULE new;

	if(len != sizeof(VCTL_ACL))
	{
		vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
			"Invalid argument for ACL add.");
		return;
	}

	bzero(&new, sizeof(new));

	if(acl->rule > 65534)
	{
		vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
			"Invalid rule number specified (0 <= r <= 65534).");
		return;
	}

	new.rule = acl->rule;
	bcopy(&acl->address, &new.address, sizeof(new.address));
	bcopy(&acl->netmask, &new.netmask, sizeof(new.netmask));
	new.action = (acl->flags & VCTL_ACL_ACCEPT) ? ACL_ACCEPT : ACL_REJECT;

	if(acl->expires)
	{
		time(&new.expires);

		if(acl->flags & VCTL_ACL_TTL)
		{
			new.expires += acl->expires;
		}
		else if(acl->expires <= new.expires)
		{
			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"Specified rule expiration time already reached.");
			return;
		}
		else
		{
			new.expires = acl->expires;
		}
	}

	if(!acldb_add(&new))
	{
		char buf[256];

		acldb_print(&new, buf, sizeof(buf));
		vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA, "add %s", buf);
	}
	else
	{
		vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
			"Failed to add ACL rule 0x%04X (%s)",
			new.rule, strerror(errno));
	}
}

/*
 * Delete an ACL rule.
 */

void vchatctl_acl_delete(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	if(len == sizeof(uint16_t))
	{
		uint16_t rule = *(uint16_t *)buf;

		if(!acldb_delete(rule))
		{
			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"delete 0x%04X", rule);
		}
		else
		{
			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"Failed to delete ACL rule 0x%04X (%s)",
					rule, strerror(errno));
		}
	}
	else
	{
		vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
			"ERROR: Invalid argument for ACL flush.");
	}
}

/*
 * Flush all ACL rules.
 */

void vchatctl_acl_flush(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	acldb_flush();
	vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA, "Flushed ACL database");
}

/*
 * Show the current server ACL setup.
 */

void vchatctl_acl_list(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	uint32_t records = acldb_records();
	register ACLRULE *rp;

	if(records && (rp = acldb_addr(0)))
	{
		char buf[256];
		register int i;

		for(i = 0; i < records; i++, rp++)
		{
			acldb_print(rp, buf, sizeof(buf));

			vchatctl_sendmsg(sun, VCTL_MSG_TEXT,
				(!i ? VCTL_FLG_MSGIT : 0), "%s", buf);
		}
	}

	vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGCT,
		"0xFFFF accept any");
}

/*
 * Check if a user name is present in the user database.
 */

void vchatctl_chk_user(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	uint8_t yesno = 0;

	if(len && userdb_find((char *)buf) >= 0)
		yesno = 1;

	vchatctl_send(sun, VCTL_MSG_ONOFF, VCTL_FLG_MSGSA, &yesno, 1);
}

/*
 * Add a user to the user database.
 */

void vchatctl_add_user(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	if(len == sizeof(VCTL_NEWUSER))
	{
		VCTL_NEWUSER *nu = (VCTL_NEWUSER *)buf;

		nu->nick[sizeof(nu->nick)-1] = 0;
		nu->passwd[sizeof(nu->passwd)-1] = 0;

		if(!userdb_add(nu->nick, nu->passwd))
		{
			userdb_save();
			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"Successfully added user '%s'.", nu->nick);
		}
		else if(errno == EEXIST)
		{
			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"User '%s' already exists.", nu->nick);
		}
		else if(errno == EBUSY)
		{
			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"Nickname '%s' is currently in use.",
				nu->nick);
		}
		else
		{
			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"Failed to add user '%s' (%s)",
				nu->nick, strerror(errno));
		}
	}
	else
	{
		vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
			"Invalid argument.");
	}
}

/*
 * Delete a user from the user database.
 */

void vchatctl_del_user(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	if(len)
	{
		if(!userdb_delete((char *)buf))
		{
			userdb_save();
			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"Deleted user '%s' from the user database.",
				(char *)buf);
		}
		else if(errno == ENOENT)
		{
			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"User '%s' not found in user database.",
				(char *)buf);
		}
		else
		{
			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"Failed to delete user '%s' (%s)",
				(char *)buf, strerror(errno));
		}
	}
	else
	{
		vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA, "No user name "
			"specified to delete.");
	}
}

/*
 * Send a message to a channel or user.
 */

void vchatctl_cnmsg(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	if(len > sizeof(VCTL_CNMSG))
	{
		VCTL_CNMSG *cn = (VCTL_CNMSG *)buf;
		char *msg = (char *)(buf + sizeof(VCTL_CNMSG));

		if(*msg)
		{
			uint32_t n = 0;
			char *lc = (char *)(buf + (len - 1));

			if(*lc == '\n' || *lc == '\r')
				*lc = 0;

			if(cn->to[0])
			{
				/* Send message to specified nickname */

				n = vcp_msgnick(cn->to, "(%s%s*%s%s%s%s%s@%s%s%s%s%s%s*%s) %s%s",
					S_BOLD, S_BLINK, S_OFF,
					S_FG_GRN, cn->from, S_OFF,
					S_BOLD, S_OFF,
					S_FG_YEL, cn->tty, S_OFF,
					S_BOLD, S_BLINK, S_OFF,
					msg, S_OFF);
			}
			else
			{
				/* Send message to specified channel */

				n = vcp_msgchan(cn->channel, (CLDATA *)0,
					"(%s%s%s%s@%s%s%s%s) %s%s",
					S_FG_GRN, cn->from, S_OFF,
					S_BOLD, S_OFF,
					S_FG_YEL, cn->tty, S_OFF,
					msg, S_OFF);
			}

			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"%d user%s notified.", n, n == 1 ? "" : "s");
		}
		else
		{
			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"No message.");
		}
	}
	else
	{
		vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
			"Invalid argument.");
	}
}

/*
 * Retrieve lastlog information.
 */

void vchatctl_lastlog(struct sockaddr_un *sun, uint16_t type, void *buf, size_t len)
{
	if(len == sizeof(VCTL_LLQUERY))
	{
		VCTL_LLQUERY *llq = (VCTL_LLQUERY *)buf;

		if(lastlog)
		{
			VCTL_LLRESP r;
			register LLOGDAT *ldat;
			register int i;
			register int n;

			/*
			 * If no max number of records is specified,
			 * then list all available records.
			 */

			if(!llq->records)
				llq->records = lastlog->header->used;

			for(i = lastlog->header->used, n = 0; i >= 0
			&& n < llq->records; --i)
			{
				/*
				 * Get the next lastlog record.
				 */

				if(!(ldat = lastlog_get_record(i)))
				{
					if(lastlog)
					{
						continue;
					}
					else
					{
						break;
					}
				}

				/*
				 * If an until date is specified, skip
				 * records that are newer than that.
				 */

				if(llq->until && ldat->signon_stamp > llq->until)
					continue;

				/*
				 * Stop on the first record that is
				 * older than our given since date.
				 */

				if(ldat->signoff_stamp < llq->since)
					break;

				/*
				 * Skip records that do not match the
				 * specified IP network.
				 */

				if((ldat->ipaddr.s_addr & llq->netmask.s_addr)
				!= (llq->address.s_addr & llq->netmask.s_addr))
					continue;

				/*
				 * Skip records not matching the
				 * specified username (if any).
				 */

				if(llq->user[0] && strcmp(llq->user, ldat->username))
					continue;

				/*
				 * Skip records not matching the
				 * specified nickname (if any).
				 */

				if(llq->nick[0] && strcmp(llq->nick, ldat->nickname))
					continue;

				/*
				 * Skip records not matching the
				 * specified hostname (if any).
				 */

				if(llq->host[0] && strcmp(llq->host, ldat->hostname))
					continue;

				/*
				 * OK, we have a record - list it.
				 */

				r.signon = ldat->signon_stamp;
				r.signoff = ldat->signoff_stamp;
				sncopy(ldat->username, r.user, sizeof(r.user));
				sncopy(ldat->nickname, r.nick, sizeof(r.nick));
				sncopy(ldat->info, r.info, sizeof(r.info));
				bcopy(&ldat->ipaddr, &r.address, sizeof(r.address));
				sncopy(ldat->hostname, r.host, sizeof(r.host));

				if(vchatctl_send(sun, VCTL_MSG_LASTLOG,
				!n ? VCTL_FLG_MSGIT : 0, &r, sizeof(r)))
				{
					/* Send error */

					return;
				}

				n++;

				usleep(600);
			}

			if((ldat = lastlog_get_record(0)))
			{
				char dtbuf[64];

				strftime(dtbuf, sizeof(dtbuf),
					"%a, %d %b %Y %H:%M:%S %z",
					localtime(&ldat->signon_stamp));
				dtbuf[sizeof(dtbuf)-1] = 0;

				vchatctl_sendmsg(sun, VCTL_MSG_TEXT, 0,
					"Lastlog begins: %s", dtbuf);
			}

			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGCT,
				"%d Records listed.", n);
		}
		else
		{
			vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"Lastlog is disabled.");
		}
	}
	else
	{
		vchatctl_sendmsg(sun, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
			"Invalid argument.");
	}
}

/*****************************************************************************
 * VCHATCTL CONTROL COMMAND VECTOR:                                          *
 *****************************************************************************/

typedef void (*VCTLCMDFUNC)(struct sockaddr_un *, uint16_t type, void *buf, size_t len);

VCTLCMDFUNC vctlcmd[] =
{
	vchatctl_ulist,
	vchatctl_broadcast,
	vchatctl_kill,
	vchatctl_svopt,
	vchatctl_svopt,
	vchatctl_svopt,
	vchatctl_svopt,
	vchatctl_fgthreshold,
	vchatctl_shutdown,
	vchatctl_dnschklvl,
	vchatctl_servername,
	vchatctl_uptime,
	vchatctl_acl_add,
	vchatctl_acl_delete,
	vchatctl_acl_flush,
	vchatctl_acl_list,
	vchatctl_chk_user,
	vchatctl_add_user,
	vchatctl_del_user,
	vchatctl_cnmsg,
	vchatctl_lastlog
};

int numvctlcmd = sizeof(vctlcmd) / sizeof(VCTLCMDFUNC);

/*****************************************************************************
 * VCHATCTL MESSAGE BUFFERING AND COMMAND EXECUTION FUNCTION:                *
 *****************************************************************************/

/*
 * Coalesce the control message <vctl> originating from <sender> into the
 * buffer, and if the message has the VCTL_FLG_MSGCT flag set, execute the
 * corresponding control command handler function upon the buffer contents.
 * If called with an argument of NULL, the function will explicitly reset
 * the buffer.
 */

void vchatctl_buffer(struct sockaddr_un *sender, VCTL *vctl)
{
	static uint8_t rcvbuf[SVMSGBUF];
	static uint16_t curtype = 0xFFFF;
	static size_t inbuf = 0;
	size_t remain;

	/*
	 * If no VCTL is provided, explicitly reset the buffer and return.
	 */

	if(!vctl || !sender)
	{
		inbuf = 0;
		curtype = 0xFFFF;
		return;
	}

	/*
	 * Start buffering when the message initiates the transaction.
	 * Successive messages must match the type we are currently buffering,
	 * otherwise the buffer is reset and the message discarded.
	 */

	if(vctl->flags & VCTL_FLG_MSGIT)
	{
		inbuf = 0;
		curtype = vctl->type;
	}
	else if(vctl->type != curtype)
	{
		inbuf = 0;
		curtype = 0xFFFF;

		return;
	}

	/*
	 * Calculate remaining buffer space.
	 */

	remain = (SVMSGBUF-1) - inbuf;

	/*
	 * Set the len field of the VCTL message to the payload length.
	 */

	vctl->len -= VCTLSIZE(0);

	/*
	 * Copy message data to buffer.
	 */

	if(vctl->len <= remain)
	{
		bcopy(&vctl->data[0], &rcvbuf[inbuf], vctl->len);
		inbuf += vctl->len;

		/*
		 * Message completes transaction?
		 */

		if(vctl->flags & VCTL_FLG_MSGCT)
		{
			/* YES */

			vlog(VLOG_NOTICE, "Executing control command $%04X "
				"received from '%s' with %d bytes of data",
				curtype, sender->sun_path, inbuf);

			rcvbuf[inbuf] = 0;
			vctlcmd[curtype](sender, curtype, &rcvbuf[0], inbuf);
		}
		else
		{
			/* NO */

			return;
		}
	}
	else
	{
		vlog(VLOG_WARNING, "Control message buffer space exhausted "
			"- message discarded");

		vchatctl_sendmsg(sender, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
			"ERROR: Control message buffer space exhausted "
			"- message discarded.");
	}

	/*
	 * Reset buffer on error or transaction completed.
	 */

	inbuf = 0;
	curtype = 0xFFFF;
}

/*****************************************************************************
 * VCHATCTL EXTERNALLY VISIBLE FUNCTIONS:                                    *
 *****************************************************************************/

/*
 * Open and initialize the server control socket.
 * Returns 0 if successful, -1 in case of failure.
 */

int vchatctl_open(void)
{
	int s;

	/*
	 * Initialize UDP socket
	 */

	if((s = socket(PF_LOCAL, SOCK_DGRAM, 0)) != -1)
	{
		saun.sun_family = AF_LOCAL;
		snprintf(saun.sun_path, sizeof(saun.sun_path), "%s%s",
			cfdata.basedir, fn_ctlsock);

  #if HAVE_STRUCT_SOCKADDR_UN_SUN_LEN
		saun.sun_len = SUN_LEN(&saun);
  #endif
		unlink(saun.sun_path);

		if(!bind(s, (struct sockaddr *) &saun, SUN_LEN(&saun)))
		{
			/*
			 * Set the socket permissions to 0660 so that
			 * owner and group may write to it.
			 */

			chmod(saun.sun_path, 0660);

			/* OK */

			ctlsock = s;

			return 0;
		}
		else
		{
			vlog(VLOG_ERR, "Failed to bind control socket, "
				"status=%d (%s)", errno, strerror(errno));
		}

		close(s);
	}
	else
	{
		vlog(VLOG_ERR, "Failed to open control socket, status=%d (%s)",
			errno, strerror(errno));
	}

	return -1;
}

/*
 * Dispose of the server control socket.
 */

void vchatctl_close(void)
{
	if(ctlsock)
	{
		close(ctlsock);
		unlink(saun.sun_path);
		ctlsock = 0;
	}
}

/*
 * Receive and process a control command, waiting
 * on the server control socket.
 */

void vchatctl_rcvcmd(void)
{
	VCTL vctl;
	int msglen;
	struct sockaddr_un fsa;
	socklen_t flen = sizeof(fsa);

	bzero(&fsa, flen);
	bzero(&vctl, sizeof(vctl));

	/*
	 * Get the message, check protocol version and message length.
	 */

	if((msglen = recvfrom(ctlsock, &vctl, sizeof(VCTL), 0, (struct sockaddr *)&fsa, &flen)) >= 0)
	{
		if(msglen >= VCTLSIZE(0))
		{
			if(vctl.version == VCTL_PROTOCOL)
			{
				if(vctl.len == msglen)
				{
					if(vctl.type < numvctlcmd)
					{
						/*
						 * Buffer the message and
						 * execute control command
						 * when all successive messages
						 * have been received.
						 */

						vchatctl_buffer(&fsa, &vctl);

						return;
					}
					else
					{
						vlog(VLOG_ERR, "Received control message for command "
							"$%04x which is not implemented", vctl.type);
	
						vchatctl_sendmsg(&fsa, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
							"ERROR: Control command $%04x not implemented.",
							vctl.type);
					}
				}
				else
				{
					vlog(VLOG_ERR, "Received truncated "
						"control message (%d/%d bytes)",
						msglen, vctl.len);

					vchatctl_sendmsg(&fsa, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
						"ERROR: Truncated message received (%d/%d bytes).",
						msglen, vctl.len);
				}
			}
			else
			{
				vlog(VLOG_ERR, "Received control message with "
					"invalid protocol version $%04x",
					vctl.version);
				vchatctl_sendmsg(&fsa, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
					"ERROR: Protocol version mismatch ($%04X/$%04X).",
					vctl.version, VCTL_PROTOCOL);
			}
		}
		else
		{
			vlog(VLOG_ERR, "Received control message of invalid "
				"length %d", msglen);
			vchatctl_sendmsg(&fsa, VCTL_MSG_TEXT, VCTL_FLG_MSGSA,
				"ERROR: Control message of invalid length "
				"received (%d/%d bytes).", msglen, VCTLSIZE(0));
		}
	}
	else
	{
		vlog(VLOG_ERR, "Failed to receive control message (%s)",
			strerror(errno));
	}

	/*
	 * Reset buffer on error.
	 */

	vchatctl_buffer((struct sockaddr_un *)0, (VCTL *)0);
}

#endif /* HAVE_UNIX_SOCKETS */
