/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: vcp.c,v 1.4 2009/08/14 11:54:52 mickey Exp $ */

#include "vserver.h"

/*
 * Accept a new client connection on the server socket, get the remote and
 * local interface addresses, check the ACL database if client is allowed
 * to connect, perform DNS checks upon the client address and/or hostname
 * and determine a hostname for the local interface. If the client passes
 * these checks and no error condition is encountered so far, then the data
 * listed below is filled into a new CLDATA record and the vcp_handshake()
 * function is called to continue with the VChat protocol negotiation. If
 * vcp_handshake returns 0 then a value of 0 is returned. In case of failure,
 * including vcp_handshake() returning a non-zero exit status, a value of
 * -1 is returned.
 *
 * CLDATA filled in by vcp_setupconn upon success:
 *
 * fd          - Client socket descriptor
 * ip_addr     - Client internet address
 * if_addr     - Local interface address
 * host        - Client host name
 * if_name     - Local interface host name
 * dnschklvl   - The DNS check level that has been determined
 */

int vcp_setupconn(void)
{
	CLDATA new;
	struct sockaddr_in peer;
	socklen_t peerlen = sizeof(peer);

	bzero(&new, sizeof(new));

	if((new.fd = accept(sock, (struct sockaddr *)&peer, &peerlen)) != -1)
	{
		struct sockaddr_in itf;
		socklen_t itflen = sizeof(itf);

		vs.vs_total_conn++;
		vs.vs_active_conn++;

		if(!getsockname(new.fd, (struct sockaddr *)&itf, &itflen))
		{
			char acl[256];

			bcopy(&peer.sin_addr, &new.ip_addr, sizeof(new.ip_addr));
			bcopy(&itf.sin_addr, &new.if_addr, sizeof(new.if_addr));

			if(acldb_authorize(peer.sin_addr, acl, sizeof(acl)))
			{
				new.dnschklvl = dns_check(&peer.sin_addr,
					new.host, sizeof(new.host));

				if(new.dnschklvl >= vsopt.dnschklvl)
				{
					if(cfdata.hostname[0])
					{
						sncopy(cfdata.hostname,
							new.if_name,
							sizeof(new.if_name));
					}
					else
					{
						sncopy(iptoname(itf.sin_addr),
							new.if_name,
							sizeof(new.if_name));
					}

					vlog(VLOG_INFO, "Connection accepted HOST: %s:%d IF: %s:%d DNSLVL: %d ACL: %s",
						iptostr(peer.sin_addr), ntohs(peer.sin_port),
						iptostr(itf.sin_addr), ntohs(itf.sin_port),
						new.dnschklvl, acl);

					if(!vcp_handshake(&new))
					{
						/* OK */

						return 0;
					}
				}
				else
				{
					vlog(VLOG_NOTICE, "Connection rejected HOST: %s:%d IF: %s:%d REASON: Level %d dns clearance required, level %d too weak",
						iptostr(peer.sin_addr), ntohs(peer.sin_port),
						iptostr(itf.sin_addr), ntohs(itf.sin_port),
						vsopt.dnschklvl,
						new.dnschklvl);

					vio_write(new.fd, ident(CONN_EDNSFAIL), sizeof(VIDENT), 1);
				}
			}
			else
			{
				vlog(VLOG_NOTICE, "Connection rejected HOST: %s:%d IF: %s:%d ACL: %s",
					iptostr(peer.sin_addr), ntohs(peer.sin_port),
					iptostr(itf.sin_addr), ntohs(itf.sin_port),
					acl);

				vio_write(new.fd, ident(CONN_EDENIED), sizeof(VIDENT), 1);
			}
		}
		else
		{
			vlog(VLOG_ERR, "Failed to getsockname(2) for client "
				"socket %d (%s)", new.fd, strerror(errno));

			vs.vs_err_conn++;
		}

		vs.vs_active_conn--;
#ifdef WIN32
		closesocket(new.fd);
#else
		close(new.fd);
#endif
	}
	else
	{
		vlog(VLOG_ERR, "Failed to accept(2) client, status=%d (%s)",
			new.fd, strerror(errno));

		vs.vs_err_conn++;
	}

	vlog(VLOG_INFO, "Client disconnected");

	return -1;
}

/*
 * Perform the initial VChat protocol handshake with the client identified
 * by the CLDATA structure pointed to by <new> that has previously been
 * setup by the vcp_setupconn() function. If no errors are encountered and
 * the client supplied a valid protocol, client type and magic number, the
 * data listed below is added to the CLDATA record pointed to by <new> and
 * the vcp_login() function is called to continue with the VChat protocol
 * login procedure. If the vcp_login() function returns with a status of 0,
 * then a value of 0 is returned. In case of failure, including vcp_login()
 * returning with a non-zero exit status, a value of -1 is returned and the
 * contents of the CLDATA <new> are indeterminate.
 *
 * Data added to CLDATA <new> by vcp_handshake() upon success:
 *
 * type       - Client type (CLIENT_FRONTEND or CLIENT_GATEWAY)
 * protocol   - Protocol revision used by client
 * ident      - Software identification string submitted by client
 * channel    - Initial conference channel number
 *
 * Additionally, if the client identified itself as type CLIENT_GATEWAY, the
 * CLDATA field 'ip_addr' is moved to 'gw_addr' and 'host' to 'gw_name'
 * respectively.
 */

int vcp_handshake(CLDATA *new)
{
	if(vio_write(new->fd, ident(CONN_OK), sizeof(VIDENT), CONN_TIMEOUT) == sizeof(VIDENT))
	{
		VIDENT vid;

		if(vio_read(new->fd, &vid, sizeof(vid), CONN_TIMEOUT) == sizeof(vid))
		{
			uint8_t status;

			vid.type  = ntohl(vid.type);
			vid.proto = ntohl(vid.proto);
			vid.magic = ntohl(vid.magic);
			vid.ident[VPIDENTSIZE] = 0;

			if(vid.magic == VCLIENT_MAGIC)
			{
				if(MAJOR_REV(vid.proto) == MAJOR_REV(PROTO_REV))
				{
					/*
					 * Extract the initial channel number
					 * from the most significant 24 bits
					 * of the "type" field.
					 */

					new->channel = ((vid.type
						& 0xffffff00) >> 8);

					if(vid.type & 0x80000000)
						new->channel |= 0xff000000;

					vid.type &= 0x000000ff;

					/*
					 * Check client type,
					 */

					if(vid.type == CLIENT_FRONTEND
					|| vid.type == CLIENT_GATEWAY)
					{
						vlog(VLOG_INFO, "Remote client IDENT: %s TYPE: %s PROTOCOL: %x-%x",
							vid.ident,
							vid.type == CLIENT_FRONTEND ?
							"frontend" : "gateway",
							MAJOR_REV(vid.proto),
							MINOR_REV(vid.proto));

						status = CONN_OK;

						if(vio_write(new->fd, &status,
							1, CONN_TIMEOUT) == 1)
						{
							/*
							 * Fill in client type
							 * protocol and ident.
							 */

							new->type = vid.type;
							new->protocol = vid.proto;
							sncopy(vid.ident, new->ident, sizeof(new->ident));

							/*
							 * If the client is a
							 * gateway, move IP
							 * and host name to
							 * that of gateway.
							 */

							if(vid.type == CLIENT_GATEWAY)
							{
								bcopy(&new->ip_addr, &new->gw_addr, sizeof(new->gw_addr));
								sncopy(new->host, new->gw_name, sizeof(new->gw_name));
								bzero(&new->ip_addr, sizeof(new->ip_addr));
								bzero(&new->host, sizeof(new->host));
							}

							/* OK */

							if(!vcp_login(new))
							{
								return 0;
							}
						}
						else
						{
							vlog(VLOG_ERR, "Failed to send status (%s)",
								strerror(errno));
						}
					}
					else
					{
						vlog(VLOG_ERR, "Client type $%02x invalid",
							vid.type);

						status = CONN_ECLTYPE;
					}
				}
				else
				{
					vlog(VLOG_ERR, "Invalid major protocol"
						" revision $%04x",
						MAJOR_REV(vid.proto));

					status = CONN_EPROTO;
				}
			}
			else
			{
				vlog(VLOG_ERR, "Magic number mismatch "
					"($%08x/$%08x)", vid.magic,
					VCLIENT_MAGIC);

				status = CONN_EMAGIC;
			}

			/* Send error status if any */

			if(status != CONN_OK)
				vio_write(new->fd, &status, 1, 1);
		}
		else
		{
			vlog(VLOG_ERR, "Failed waiting for client VIDENT (%s)",
				strerror(errno));
		}
	}
	else
	{
		vlog(VLOG_ERR, "Failed to send VIDENT packet (%s)",
			strerror(errno));
	}

	return -1;
}

/*
 * Perform the VChat protocol login procedure with the client identified by
 * the CLDATA pointed to by <new>. For clients of type CLIENT_GATEWAY the
 * vcp_gwcheck() function is called, which performs additional ACL and DNS
 * checks upon the "end user" internet address, supplied as part of the login
 * procedure. If no error is encountered and the client has submitted a valid
 * nickname, the supplied nickname is checked against the user database for
 * existence. If it is found within the user database, the vcp_auth() function
 * is called to perform the additional authorization protocol stage. If the
 * vcp_auth() function returns a status of 0, a value of 0 is returned.
 * In case the supplied nickname is not found within the user database, it is
 * checked against the client database and rejected if it is already in use.
 * Otherwise the client option flags within the CLDATA record are set to the
 * default and the vcp_finish() function is called in order to add the CLDATA
 * <new> to the client database and send the final CONN_OK status to the
 * client. If vcp_finish() returns with a status of 0, then a value of 0 is
 * returned. If at any point the client is rejected for whatever reason, or
 * an error condition arises, including vcp_gwchek(), vcp_auth() and/or
 * vcp_finish() returning an error status, a value of -1 is returned and
 * the contents of the CLDATA <new> are indeterminate.
 *
 * CLDATA filled in by vcp_login() upon success:
 *
 * nick      - The nickname of the client
 * nicklen   - Length of nickname excluding the terminating NUL character
 * user      - The username of the client
 * ip_addr   - (GATEWAY) The internet address of the user behind the gateway
 * host      - (GATEWAY) The hostname of the user behind the gateway
 * dnschklvl - (GATEWAY) The lower of the gateway and client address dnslevel
 * cl_opts   - Client options, either default or taken from user record
 */

int vcp_login(CLDATA *new)
{
	VLOGINDAT vl;

	if(vio_read(new->fd, &vl, sizeof(vl), CONN_TIMEOUT) == sizeof(vl))
	{
		uint8_t status = CONN_OK;

		vl.uname[VPUSERSIZE] = 0;
		vl.nick[VPNICKSIZE] = 0;

		/*
		 * Clients of type FRONTEND which did not supply a
		 * nickname, receive their username as nickname.
		 */

		if(new->type == CLIENT_FRONTEND && !vl.nick[0])
			sncopy(vl.uname, vl.nick, sizeof(vl.nick));

		/*
		 * For clients of type GATEWAY we call vcp_gwcheck()
		 * in order to perform additional ACL and DNS checks
		 * upon the internet address of the "end user" which
		 * has been supplied as part of the VLOGINDAT packet.
		 * The vcp_gwcheck() function returns a partial status
		 * which is either an error status or CONN_OK when we
		 * should go ahead.
		 */

		if(new->type == CLIENT_GATEWAY)
			status = vcp_gwcheck(new, &vl);

		if(status == CONN_OK)
		{
			/*
			 * Accept only clients that have a valid nickname.
			 */

			if(nick_ok(vl.nick))
			{
				USERDATA *ud;

				sncopy(vl.uname, new->user, sizeof(new->user));
				new->nicklen = sncopy(vl.nick, new->nick, sizeof(new->nick));

				if((ud = userdb_findaddr(vl.nick)))
				{
					/* Registered user needs auth */

					if(new->protocol >= 0x0004000e)
					{
						/*
						 * Protocol revision >= 4-e
						 *
						 * Perform authorization and
						 * return status.
						 */

						if(!vcp_auth(new, ud))
						{
							/* OK */

							return 0;
						}
					}
					else
					{
						/*
						 * Protocol Revision <= 4-d
						 *
						 * COMPAT:
						 * =======
						 *
						 * Client can not perform
						 * authorization so return a
						 * status of CONN_ENDUP here.
						 */

						vlog(VLOG_NOTICE, "Supplied nickname '%s' requires authentication, "
							"but client protocol %x-%x does not support this",
							new->nick, MAJOR_REV(new->protocol),
							MINOR_REV(new->protocol));

						status = CONN_ENDUP;
					}
				}
				else if(errno == ENOENT)
				{
					/* Unregistered user */

					if(!(vsopt.flags & VO_REGONLY))
					{
						/*
						 * Check for duplicated
						 * nicknames. Unregistered
						 * nicknames may only be
						 * present once at every
						 * given time.
						 */

						if(!cldb_chkpresent(vl.nick))
						{
							new->cl_opts = DEF_CLOPTMASK;

							/*
							 * Add CLDATA record to
							 * client database and
							 * send final CONN_OK.
							 */

							if(!vcp_finish(new))
							{
								/* OK */

								return 0;
							}
						}
						else
						{
							vlog(VLOG_NOTICE, "Duplicated nickname '%s' invalid",
								vl.nick);

							status = CONN_ENDUP;
						}
					}
					else
					{
						vlog(VLOG_NOTICE, "Nickname '%s' not registered",
							vl.nick);

						status = CONN_EDENIED;
					}

				}
				else
				{
					vlog(VLOG_ERR, "User database failure (%s)",
						strerror(errno));

					status = CONN_ENOMEM;
				}
			}
			else
			{
				vlog(VLOG_NOTICE, "Supplied nickname '%s' invalid",
					vl.nick);

				status = CONN_ENINVAL;
			}
		}

		/* Send error status if any */

		if(status != CONN_OK)
			vio_write(new->fd, &status, 1, 1);
	}
	else
	{
		vlog(VLOG_ERR, "Failed to read VLOGINDAT (%s)",
			strerror(errno));
	}

	return -1;
}

/*
 * Check the actual client internet address supplied by a gateway client
 * against the ACL and perform addtional DNS checks upon that address.
 * Upon successful completion a partial status of CONN_OK is returned
 * and the data listed below has been added to the CLDATA <new>. Otherwise
 * an error status is returned.
 *
 * CLDATA filled in upon success:
 *
 * uname     - Since gateway clients have no username, this is set to "<none>"
 * ip_addr   - The internet address of the user behind the gateway
 * host      - The hostname of the user behind the gateway
 * dnschklvl - The lower of the gateway / user address dns check level
 */

uint8_t vcp_gwcheck(CLDATA *new, VLOGINDAT *vl)
{
	uint8_t status;
	char acl[256];

	if(acldb_authorize(vl->remote_user_ip, acl, sizeof(acl)))
	{
		uint8_t lvl;

		bcopy(&vl->remote_user_ip, &new->ip_addr, sizeof(new->ip_addr));

		lvl = dns_check(&new->ip_addr, new->host, sizeof(new->host));

		if(lvl >= vsopt.dnschklvl)
		{
			/*
			 * The DNS check level for this client will be the
			 * lower of the gateway address check level and the
			 * client address check level.
			 */

			if(lvl < new->dnschklvl)
				new->dnschklvl = lvl;

			/*
			 * Connections of type GATEWAY do not
			 * have a username.
			 */

			snprintf(vl->uname, sizeof(vl->uname), "<none>");

			vlog(VLOG_INFO, "Gateway user accepted HOST: %s VIA: %s DNSLVL: %d ACL: %s",
				iptostr(vl->remote_user_ip),
				iptostr(new->gw_addr),
				new->dnschklvl, acl);

			/* OK */

			status = CONN_OK;
		}
		else
		{
			vlog(VLOG_NOTICE, "Gateway user rejected HOST: %s VIA: %s REASON: Level %d dns clearance required, level %d too weak.",
				iptostr(vl->remote_user_ip),
				iptostr(new->gw_addr),
				vsopt.dnschklvl, lvl);

			status = CONN_EDNSFAIL;
		}
	}
	else
	{
		vlog(VLOG_NOTICE, "Gateway user rejected HOST: %s VIA: %s ACL: %s",
			iptostr(vl->remote_user_ip), iptostr(new->gw_addr),
			acl);

		status = CONN_EDENIED;
	}

	return status;
}

/*
 * Perform the VChat protocol authorization procedure, that is required for
 * registered nicknames, with the client identified by the CLDATA pointed to
 * by <new> and the user database record pointed to by <ud>. If no error is
 * encountered and the client supplied the correct password for the specified
 * nickname, then the data listed below is added to the CLDATA <new> and the
 * vcp_finish() function is called in order to add the CLDATA <new> to the
 * client database and send the final CONN_OK status to the client. When the
 * vcp_finish() function returns with a status of 0, the timestamp of last
 * login within the user database record <ud> is updated to the current time,
 * the last logout timestamp is set to 0xffffffff and the session count is
 * incremented by one. Thereafter the modified user database is saved back
 * to file and a value of 0 is returned. If at any point the client is rejected
 * or an error condition is encountered, including the vcp_finish() function
 * returning a non-zero status, a value of -1 is returned and the contents
 * of the CLDATA <new> are indeterminate.
 *
 * CLDATA filled in by vcp_auth() upon success:
 *
 * accperm   - Access permission mask taken from user database record
 * modperm   - Modification permission mask taken from user database record
 * cl_opts   - Client option mask taken from user database record
 * flags     - The CFLAG_REGISTERED bit is set to indicate a registered user
 * lastlogin - Time of last login from user database record
 */

int vcp_auth(CLDATA *new, USERDATA *ud)
{
	uint8_t status = CONN_AUTHREQ;

	if(vio_write(new->fd, &status, 1, CONN_TIMEOUT) == 1)
	{
		VPASSWD	vpwd;

		if(vio_read(new->fd, &vpwd, sizeof(vpwd), CONN_TIMEOUT) == sizeof(vpwd))
		{
			vpwd.passwd[VPPASSWDSIZE] = 0;

			/*
			 * Check the supplied password against
			 * the user database record.
			 */

			if(!strncmp(vpwd.passwd, ud->passwd, VPPASSWDSIZE))
			{
				/*
				 * Password matched.
				 * Copy permission masks and client options
				 * for this client from user database record.
				 */

				vlog(VLOG_INFO, "Password ok for nickname '%s'",
					new->nick);

				new->accperm = ud->accperm;
				new->modperm = ud->modperm;
				new->cl_opts = ud->cl_opts;

				new->flags |= CFLAG_REGISTERED;

				new->lastlogin = ud->last_signon;

				/*
				 * Add new CLDATA record to client database
				 * and send the final CONN_OK status.
				 */

				status = CONN_OK;

				if(!vcp_finish(new))
				{
					/*
					 * Update last login/logoff timestamps
					 * and session counter in user database
					 * record, and save the modified data
					 * back.
					 */

					time(&ud->last_signon);
					ud->last_signoff = 0xffffffff;
					++ud->sessioncnt;

					userdb_save();

					/* OK */

					return 0;
				}
			}
			else
			{
				vlog(VLOG_NOTICE, "Invalid password for nickname '%s'",
					new->nick);

				status = CONN_EAUTHFAIL;
			}

			/* Send error status if any */

			if(status != CONN_OK)
				vio_write(new->fd, &status, 1, 1);
		}
		else
		{
			vlog(VLOG_ERR, "Failed to read VPASSWD (%s)",
				strerror(errno));
		}
	}
	else
	{
		vlog(VLOG_ERR, "Failed to send status (%s)", strerror(errno));
	}

	return -1;
}

/*
 * Finish the VChat protocol negotiation by adding the CLDATA pointed to by
 * <new> to the client database and sending the final status byte to the
 * client. In case a database error is encountered while adding <new>, a
 * status of CONN_ENOMEM is sent to the client and a value of -1 is returned.
 * Upon success, a status of CONN_OK is sent to the client. If sending of the
 * final CONN_OK status fails, the newly added CLDATA record is immediately
 * removed from the client database and a value of -1 is returned. Provided
 * that neither adding the new CLDATA record nor sending the final CONN_OK
 * status failed, the cmd_signon() function is called to sign on the new
 * client to the conference and to inform others about the new user.
 * After cmd_signon() has returned, a value of 0 is returned.
 */

int vcp_finish(CLDATA *new)
{
	uint8_t status;
	CLDATA *rec;

	/*
	 * Set the timestamps on the new CLDATA record.
	 */

	time(&new->login);
	new->stamp = new->login;
	mt_stamp(&new->lastsend);

	/*
	 * Add the new CLDATA record to the client database.
	 */

	if((rec = cldb_add(new)))
	{
		/*
		 * Send final CONN_OK status.
		 */

		status = CONN_OK;

		if(vio_write(new->fd, &status, 1, CONN_TIMEOUT) == 1)
		{
			/* OK */

			cmd_signon(rec);

			return 0;
		}
		else
		{
			vlog(VLOG_ERR, "Failed to send final OK status (%s)",
				strerror(errno));

			if(cldb_delete(rec))
			{
				vlog(VLOG_ERR, "Failed to delete client database record (%s)",
					strerror(errno));
			}
		}
	}
	else
	{
		vlog(VLOG_ERR, "Client database failure (%s)", strerror(errno));
		status = CONN_ENOMEM;
		vio_write(new->fd, &status, 1, 1);
	}

	return -1;
}

/*
 * Send a VChat protocol text message of type SERV_MSG to a client.
 * The text message supplied by the format string <fmt> and it's arguments
 * is formatted into a buffer and then sent as a SERV_MSG type message to the
 * client that is identified by the CLDATA pointed to by <cp>. If an error
 * condition arises whilst sending the data, the error flag within <cp> will
 * be raised, and the error field of <cp> is set to indicate the type of error.
 */

void vcp_msg(CLDATA *cp, char *fmt, ...)
{
	va_list xp;

#if HAVE_SYS_UIO_H
	char sndbuf[SVMSGBUF];
	struct iovec iov[2];
	VMSG svmsg;
#else
	char sndbuf[SVMSGBUF + sizeof(VMSG)];
	VMSG *svmsg = (VMSG *)&sndbuf[0];
	char *datbuf = &sndbuf[sizeof(VMSG)];
	size_t len;
#endif

	va_start(xp, fmt);

	if(cp->flags & CFLAG_ERROR)
	{
		va_end(xp);
		return;
	}

#if HAVE_SYS_UIO_H
	
	iov[0].iov_base = &svmsg;
	iov[0].iov_len = sizeof(VMSG);
	iov[1].iov_base = &sndbuf[0];
	iov[1].iov_len = vsnprintf(sndbuf, sizeof(sndbuf), fmt, xp);

	svmsg.cmd = htonl(SERV_MSG);
	svmsg.len = htonl(iov[1].iov_len);

	if(vio_writev(cp->fd, &iov[0], 2, CONN_TIMEOUT) == -1)
#else
	len = vsnprintf(datbuf, SVMSGBUF, fmt, xp);

	svmsg->cmd = htonl(SERV_MSG);
	svmsg->len = htonl(len);

	len += sizeof(VMSG);

	if(vio_write(cp->fd, &sndbuf[0], len, CONN_TIMEOUT) != len)
#endif
	{
		cp->flags |= CFLAG_ERROR;
		cp->error = errno;
	}

	va_end(xp);
}

/*
 * Send a VChat protocol text message of type SERV_MSG to all clients matching
 * a particular nickname. The text message supplied by the format string <fmt>
 * and it's arguments is formatted into a buffer and then sent as a SERV_MSG
 * type message to all currently connected clients whose nickname matches the
 * given <nick>. If an error condition arises whilst sending the data, the
 * error flag within the CLDATA record of the corresponding client will be
 * raised, and the error field is set to indicate the type of error. The
 * return value of this function is the number of clients the message has been
 * delivered to.
 */

uint32_t vcp_msgnick(char *nick, char *fmt, ...)
{
	va_list xp;
	uint32_t records = cldb_records();
	register CLDATA *cp;
	register uint32_t n = 0;

	va_start(xp, fmt);

	if(records && (cp = cldb_addr(0)))
	{
		register int i;
#if HAVE_SYS_UIO_H
		char sndbuf[SVMSGBUF];
		struct iovec iov[2];
		VMSG svmsg;

		iov[0].iov_base = &svmsg;
		iov[0].iov_len = sizeof(VMSG);
		iov[1].iov_base = &sndbuf[0];
		iov[1].iov_len = vsnprintf(sndbuf, sizeof(sndbuf), fmt, xp);

		svmsg.cmd = htonl(SERV_MSG);
		svmsg.len = htonl(iov[1].iov_len);
#else
		char sndbuf[SVMSGBUF + sizeof(VMSG)];
		VMSG *svmsg = (VMSG *)&sndbuf[0];
		char *datbuf = &sndbuf[sizeof(VMSG)];
		size_t len;

		len = vsnprintf(datbuf, SVMSGBUF, fmt, xp);

		svmsg->cmd = htonl(SERV_MSG);
		svmsg->len = htonl(len);

		len += sizeof(VMSG);
#endif
		for(i = 0; i < records; i++, cp++)
		{
			if(strcmp(cp->nick, nick))
				continue;

			n++;	/* Found one */

#if HAVE_SYS_UIO_H
			if(vio_writev(cp->fd, &iov[0], 2, CONN_TIMEOUT) == -1)
#else
			if(vio_write(cp->fd, &sndbuf[0], len, CONN_TIMEOUT) != len)
#endif
			{
				cp->flags |= CFLAG_ERROR;
				cp->error = errno;
			}
		}
	}

	va_end(xp);

	return n;
}

/*
 * Send a VChat protocol text message of type SERV_MSG to all clients on the
 * conference channel specified by <channel>. The text message supplied by the
 * format string <fmt> and it's arguments is formatted into a buffer and then
 * sent as a SERV_MSG type message to all clients that are on <channel>. If an
 * error condition arises whilst sending the data, the error flag within the
 * CLDATA record of the corresponding client will be raised, and the error
 * field is set to indicate the type of error. The return value of this
 * function is the number of clients the message has been delivered to.
 */

uint32_t vcp_msgchan(int32_t channel, CLDATA *notme, char *fmt, ...)
{
	va_list xp;
	uint32_t records = cldb_records();
	register CLDATA *cp;
	register uint32_t n = 0;

	va_start(xp, fmt);

	if(records && (cp = cldb_addr(0)))
	{
		register int i;
#if HAVE_SYS_UIO_H
		char sndbuf[SVMSGBUF];
		struct iovec iov[2];
		VMSG svmsg;

		iov[0].iov_base = &svmsg;
		iov[0].iov_len = sizeof(VMSG);
		iov[1].iov_base = &sndbuf[0];
		iov[1].iov_len = vsnprintf(sndbuf, sizeof(sndbuf), fmt, xp);

		svmsg.cmd = htonl(SERV_MSG);
		svmsg.len = htonl(iov[1].iov_len);
#else
		char sndbuf[SVMSGBUF + sizeof(VMSG)];
		VMSG *svmsg = (VMSG *)&sndbuf[0];
		char *datbuf = &sndbuf[sizeof(VMSG)];
		size_t len;

		len = vsnprintf(datbuf, SVMSGBUF, fmt, xp);

		svmsg->cmd = htonl(SERV_MSG);
		svmsg->len = htonl(len);

		len += sizeof(VMSG);
#endif
		for(i = 0; i < records; i++, cp++)
		{
			if(cp->channel != channel || (notme && cp == notme))
				continue;

			n++;	/* Found one */

#if HAVE_SYS_UIO_H
			if(vio_writev(cp->fd, &iov[0], 2, CONN_TIMEOUT) == -1)
#else
			if(vio_write(cp->fd, &sndbuf[0], len, CONN_TIMEOUT) != len)
#endif
			{
				cp->flags |= CFLAG_ERROR;
				cp->error = errno;
			}
		}
	}

	va_end(xp);

	return n;
}

/*
 * Send a VChat protocol text message of type SERV_MSG to all clients.
 * The user who is sending the broadcast message is identified by <from>.
 * If the message is sent from a terminal via vchatctl(8), the terminal
 * line the message originates from should be supplied in <tty>, otherwise
 * <tty> should be specified as NULL. The broadcast message supplied by the
 * format string <fmt> and it's arguments is formatted into a buffer and then
 * sent as a SERV_MSG type message to all currently connected clients. If an
 * error condition arises whilst sending the data, the error flag within the
 * CLDATA record of the corresponding client will be raised, and the error
 * field is set to indicate the type of error. The return value of this
 * function is the number of clients the message has been delivered to.
 */

uint32_t vcp_broadcast(char *from, char *tty, char *fmt, ...)
{
	va_list xp;
	uint32_t records = cldb_records();
	register CLDATA *cp;

	va_start(xp, fmt);

	if(records && (cp = cldb_addr(0)))
	{
		register int i;
		char *s;
		size_t len;
		size_t maxlen = SVMSGBUF;
		char dtbuf[DT_SIZE];
		time_t now;
#if HAVE_SYS_UIO_H
		char sndbuf[SVMSGBUF];
		struct iovec iov[2];
		VMSG svmsg;

		iov[0].iov_base = &svmsg;
		iov[0].iov_len = sizeof(VMSG);
		iov[1].iov_base = s = &sndbuf[0];

		svmsg.cmd = htonl(SERV_MSG);
#else
		char sndbuf[SVMSGBUF + sizeof(VMSG)];
		VMSG *svmsg = (VMSG *)&sndbuf[0];
		s = &sndbuf[sizeof(VMSG)];

		svmsg->cmd = htonl(SERV_MSG);
#endif
		/*
		 * Format broadcast message.
		 */

		len = snprintf(s, maxlen, "* Broadcast message from ");
		maxlen -= len;
		s += len;

		if(tty)
		{
			char host[128];

			if(gethostname(&host[0], sizeof(host)))
				snprintf(host, sizeof(host), "<unknown>");

			len = snprintf(s, maxlen, "%s%s%s%s@%s%s%s%s on %s%s%s:\n%s",
				S_FG_MAG, from, S_OFF,
				S_BOLD, S_OFF,
				S_FG_GRN, host, S_OFF,
				S_FG_CYN, tty, S_OFF,
				S_BOLD);
		}
		else
		{
			len = snprintf(s, maxlen, "%s%s%s:\n%s",
				S_FG_YEL, from, S_OFF, S_BOLD);
		}

		maxlen -= len;
		s += len;

		len = vsnprintf(s, maxlen, fmt, xp);

		maxlen -= len;
		s += len;

		time(&now);
		strftime(dtbuf, sizeof(dtbuf), DT_FMT, localtime(&now));
		dtbuf[sizeof(dtbuf)-1] = 0;

		len = snprintf(s, maxlen, "%s\n* EOT - %s", S_OFF, dtbuf);

		maxlen -= len;
		len = SVMSGBUF - maxlen;

#if HAVE_SYS_UIO_H
		iov[1].iov_len = len;
		svmsg.len = htonl(len);
#else
		svmsg->len = htonl(len);
		len += sizeof(VMSG);
#endif
		/*
		 * Send broadcast message.
		 */

		for(i = 0; i < records; i++, cp++)
		{
#if HAVE_SYS_UIO_H
			if(vio_writev(cp->fd, &iov[0], 2, CONN_TIMEOUT) == -1)
#else
			if(vio_write(cp->fd, &sndbuf[0], len, CONN_TIMEOUT) != len)
#endif
			{
				cp->flags |= CFLAG_ERROR;
				cp->error = errno;
			}
		}
	}

	va_end(xp);

	return records;
}

/*
 * Send a VChat protocol message of arbitrary message type to a client.
 * A message of the type indicated by <packtype> with accompanying <data> of
 * <len> bytes is sent to the client indicated by the CLDATA pointed to by
 * <cp>. If an error condition arises whilst sending the message to the client,
 * the error flag within <cp> is raised and the error field of <cp> is set
 * to indicate the type of error.
 */

void vcp_snd(CLDATA *cp, uint32_t packtype, void *data, size_t len)
{
#if HAVE_SYS_UIO_H
	struct iovec	iov[2];
	VMSG		svmsg;

	if(cp->flags & CFLAG_ERROR)
		return;

	svmsg.cmd = htonl(packtype);
	svmsg.len = (data && len > 0) ? htonl(len) : 0;

	iov[0].iov_base = &svmsg;
	iov[0].iov_len = sizeof(VMSG);
	iov[1].iov_base = data;
	iov[1].iov_len = len;

	if(vio_writev(cp->fd, &iov[0], ((data && len > 0) ? 2 : 1), CONN_TIMEOUT) == -1)
	{
		cp->flags |= CFLAG_ERROR;
		cp->error = errno;
	}

#else	/* !HAVE_SYS_UIO_H */

	size_t	bufsize = (sizeof(VMSG) + ((data && len > 0) ? len : 0));
	uint8_t	sndbuf[bufsize];
	VMSG	*svmsg = (VMSG *)&sndbuf;
	void	*dbuf = &sndbuf[sizeof(VMSG)];

	if(cp->flags & CFLAG_ERROR)
		return;

	svmsg->cmd = htonl(packtype);
	svmsg->len = (data && len > 0) ? htonl(len) : 0;

	if(data && len > 0)
		bcopy(data, dbuf, len);

	if(vio_write(cp->fd, &sndbuf[0], bufsize, CONN_TIMEOUT) != bufsize)
	{
		cp->flags |= CFLAG_ERROR;
		cp->error = errno;
	}

#endif	/* HAVE_SYS_UIO_H */
}
