/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: vio.c,v 1.9 2009/08/14 11:54:52 mickey Exp $ */

#include "vserver.h"

#ifndef WIN32	/* ----------------------- UN*X ---------------------------- */

#if HAVE_SIGNAL_H
  #include <signal.h>
#endif

#if !HAVE_SIGNALS
  #error "Signals are not supported, but necessary for vio_*() to work"
#endif
#if !HAVE_SETJMP
  #error "Your system does not support setjmp(3) which is needed"
#endif

/*
 * Set up a timer that delivers SIGALRM when it expires.
 * When <sec> is specified as 0, the timer is disabled.
 */

void vio_settimeout(time_t sec)
{
#if HAVE_SETITIMER
	struct itimerval val;

	val.it_value.tv_sec = sec;
	val.it_value.tv_usec = 0;
	val.it_interval.tv_sec = 0;
	val.it_interval.tv_usec = 0;

	setitimer(ITIMER_REAL, &val, NULL);
#elif HAVE_ALARM
	alarm(sec);
#else
  #error "Your system neither supports setitimer(2) nor alarm(3)"
#endif
	vio_tovalid = sec ? 1 : 0;
}

/*
 * Attempt to write <nbytes> of data from the buffer pointed to by <buf>
 * to the socket descriptor <fd>. If the data cannot be written within
 * <tout> seconds, this function returns -1 with errno set to ETIMEDOUT.
 */

ssize_t vio_write(int fd, void *buf, size_t nbytes, time_t tout)
{
	if(!setjmp(vio_jmpbuf))
	{
		ssize_t ret;
		ssize_t sent = 0;

		vio_settimeout(tout);

		while(sent < nbytes)
		{
			if((ret = write(fd, buf, (nbytes - sent))) >= 0)
			{
				buf = (uint8_t *)buf + ret;
				sent += ret;
			}
			else
			{
				vio_settimeout(0);
				return ret;
			}
		}

		vio_settimeout(0);

		return sent;
	}
	else
	{
		vio_tovalid = 0;
		errno = ETIMEDOUT;
		return -1;
	}
}

#if HAVE_SYS_UIO_H

/*
 * Attempt to write the contents of the buffers pointed to by the iovec array
 * <iov> whose size is determined by <iovcnt> to the socket descriptor <fd>.
 * If the write operation does not finish within <tout> seconds this
 * function returns -1 with errno set to ETIMEDOUT.
 */

ssize_t vio_writev(int fd, struct iovec *iov, int iovcnt, time_t tout)
{
	if(!setjmp(vio_jmpbuf))
	{
		ssize_t x;

		vio_settimeout(tout);

		x = writev(fd, iov, iovcnt);

		vio_settimeout(0);

		return (x > 0 ? x : -1);
	}
	else
	{
		vio_tovalid = 0;
		errno = ETIMEDOUT;
		return -1;
	}
}

#endif /* HAVE_SYS_UIO_H */

/*
 * Attempt to read <nbytes> from socket descriptor <fd> into the buffer
 * pointed to by <buf>. If the corresponding number of bytes cannot be
 * read withing <tout> seconds, the function returns -1 with errno set
 * to ETIMEDOUT.
 */

ssize_t vio_read(int fd, void *buf, size_t nbytes, time_t tout)
{
	if(!setjmp(vio_jmpbuf))
	{
		ssize_t ret;
		ssize_t got = 0;

		vio_settimeout(tout);

		while(got < nbytes)
		{
			switch((ret = read(fd, buf, (nbytes - got))))
			{
				case 0:		errno = EPIPE;
				case -1:	vio_settimeout(0);
						return -1;

				default:	got += ret;
						buf = (uint8_t *)buf + ret;
						break;
			}
		}

		vio_settimeout(0);

		return got;
	}
	else
	{
		vio_tovalid = 0;
		errno = ETIMEDOUT;
		return -1;
	}
}



#else /* ------------------------------ WIN32 ------------------------------ */


/*************************************
 * WINDOWS 32 low level I/O routines *
 *************************************/

DWORD WINAPI WriteOp(LPVOID arg)
{
	IO_OP	*op	= (IO_OP *)arg;
	DWORD	sent	= 0;
	void	*buf	= op->buf;
	int	ret;

	while(sent < op->len)
	{
		if((ret = send(op->fd, buf, (op->len - sent), 0)) != SOCKET_ERROR)
		{
			buf = (uint8_t *)buf + ret;
			sent += ret;
		}
		else
		{
			ExitThread(-1);
		}
	}

	return sent;
}

ssize_t vio_write(int fd, void *buf, size_t nbytes, time_t tout)
{
	IO_OP	op;
	HANDLE	hOutputThread		= NULL;
	DWORD	dwOutputThreadID	= 0;
	DWORD	dwRetCode		= -1;

	/*
	 * Create a Thread that will perform the requested output operation.
	 */

	op.buf = buf;
	op.len = nbytes;
	op.fd = fd;

	if((hOutputThread = CreateThread(NULL, 0, WriteOp, &op, 0, &dwOutputThreadID)) == NULL)
	{
		vlog(VLOG_ERR, "Failed to create input operation thread.");
		return -1;
	}

	/*
	 * Wait for the timeout period for the output operation to
	 * complete by itself, if the timeout expires, then abort
	 * the output operation in progress and return error status.
	 */

	switch(WaitForSingleObject(hOutputThread, (tout * 1000)))
	{
		case WAIT_OBJECT_0:	/* The Thread has exited by itself */
					break;
		case WAIT_TIMEOUT:	/* The timeout period expired */
					TerminateThread(hOutputThread, -1);
					break;
	}

	GetExitCodeThread(hOutputThread, &dwRetCode);

	return dwRetCode;
}

DWORD WINAPI ReadOp(LPVOID arg)
{
	IO_OP	*op	= (IO_OP *)arg;
	DWORD	got	= 0;
	void	*buf	= op->buf;
	int	ret;

	/*
	 * Perform the input operation described by the IO_OP.
	 */

	while(got < op->len)
	{
		switch(ret = recv(op->fd, op->buf, (op->len - got), 0))
		{
			case SOCKET_ERROR:	ExitThread(-1);
			case 0:			ExitThread(-1);
			default:		got += ret;
						buf = (uint8_t *)buf + ret;
						break;
		}
	}

	return got;
}
  
ssize_t vio_read(int fd, void *buf, size_t nbytes, time_t tout)
{
	HANDLE	hInputThread	= NULL;
	DWORD	dwInputThreadID	= 0;
	DWORD	dwRetCode	= -1;
	IO_OP	op;

	/*
	 * Create a Thread that will perform the requested input operation.
	 */

	op.buf = buf;
	op.len = nbytes;
	op.fd = fd;
	
	if((hInputThread = CreateThread(NULL, 0, ReadOp, &op, 0, &dwInputThreadID)) == NULL)
	{
		vlog(VLOG_ERR, "Failed to create input operation thread.");
		return -1;
	}

	/*
	 * Wait for the timeout period for the input operation to
	 * complete by itself, if the timeout expires, then abort
	 * the input operation in progress and return error status.
	 */

	switch(WaitForSingleObject(hInputThread, (tout * 1000)))
	{
		case WAIT_OBJECT_0:	/* The Thread has exited by itself */
					break;
		case WAIT_TIMEOUT:	/* The timeout period expired */
					TerminateThread(hInputThread, -1);
					break;
	}

	GetExitCodeThread(hInputThread, &dwRetCode);

	return dwRetCode;
}

#endif /* WIN32 */
