/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: vlog.c,v 1.13 2009/08/14 11:54:52 mickey Exp $ */

#include "vserver.h"

/*
 * Prepare for logging.
 */

void vlog_open(void)
{
#if HAVE_SYSLOG
	if(cfdata.flags & CF_LOG_SYSLOG)
	{
		openlog(prog_name, LOG_NDELAY, LOG_LOCAL1);
	}
#endif
	if(cfdata.flags & CF_LOG_FILE)
	{
		if((lgf = fopen(cfdata.logfile, "a")) == NULL)
		{
			fprintf(stderr, "%s [WARNING] Failed to open log file '%s' for output (%s)\n",
				prog_name, cfdata.logfile, strerror(errno));
			fprintf(stderr, "%s [WARNING] Logging to file disabled\n\n", prog_name);
			cfdata.flags &= ~CF_LOG_FILE;
		}
		else
		{
#if SETVBUF_REVERSED
			setvbuf(lgf, _IONBF, NULL, 0);
#else
			setvbuf(lgf, NULL, _IONBF, 0);
#endif

		}
	}

	vlog(VLOG_INFO, "Log started");
}

/*
 * Finish using vserver logging.
 */

void vlog_close(void)
{
	vlog(VLOG_INFO, "Log closed");

#if HAVE_SYSLOG
	if(cfdata.flags & CF_LOG_SYSLOG)
	{
		closelog();
	}
#endif
	if(cfdata.flags & CF_LOG_FILE)
	{
		if(lgf)
		{
			fclose(lgf);
			lgf = NULL;
		}
	}
}

/*
 * Write the message specified by <fmt> and it's arguments to the server log.
 */

void vlog(int pri, char *fmt, ...)
{
	char	logmsg[1024];
	char	*prefix;
	va_list ap;

	va_start(ap, fmt);

	/*
	 * Select logging prefix according to log priority.
	 */

	switch(pri)
	{
		case VLOG_EMERG:	prefix="[EMERGENCY] ";
					break;
		case VLOG_ALERT:	prefix="[ALERT] ";
					break;
		case VLOG_CRIT:		prefix="[CRITICAL] ";
					break;
		case VLOG_ERR:		prefix="[ERROR] ";
					break;
		case VLOG_WARNING:	prefix="[WARNING] ";
					break;
		default:		prefix="";
					break;
	}

	/*
	 * Format the log message.
	 */

	vsnprintf(logmsg, sizeof(logmsg), fmt, ap);

	/*
	 * Log to syslog when enabled.
	 */

#if HAVE_SYSLOG
	if(cfdata.flags & CF_LOG_SYSLOG)
	{
		syslog(pri, "%s%s", prefix, logmsg);
	}
#endif
	/*
	 * Log to file when enabled.
	 */

	if(cfdata.flags & CF_LOG_FILE)
	{
		char	dtbuf[DT_SIZE];
		time_t	now;

		time(&now);
		strftime(dtbuf, sizeof(dtbuf), DT_LOGFMT, localtime(&now));
		dtbuf[sizeof(dtbuf)-1] = 0;

		fprintf(lgf, "%s %s%s\n", dtbuf, prefix, logmsg);
	}

	/*
	 * Also write urgent messages to stderr.
	 */

	if(pri == VLOG_CRIT || pri == VLOG_ALERT || pri == VLOG_EMERG)
	{
		fprintf(stderr, "%s %s%s\n", prog_name, prefix, logmsg);
	}

	va_end(ap);
}

/*
 * Rotate the logfile upon receipt of a SIGUSR1 signal.
 */

void rotate_logfile(void)
{
	register int	i;
	char		tmp1[PATH_MAX];
	char		tmp2[PATH_MAX];
	struct stat	st;

	/*
	 * Do nothing if logging to file is disabled.
	 */

	if(!(cfdata.flags & CF_LOG_FILE) || lgf == NULL)
		return;

	/*
	 * Close logfile
	 */

	fclose(lgf);

	/*
	 * Shift logfile generations
	 */

	for(i = cfdata.rotate_log ; i >= 0 ; --i)
	{
		snprintf(tmp1, sizeof(tmp1), "%s;%d", cfdata.logfile, i);

		if(stat(tmp1, &st) == -1)
			continue;

		if(i == cfdata.rotate_log)
		{
			/*
			 * This file has hit the limit!
			 * -delete- it!
			 */

			unlink(tmp1);
			continue;
		}

		/*
		 * Shift file up one number
		 */

		snprintf(tmp2, sizeof(tmp2), "%s;%d", cfdata.logfile, i+1);

		rename(tmp1, tmp2);
	}

	/*
	 * Rename current logfile to logfile.0
	 */

	snprintf(tmp2, sizeof(tmp2), "%s;0", cfdata.logfile);
	rename(cfdata.logfile, tmp2);

	/*
	 * Open new logfile
	 */

	if((lgf = fopen(cfdata.logfile, "a")) == NULL)
	{
		fprintf(stderr, "%s: Failed to open logfile '%s' for output (%s)\n",  
			prog_name, cfdata.logfile, strerror(errno));
		fprintf(stderr, "%s: Logging to file disabled due to error\n\n", prog_name);
			cfdata.flags &= ~CF_LOG_FILE;
	}
	else
	{
#if SETVBUF_REVERSED
		setvbuf(lgf, _IONBF, NULL, 0);
#else
		setvbuf(lgf, NULL, _IONBF, 0);
#endif
		vlog(VLOG_INFO, "New logfile started");
	}
}
