/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: vserver.c,v 1.22 2009/08/14 11:54:52 mickey Exp $ */

#include "vserver.h"
#include "vctlproto.h"

#if HAVE_SYS_SELECT_H
  #include <sys/select.h>
#endif

#ifdef WIN32
  #include <io.h>
  #include <tchar.h>
#endif

/*
 * The VChat conference server main entry point.
 */

int main(int argc, char *argv[])
{
#ifdef WIN32
	SERVICE_TABLE_ENTRY sdpt[2];
#endif
	int ret;

	/*
	 * Determine program name.
	 */

	if((prog_name = strrchr(argv[0], DIRSEP)) != NULL)
		prog_name++;
	else
		prog_name = argv[0];

#ifdef WIN32
	upperstr(prog_name);
#endif
	/*
	 * Setup configuration base defaults.
	 */

	mkdefaults();

	/*
	 * Process commandline arguments.
	 */

	if(process_cmdline(argc, argv))
		return -1;

	/*
	 * If CF_CREATE_UDB is set in the flags, then
	 * create a new user database file and exit.
	 */

	if(cfdata.flags & CF_CREATE_UDB)
		return userdb_create();

#ifdef WIN32
	/*
	 * If CF_INSTALL is set in the flags, then
	 * install the service and exit.
	 */

	if(cfdata.flags & CF_INSTALL)
	{
		ret = InstallService();

		fprintf(stderr, ret ? "Failed to install service '%s'\n" :
			"Service '%s' installed sucessfully\n",
			cfdata.svc_name);

		/*
		 * If there was a problem installing the service
		 * then delete the leftovers.
		 */

		if(ret)
			UninstallService();

		return ret;
	}

	/*
	 * If CF_UNINSTALL is set in the flags, then
	 * uninstall the service and exit.
	 */

	if(cfdata.flags & CF_UNINSTALL)
	{
		ret = UninstallService();

		fprintf(stderr, ret ? "Failed to uninstall service '%s'\n" :
			"Service '%s' uninstalled successfully\n",
			cfdata.svc_name);

		return ret;
	}
#endif

#if HAVE_DAEMON
	/*
	 * Detach from controlling terminal if running as daemon
	 */

	if(cfdata.flags & CF_DETACH)
	{
		if(daemon(1, 1) == -1)
		{
			fprintf(stderr, "%s: error detaching from controlling terminal (%s)\n",
				prog_name, strerror(errno));
		}
	}
#endif

#ifdef WIN32
	/*
	 * Initialize the service dispatch table with the servic name
	 * that we will be using and the service main function. Then
	 * register it with the ServiceControlManager.
	 */

	sdpt[0].lpServiceName = TEXT(cfdata.svc_name);
	sdpt[0].lpServiceProc = (LPSERVICE_MAIN_FUNCTION)ServiceMain;
	sdpt[1].lpServiceName = NULL;
	sdpt[1].lpServiceProc = NULL;

	if(!StartServiceCtrlDispatcher(sdpt))
		LogMessage(VLOG_ERR, "StartServiceCtrlDispatcher failed.");

	/*
	 * We're done here. Service control manager will call
	 * the ServiceMain() function in order to start the
	 * service.
	 */

	return 0;

#else /* !WIN32 */

	/*
	 * Perform global initialization.
	 */

	if(global_init())
		return -1;

	/*
	 * Run the service.
	 */

	ret = Service();

	/*
	 * Perform global deinitialization
	 * and return service exit status.
	 */

	global_exit();

	return ret;
#endif
}

/*
 * The function that performs service operation.
 */

int Service(void)
{
	register CLDATA	*cp;
	register int	i;

	uint8_t		buffer[SVMSGBUF];
	VMSG		svmsg;
	SVCMD		*pcmd;
	fd_set		rfdset;
	uint32_t	records;
	uint32_t	cmd;
	size_t		len;
	time_t		tick = 0;
	struct timeval	t_start;
	struct timeval	t_accept;
	struct timeval	*elapsed;
	struct timeval	sTimeout;
	int		sel;
	int		ret = 0;

#ifdef WIN32
	/*
	 * The service is now running.
	 */

	ssStatus.dwWaitHint = 0;
	ssStatus.dwCheckPoint = 0;
	ssStatus.dwControlsAccepted |= SERVICE_ACCEPT_STOP;
	ssStatus.dwControlsAccepted |= SERVICE_ACCEPT_SHUTDOWN;
	ssStatus.dwCurrentState = SERVICE_RUNNING;

	SetServiceStatus(sshStatusHandle, &ssStatus);

	LogMessage(VLOG_INFO, "Starting normal server operation");
#endif
	/*
	 * Server main-loop
	 */

	while(!(cfdata.flags & CF_SHUTDOWN))
	{
		FD_ZERO(&rfdset);
#if HAVE_UNIX_SOCKETS
		FD_SET(ctlsock, &rfdset);
#endif
		/*
		 * Get the number of clients we'll be serving this run.
		 */

		records = cldb_records();

		/*
		 * With flood protection enabled, we will only accept
		 * new client connections on the server socket, if the
		 * time that has elapsed since we did the last accept(2)
		 * is greater than the flood protect time.
		 */

		if(vsopt.flags & VO_FLOODGUARD)
		{
			elapsed = mt_elapsed_now(&t_accept);

			if(timercmp(elapsed, &vsopt.t_guard, >))
			{
				FD_SET(sock, &rfdset);
			}
		}
		else
		{
			FD_SET(sock, &rfdset);
		}

		/*
		 * Put client fds into descriptor set.
		 */

		if(records && (cp = cldb_addr(0)))
		{
			for(i = 0; i < records; i++, cp++)
			{
				/*
				 * If flood protection is enabled and the client
				 * is not registered, select the client only if
				 * the time that has elapsed since the client last
				 * sent data is greater than the flood protect time.
				 */

				if((vsopt.flags & VO_FLOODGUARD)
				&& !(cp->flags & CFLAG_REGISTERED))
				{
					elapsed = mt_elapsed_now(&(cp->lastsend));

					if(timercmp(elapsed, &vsopt.t_guard, >))
					{
						FD_SET(cp->fd, &rfdset);
					}
				}
				else
				{
					FD_SET(cp->fd, &rfdset);
				}
			}
		}

		/*
		 * The select timeout needs to be set again during each loop
		 * since there are stupid UN*Xes believing they should have
		 * changed the timeout value on return of select(2) *shrug*
		 */

		sTimeout.tv_sec = 1;
		sTimeout.tv_usec = 0;

		/*
		 * Wait until something happens on our descriptors.
		 */

		sel = select(FD_SETSIZE, &rfdset, (fd_set *)NULL, (fd_set *)NULL, &sTimeout);

#ifdef WIN32
		if(sel == SOCKET_ERROR)
		{
			int errcod = WSAGetLastError();

			vlog(VLOG_CRIT, "select() returned status %d (%s) - server exiting",
					errcod, strerror(errcod));

			ret = -1; /* Return ERROR status */

			break;
		}
#else
		if(sel < 0)
		{
			if(errno == EINTR)
			{
				continue;
			}
			else
			{
				vlog(VLOG_CRIT, "select() failed (%s) "
					"- server exiting",
					strerror(errno));

				ret = -1; /* Return ERROR status */

				break;
			}
		}
#endif
		else if(sel == 0) /* Select timeout has expired */
		{
			/*
			 * Increment tick counter.
			 */

			tick++;

			/*
			 * Delete ACL rules that have expired.
			 */

			acldb_delete_expired();

			/*
			 * Perform reorganization of RMS databases
			 * as necessary every 60 seconds.
			 */

			if(!(tick % RMS_REORG_INTERVAL))
			{
				if(rms_reorg())
				{
					vlog(VLOG_CRIT, "rms_reorg() failed (%s)",
						strerror(errno));

					ret = -1; /* Return ERROR status */

					break;
				}
			}

			continue;
		}

		/*
		 * Check for new data from our clients and process it.
		 */

		if(records && (cp = cldb_addr(0)))
		{
			for(i = 0; i < records; i++, cp++)
			{
				/*
				 * Check for data waiting.
				 */

				if(!FD_ISSET(cp->fd, &rfdset))
					continue;

				/*
				 * Get command packet.
				 */

				if(vio_read(cp->fd, &svmsg, sizeof(VMSG), CONN_TIMEOUT) == -1)
				{
					cp->flags |= CFLAG_ERROR;
					cp->error = errno;
					continue;
				}

				/*
				 * Extract command and length with correct
				 * byteorder.
				 */

				cmd = ntohl(svmsg.cmd);
				len = ntohl(svmsg.len);

				/*
				 * Apply buffer size limit if necessary.
				 */

				if(len >= SVMSGBUF)
					len = SVMSGBUF - 1;

				/*
				 * Get appended data packet if any.
				 */
				
				if(len > 0 && vio_read(cp->fd, &buffer[0],
					len, CONN_TIMEOUT) == -1)
				{
					cp->flags |= CFLAG_ERROR;
					cp->error = errno;
					continue;
				}

				/*
				 * Ascertain that the buffer is NUL terminated.
				 */

				buffer[len] = 0;

				/*
				 * Get the command descriptor that is
				 * associated with the specified command.
				 */

				pcmd = &svcmd[cmdidx(cmd)];

				/*
				 * Increment request counters.
				 */

				vs.vs_requests_total++;
				pcmd->requests++;

				/*
				 * Check if the command is implemented.
				 */

				if(pcmd->func)
				{
					/*
					 * Update timestamp of client's last
					 * activity, unless when measuring
					 * round trip delays via RTCMP.
					 */

					if(cmd != CMD_RTCMP)
						mt_stamp(&(cp->lastsend));

					/*
					 * Check if the client is permitted
					 * to execute the given command.
					 */

					if(!pcmd->perm
					|| (cp->accperm & pcmd->perm)
					== pcmd->perm)
					{
						/*
						 * Execute the command and
						 * measure execution time.
						 */
				
						mt_stamp(&t_start);
						pcmd->func(cp, buffer, len);
						elapsed = mt_elapsed_now(&t_start);

						/*
						 * Store command execution
						 * time for statistical
						 * processing.
						 */

						mt_store(elapsed);
					}
					else
					{
						vcp_msg(cp, "* %s%s%s security clearance required",
							S_FG_RED,
							perm_str(pcmd->perm),
							S_OFF);
					}
				}
				else
				{
					vcp_msg(cp, "* Command $%08X not implemented", cmd);
					vlog(VLOG_WARNING, "WARNING: Unknown client command CMD: $%08x USER: %s NICK: %s FROM: %s",
						cmd, cp->user, cp->nick,
						cp->host);
				}
			}
		}

#if HAVE_UNIX_SOCKETS
		/*
		 * Process UDP control messages received
		 * on the server control socket.
		 */

		if(FD_ISSET(ctlsock, &rfdset))
		{
			vchatctl_rcvcmd();
		}
#endif
		/*
		 * Signoff clients that have one of the SIGNOFF,
		 * KILL, EXCLUDE or ERROR flags set.
		 */

		if(records && (cp = cldb_addr(records - 1)))
		{
			for(i = (records - 1); i >= 0; i--, cp--)
			{
				if((cp->flags & (CFLAG_SIGNOFF | CFLAG_KILL | CFLAG_EXCLUDE)))
				{
					signoff(cp);
				}
				else if(cp->flags & CFLAG_ERROR)
				{
					vlog(VLOG_ERR, "Error condition on client descriptor (%s) DESC: %d USER: %s NICK: %s HOST: %s",
						strerror(cp->error), cp->fd,
						cp->user, cp->nick,
						cp->host);

					++vs.vs_err_conn;
					signoff(cp);
				}
			}
		}

		/*
		 * If there's data available on the server socket
		 * then a new client wants to connect.
		 */

		if(FD_ISSET(sock, &rfdset))
		{
			mt_stamp(&t_accept);
			vcp_setupconn();
		}

	} /*** END OF SERVER MAIN LOOP ***/

	vlog(VLOG_NOTICE, "Shutting down operations");

	/*
	 * Signoff all connected clients.
	 */

	records = cldb_records();

	if(records && (cp = cldb_addr(records - 1)))
	{
		for(i = (records - 1); i >= 0; i--, cp--)
		{
			cp->flags |= CFLAG_SHUTDOWN;
			signoff(cp);
		}
	}

	/*
	 * Service Exit.
	 */

	return ret;
}
