/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: vtopic.c,v 1.14 2009/08/14 11:54:52 mickey Exp $ */

#include "vserver.h"

/*
 * Open and initialize the topic database. Returns 0 if successful,
 * otherwise a value of -1 is returned.
 */

int topicdb_open(void)
{
	int ret = -1;

	if(!rms_open(RMS_TOPIC, sizeof(VTOPIC), fn_topicdb, RMS_O_CREATE, 0600))
	{
		int records = rms_records(RMS_TOPIC);

		if(records > 0)
		{
			register VTOPIC *vt = rms_addr(RMS_TOPIC, 0);

			if(vt)
			{
				register int i;

				/*
				 * Update time stamp of all channels upon startup.
				 */

				for(i = 0; i < records; i++, vt++)
					time(&vt->stamp);

				vlog(VLOG_INFO, "Topic database loaded (%d topics)", records);

				/* OK */

				ret = 0;
			}
			else
			{
				vlog(VLOG_ERR, "Failed to get address for topic record 0");
			}
		}
		else if(!records)
		{
			vlog(VLOG_INFO, "Topic database initialized");

			/* OK */

			ret = 0;
		}
		else
		{
			vlog(VLOG_ERR, "Failed to get number of topic records (%s)",
				strerror(errno));

			rms_close(RMS_TOPIC);
		}
	}
	else
	{
		vlog(VLOG_ERR, "Failed to open topic database '%s' (%s)",
			fn_topicdb, strerror(errno));
	}

	return ret;
}

/*
 * Close the topic database and free any memory possibly in use.
 * Returns 0 if successful, otherwise a value of -1 is returned.
 */

int topicdb_close(void)
{
	return rms_close(RMS_TOPIC);
}

/*
 * Save all fixed channel topics to the topic database file. Upon successful
 * operation the number of bytes written to the file is returned. Otherwise
 * a value of -1 is returned.
 */

ssize_t topicdb_save(void)
{
	ssize_t ret = -1;
	int records = rms_records(RMS_TOPIC);

	if(records >= 0)
	{
		int fd = open(fn_topicdb, (O_WRONLY | O_CREAT
			| O_TRUNC | O_BINARY | O_SEQUENTIAL), 0600);

		if(fd >= 0)
		{
			if(records > 0)
			{
				register VTOPIC *vt = rms_addr(RMS_TOPIC, 0);	

				if(vt)
				{
					register int i;

					for(i = 0, ret = 0; i < records; i++, vt++)
					{
						if(!(vt->flags & VT_FIXED))
							continue;

						if(write(fd, vt, sizeof(VTOPIC)) == sizeof(VTOPIC))
						{
							ret += sizeof(VTOPIC);
						}
						else
						{
							ret = -1;
							break;
						}
					}
				}
			}

			if(close(fd))
				ret = -1;
		}
	}

	return ret;
}

/*
 * Set up a fixed <topic> for <channel>. If a topic already exists
 * for <channel>, then it is updated with <topic> and fixed if it has
 * not already been of fixed type. Otherwise a new fixed topic will
 * be created for <channel>. Upon successful operation a value of 0 is
 * returned. Otherwise a value of -1 is returned and the global variable
 * errno is set to indicate the error.
 */

int topicdb_fix(int32_t channel, char *topic)
{
	int idx = topicdb_find(channel);
	int ret = -1;

	if(idx >= 0)			/* Update existing topic */
	{
		VTOPIC *vt = (VTOPIC *)rms_addr(RMS_TOPIC, idx);

		if(vt)
		{
			vt->flags |= VT_FIXED;
			sncopy(topic, vt->topic, sizeof(vt->topic));

			ret = (topicdb_save() >= 0 ? 0 : -1);
		}
	}
	else if(errno == ENOENT)	/* Create new fixed topic */
	{
		VTOPIC new;

		bzero(&new, sizeof(new));
		new.channel = channel;
		new.flags = VT_FIXED;
		sncopy(topic, new.topic, sizeof(new.topic));
		time(&new.stamp);

		if(rms_add_rec(RMS_TOPIC, &new))
			ret = (topicdb_save() >= 0 ? 0 : -1);
	}

	return ret;
}

/*
 * Unfix the topic for <channel>. If there are no users on <channel> then
 * the topic for <channel> will be removed from the database. Otherwise
 * the topic for <channel> remains but is converted to non-fixed type.
 * Upon successful operation a value of 0 is returned. Otherwise a value
 * of -1 is returned and the global variable errno is set to indicate the
 * error.
 *
 * The topicdb_unfix() function succeeds unless:
 *
 * ENOATTR   - The topic for <channel> is not of fixed type.
 *
 * Furthermore errno may contain any specific error code set by the
 * topicdb_save() function if permanent data storage of fixed topics failed.
 */

int topicdb_unfix(int32_t channel)
{
	int idx = topicdb_find(channel);
	int ret = -1;

	if(idx >= 0)
	{
		VTOPIC *vt = (VTOPIC *)rms_addr(RMS_TOPIC, idx);

		if(vt && (vt->flags & VT_FIXED))
		{
			if(chatters(channel) > 0)	/* -> non-fixed */
			{
				vt->flags &= ~VT_FIXED;
				ret = (topicdb_save() >= 0 ? 0 : -1);
			}
			else				/* Remove topic */
			{
				if(!rms_del_rec(RMS_TOPIC, idx))
					ret = (topicdb_save() >= 0 ? 0 : -1);
			}
		}
		else if(vt && !(vt->flags & VT_FIXED))
		{
			errno = ENOATTR;	/* Topic is not fixed */
		}
	}

	return ret;
}

/*
 * Set/Update the topic for <channel> to <topic>. Upon successful operation
 * a value of 0 is returned. Otherwise a value of -1 is returned and the
 * global variable errno is set to indicate the error.
 *
 * The topicdb_set() function succeeds unless:
 *
 * EPERM    - The topic for <channel> may not be updated, cause it is of
 *            fixed type.
 *
 * Furthermore errno may contain any specific error code set by the
 * rms_add_rec() function if it fails to add a new topic to the database.
 */

int topicdb_set(int32_t channel, char *topic)
{
	int idx = topicdb_find(channel);
	int ret = -1;

	if(idx >= 0)			/* Update existing */
	{
		VTOPIC *vt = (VTOPIC *)rms_addr(RMS_TOPIC, idx);

		if(vt)
		{
			if(!(vt->flags & VT_FIXED))
			{
				sncopy(topic, vt->topic, sizeof(vt->topic));
				time(&vt->stamp);

				/* OK */

				ret = 0;
			}
			else
			{
				errno = EPERM;
			}
		}
	}
	else if(errno == ENOENT)	/* Create new */
	{
		VTOPIC new;

		bzero(&new, sizeof(new));
		new.channel = channel;
		sncopy(topic, new.topic, sizeof(new.topic));
		time(&new.stamp);

		ret = (rms_add_rec(RMS_TOPIC, &new) ? 0 : -1);
	}

	return ret;
}

/*
 * Expire the topic for <channel> when there is only one user on that
 * channel (which is about to leave the channel) and the channel topic
 * is not of fixed type. Upon successful operation a value of 0 is
 * returned. If the topic could not be removed from the database a
 * value of -1 is returned.
 */

int topicdb_expire(int32_t channel)
{
	int idx = topicdb_find(channel);
	int ret = 0;

	if(idx >= 0)
	{
		VTOPIC *vt = (VTOPIC *)rms_addr(RMS_TOPIC, idx);

		/*
		 * Remove the channel topic, if it is not fixed
		 * and the last user is about to leave the channel.
		 */

		if(vt && !(vt->flags & VT_FIXED) && chatters(channel) == 1)
		{
			ret = rms_del_rec(RMS_TOPIC, idx);
		}
	}

	return ret;
}

/*
 * Get the number of topics currently stored within the topic database.
 * On successful completion the non-negative number of records is returned.
 * Otherwise a value of -1 is returned and the global variable errno
 * is set to indicate the error.
 */

int topicdb_records(void)
{
	return rms_records(RMS_TOPIC);
}

/*
 * Get the memory address of the topic database record that is
 * indexed by <idx>. On successful completion the address is
 * returned. Otherwise a value of NULL is returned and the global
 * variable errno is set to indicate the error.
 */

VTOPIC *topicdb_addr(int idx)
{
	return rms_addr(RMS_TOPIC, idx);
}

/*
 * Find the topic for conference channel <channel> within the topic
 * database and return the index of the topic record. If successful,
 * the non-negative index of the record is returned. Otherwise a value
 * of -1 is returned and the global variable errno is set to indicate
 * the error.
 *
 * The topicdb_find() function succeeds unless:
 *
 * ENOENT   - No topic for the specified <channel> was found in the database.
 *
 * Furthermore errno may contain any specific error code set by the
 * rms_records() function it it fails.
 */

int topicdb_find(int32_t channel)
{
	int records = rms_records(RMS_TOPIC);

	if(records > 0)
	{
		register VTOPIC	*vt = (VTOPIC *)rms_addr(RMS_TOPIC, 0);

		if(vt)
		{
			register int i;

			for(i = 0; i < records; i++, vt++)
			{
				if(vt->channel == channel)
					return i;
			}

			errno = ENOENT;
		}
	}
	else if(!records)
	{
		errno = ENOENT;
	}

	return -1;
}

/*
 * Find the topic for conference channel <channel> within the topic
 * database and return the memory address of the topic record. If
 * successful, the address of the record is returned. Otherwise a value
 * of NULL is returned and the global variable errno is set to indicate
 * the error.
 *
 * The topicdb_findaddr() function succeeds unless:
 *
 * ENOENT   - No topic for the specified <channel> was found in the database.
 *
 * Furthermore errno may contain any specific error code set by the
 * rms_records() function it it fails.
 */

VTOPIC *topicdb_findaddr(int32_t channel)
{
	int records = rms_records(RMS_TOPIC);

	if(records > 0)
	{
		register VTOPIC	*vt = (VTOPIC *)rms_addr(RMS_TOPIC, 0);

		if(vt)
		{
			register int i;

			for(i = 0; i < records; i++, vt++)
			{
				if(vt->channel == channel)
					return vt;
			}

			errno = ENOENT;
		}
	}
	else if(!records)
	{
		errno = ENOENT;
	}

	return (VTOPIC *)0;
}

/*
 * Update the channel last activity timestamp for the channel
 * specified by <channel> to the current time.
 */

void touch_channel(int32_t channel)
{
	VTOPIC *vt = topicdb_findaddr(channel);
	
	if(vt)
		time(&vt->stamp);
}
