/*****************************************************************************
 * This file is part of the VChat conference server distribution.            *
 *                                                                           *
 * Copyright (c) 1993 - 2009 Andreas S. Wetzel                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * The contents of this file are subject to the terms of the GNU General     *
 * Public License (GPL), version 2 only ("the License") as published by the  *
 * Free Software Foundation.                                                 *
 *                                                                           *
 * This Software comes with ABSOLUTELY NO WARRANTY.                          *
 * See the License for more details. You should have received a copy of the  *
 * License along with this distribution; See the file COPYING. If not, write *
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,   *
 * Boston, MA 02110-1301 USA.                                                *
 *****************************************************************************/
 
/* $Id: win32util.c,v 1.15 2009/08/14 11:54:53 mickey Exp $ */

#ifdef WIN32	/* --------> This file only valid for Windows 32 <--------- */

#include "vserver.h"

#include <io.h>
#include <tchar.h>

#define	SVC_KEY	"SYSTEM\\CurrentControlSet\\Services"

/*
 * Get configuration data from Windows Registry.
 */

int GetRegistryData(void)
{
	HKEY	key;
	char	keyname[MAX_PATH+1];
	char	vname[MAX_PATH+1];
	char	data[MAX_PATH+1];
	DWORD	vnsize	= MAX_PATH;
	DWORD	dsize	= MAX_PATH;
	DWORD	vidx	= 0;
	DWORD	vt;
	DWORD	dwRet;

	/*
	 * On Windows we may run multiple VChat servers each as a separate
	 * service in SCM and each with it's own configuration data stored
	 * in the registry. The parameters are loaded from the registry key
	 * [HLKM\System\CurrentControlSet\Services\<service-name>\Parameter]
	 */

	snprintf(keyname, sizeof(keyname), "%s\\%s\\Parameter",
		SVC_KEY, cfdata.svc_name);

	if((dwRet = RegOpenKeyEx(HKEY_LOCAL_MACHINE, keyname, 0,
		KEY_QUERY_VALUE, &key)))
	{
		LogMessage(VLOG_ERR, "Failed to open key 'HKLM\\%s' (%s)",
			keyname, winerror(dwRet));
		return -1;
	}

	LogMessage(VLOG_INFO, "Loading config '%s' ...", keyname);

	/*
	 * Step through all values for the "Parameter" key
	 * and look out for known value names.
	 */

	while(!RegEnumValue(key, vidx, vname, &vnsize, NULL, &vt, data, &dsize))
	{
		if(!strcmp(vname, "BaseDirectory") && vt == REG_SZ)
		{
			size_t len = sncopy(data, cfdata.basedir, sizeof(cfdata.basedir));

			if(len && cfdata.basedir[len-1] != DIRSEP)
			{
				snprintf(&cfdata.basedir[len],
					sizeof(cfdata.basedir) - len,
					"%c", DIRSEP);
			}
		}
		else if(!strcmp(vname, "HostName") && vt == REG_SZ)
		{
			sncopy(data, cfdata.hostname,
				sizeof(cfdata.hostname));
		}
		else if(!strcmp(vname, "BindAddress") && vt == REG_SZ)
		{
			cfdata.bind_addr.s_addr = inet_addr(data);

			if(cfdata.bind_addr.s_addr == INADDR_NONE)
			{
				LogMessage(VLOG_ERR, "invalid address '%s' "
					"reverting to INADDR_ANY", data);
				cfdata.bind_addr.s_addr = INADDR_ANY;
			}
		}
		else if(!strcmp(vname, "Port") && vt == REG_DWORD)
		{
			cfdata.port = (uint16_t)*(DWORD *)data;
		}
		else if(!strcmp(vname, "LogEnable") && vt == REG_DWORD)
		{
			if(*((DWORD *)data))
				cfdata.flags |= CF_LOG_FILE;
		}
		else if(!strcmp(vname, "LogFile") && vt == REG_SZ)
		{
			sncopy(data, cfdata.logfile, sizeof(cfdata.logfile));
		}
		else if(!strcmp(vname, "LastlogRecords") && vt == REG_DWORD)
		{
			cfdata.lastlog_recs = *(DWORD *)data;

			if(cfdata.lastlog_recs != 0
			&& cfdata.lastlog_recs < LASTLOG_MIN)
			{
				cfdata.lastlog_recs = LASTLOG_MIN;
			}
			else if(cfdata.lastlog_recs > LASTLOG_MAX)
			{
				cfdata.lastlog_recs = LASTLOG_MAX;
			}
		}
		else if(!strcmp(vname, "MaxPending") && vt == REG_DWORD)
		{
			cfdata.max_pending = *((DWORD *)data);
		}
		else if(!strcmp(vname, "RotateLog") && vt == REG_DWORD)
		{
			cfdata.rotate_log = *((DWORD *)data);
		}

		++vidx;
		vnsize = dsize = MAX_PATH;
	}

	/*
	 * Close the registry key.
	 */

	RegCloseKey(key);

	/*
	 * Relocate pathnames if the base directory has changed.
	 */

	LogMessage(VLOG_INFO, "Config loaded successfully.");

	return 0;
}

/*
 * The ServiceMain() function is called by the
 * ServiceControlManager to start up this service.
 */

VOID WINAPI ServiceMain(DWORD dwArgc, LPTSTR *lpszArgv)
{
	/*
	 * Initialize the service status structure.
	 */

	ssStatus.dwServiceType			= SERVICE_WIN32_OWN_PROCESS;
	ssStatus.dwCurrentState			= SERVICE_STOPPED;
	ssStatus.dwControlsAccepted		= 0;
	ssStatus.dwWin32ExitCode		= NO_ERROR;
	ssStatus.dwServiceSpecificExitCode	= NO_ERROR;
	ssStatus.dwCheckPoint			= 0;
	ssStatus.dwWaitHint			= 1000;

	/*
	 * Register the ServiceControlHandler() function for this service.
	 */

	if(!(sshStatusHandle = RegisterServiceCtrlHandler(
		TEXT(cfdata.svc_name), ServiceControlHandler)))
	{
		LogMessage(VLOG_ERR, "Failed to register service control handler");
		return;
	}

	/*
	 * The service start is currently pending.
	 */

	ssStatus.dwCurrentState = SERVICE_START_PENDING;
	SetServiceStatus(sshStatusHandle, &ssStatus);

	/*
	 * Load the configuration data from the registry,
	 * perform global initialization and run the service.
	 */

	if((ssStatus.dwServiceSpecificExitCode = GetRegistryData())
	|| (ssStatus.dwServiceSpecificExitCode = global_init())
	|| (ssStatus.dwServiceSpecificExitCode = Service()))
	{
		ssStatus.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
	}

	/*
	 * The service stop is currently pending.
	 */

	ssStatus.dwCurrentState = SERVICE_STOP_PENDING;
	SetServiceStatus(sshStatusHandle, &ssStatus);


	/*
	 * Perform global deinitialization.
	 */

	global_exit();

	/*
	 * The service has now stopped operating.
	 */

	ssStatus.dwWaitHint = 0;
	ssStatus.dwCheckPoint = 0;
	ssStatus.dwControlsAccepted &= ~SERVICE_ACCEPT_STOP;
	ssStatus.dwControlsAccepted &= ~SERVICE_ACCEPT_SHUTDOWN;
	ssStatus.dwCurrentState = SERVICE_STOPPED;

	SetServiceStatus(sshStatusHandle, &ssStatus);
}

/*
 * The ServiceControlHandler() function is called by the
 * ServiceControlManager to start or stop this service or
 * to get it's current status.
 */

VOID WINAPI ServiceControlHandler(DWORD dwCtlCode)
{
	/*
	 * Handle the requested control code.
	 */

	if(dwCtlCode == SERVICE_CONTROL_STOP
	|| dwCtlCode == SERVICE_CONTROL_SHUTDOWN)
	{
		/*
		 * The service stop is pending.
		 */

		ssStatus.dwCurrentState = SERVICE_STOP_PENDING;

		SetServiceStatus(sshStatusHandle, &ssStatus);

		/*
		 * Request the service to shut down.
		 * This is accomplished by setting the shutdown flag
		 * and then closing the server socket, which will cause
		 * the select() call in the server main loop to return
		 * immediately.
		 */

		cfdata.flags |= CF_SHUTDOWN;
		closesocket(sock);
		sock = -1;
	}

	/*
	 * In any case, report the current state of this service.
	 */

	SetServiceStatus(sshStatusHandle, &ssStatus);
}

/*
 * Install the service in the ServiceControlManager database.
 */

int InstallService()
{
	int		ret = -1;
	SC_HANDLE	scm;
	char		keyname[MAX_PATH+1];
	char		path[MAX_PATH+1];
	HKEY		key;
	DWORD		dwRet;
	char		*s;
	struct stat	st;

	/*
	 * Install service to SCManager database.
	 */

	fprintf(stderr, "\nInstalling service '%s' to ServiceControlManager "
		"database\n", cfdata.svc_name);

	if((scm = OpenSCManager(0, 0, SC_MANAGER_CREATE_SERVICE)))
	{
		if(GetModuleFileName(0, path, sizeof(path)) > 0)
		{
			SC_HANDLE svc;

			if(*cfdata.svc_config)
			{
				char cfa[MAX_PATH+1];

				snprintf(cfa, MAX_PATH, " /CONFIG=%s",
					cfdata.svc_config);

				strncat(path, cfa, MAX_PATH - strlen(path));
			}

			if((svc = CreateService(scm, cfdata.svc_name,
				cfdata.svc_name, SERVICE_ALL_ACCESS,
				SERVICE_WIN32_OWN_PROCESS, SERVICE_AUTO_START,
				SERVICE_ERROR_IGNORE, path, 0, 0, 0, 0, 0)))
			{
				ret = 0;

				CloseServiceHandle(svc);
			}
			else
			{
				fprintf(stderr, "Failed to create service (%s)\n",
				winerror(GetLastError()));
			}
		}
		else
		{
			fprintf(stderr, "Failed to get module filename (%s)\n",
				winerror(GetLastError()));
		}

		CloseServiceHandle(scm);
	}
	else
	{
		fprintf(stderr, "Failed to open SCManager (%s)\n",
			winerror(GetLastError()));
	}

	/*
	 * If something has failed, return now.
	 */

	if(ret)
		return ret;

	/*
	 * Reset return value.
	 */

	ret = -1;

	/*
	 * Cut off the executable name from the path, to obtain a
	 * base directory for locating associated files.
	 */

	if((s = strrchr(path, '\\')))
	{
		/*
		 * Append service config name to base path
		 * if it is specified.
		 */

		if(*cfdata.svc_config)
		{
			snprintf(s, sizeof(path)-(s - &path[0]),
				"\\%s", cfdata.svc_config);
		}
		else
		{
			*s = 0;
		}
	}
	else
	{
		fprintf(stderr, "Failed to obtain base directory from "
			" '%s'", path);
		return ret;
	}

	/*
	 * Open the service key, add a service description and parameters.
	 */

	snprintf(keyname, sizeof(keyname), "%s\\%s", SVC_KEY,
		cfdata.svc_name);

	fprintf(stderr, "Adding parameters to 'HKLM\\%s'\n", keyname);

	if(!(dwRet = RegOpenKeyEx(HKEY_LOCAL_MACHINE, keyname, 0, KEY_ALL_ACCESS, &key)))
	{
		if(!RegSetValueEx(key, "Description", 0, REG_SZ, SVC_DESC,
			strlen(SVC_DESC)+1))
		{
			HKEY hP;

			if(!RegCreateKeyEx(key, "Parameter", 0, NULL, 0,
				KEY_ALL_ACCESS, NULL, &hP, NULL))
			{
				uint32_t val;
				char *str;

				ret = 0; /* OK if nothing fails now */

				val = snprintf(cfdata.basedir,
					sizeof(cfdata.basedir), "%s\\", path);

				ret |= RegSetValueEx(hP, "BaseDirectory", 0,
					REG_SZ, cfdata.basedir, val+1);

				ret |= RegSetValueEx(hP, "HostName", 0,
					REG_SZ, cfdata.hostname,
					strlen(cfdata.hostname)+1);

				str = iptostr(cfdata.bind_addr);
				ret |= RegSetValueEx(hP, "BindAddress", 0,
					REG_SZ, str, strlen(str) + 1);

				ret |= RegSetValueEx(hP, "Port", 0, REG_DWORD,
					(BYTE *)&cfdata.port, sizeof(DWORD));

				val = (cfdata.flags & CF_LOG_FILE) ? 1 : 0;
				ret |= RegSetValueEx(hP, "LogEnable", 0,
					REG_DWORD, (BYTE *)&val,
					sizeof(DWORD));

				ret |= RegSetValueEx(hP, "LogFile", 0,
					REG_SZ, cfdata.logfile,
					strlen(cfdata.logfile)+1);

				if(cfdata.max_pending != DEFAULT_MAX_PENDING)
				{
					ret |= RegSetValueEx(hP, "MaxPending",
						0, REG_DWORD,
						(BYTE *)&cfdata.max_pending,
						sizeof(DWORD));
				}

				if(cfdata.lastlog_recs != DEFAULT_LASTLOG_ENTRIES)
				{
					ret |= RegSetValueEx(hP,
						"LastlogRecords", 0, REG_DWORD,
						(BYTE *)&cfdata.lastlog_recs,
						sizeof(DWORD));
				}

				RegCloseKey(hP);

				if(ret)
				{
					fprintf(stderr, "Failed to add values "
						"to key 'HKLM\\%s\\Parameter' "
						"(%s)\n", keyname,
						winerror(ret));
				}
			}
			else
			{
				fprintf(stderr, "Failed to create Parameter "
					"key in 'HKLM\\%s' (%s)\n", keyname,
					winerror(GetLastError()));
			}
		}
		else
		{
			fprintf(stderr, "Failed to add value to key "
				"'HKLM\\%s' (%s)\n", keyname,
				winerror(GetLastError()));
		}

		RegCloseKey(key);
	}
	else
	{
		fprintf(stderr, "Failed to open key 'HKLM\\%s' (%s)",
			keyname, winerror(dwRet));
	}

	/*
	 * If something has failed, return now.
	 */

	if(ret)
		return ret;

	/*
	 * Reset status.
	 */

	ret = -1;

	/*
	 * Check if the subdirectory where all data for this service
	 * will be stored to already exists, or create it if necessary.
	 */

	if(stat(path, &st))
	{
		if(errno == ENOENT)
		{
			fprintf(stderr, "Creating service base directory "
				"'%s'\n", path);

			if(!mkdir(path))
			{
				ret = 0;
			}
			else
			{
				fprintf(stderr, "Failed to create directory "
					"'%s' (%s)\n", path, strerror(errno));
			}
		}
		else
		{
			fprintf(stderr, "Failed to stat '%s' (%s)\n",
				path, strerror(errno));
		}
	}
	else
	{
		if(S_ISDIR(st.st_mode))
		{
			fprintf(stderr, "Using existing base directory '%s'\n",
				path);

			ret = 0;
		}
		else
		{
			fprintf(stderr, "Cannot create directory '%s' "
				"(File Exists)\n", path);
		}
	}

	/*
	 * If something has failed, return now.
	 */

	if(ret)
		return ret;

	/*
	 * Check if a user database file already exists or create one.
	 */

	if(!(ret = chdir(path)))
	{
		if(!stat(fn_userdb, &st))
		{
			fprintf(stderr, "Using existing user database file "
				"'%s'\n", fn_userdb);
		}
		else
		{
			fprintf(stderr, "No existing user database file "
				"'%s'\n\nPlease create it now.\n", fn_userdb);

			ret = userdb_create();
		}
	}

	return ret;
}

/*
 * Uninstall the service from the ServiceControlManager database.
 */

int UninstallService()
{
	int		ret = -1;
	SC_HANDLE	scm;

	if((scm = OpenSCManager(0, 0, SC_MANAGER_CONNECT)))
	{
		SC_HANDLE svc;

		if((svc = OpenService(scm, cfdata.svc_name,
			SERVICE_QUERY_STATUS | DELETE)))
		{
			SERVICE_STATUS ss;

			if(QueryServiceStatus(svc, &ss))
			{
				if(ss.dwCurrentState == SERVICE_STOPPED)
				{
					DeleteService(svc);
					fprintf(stderr, "Service '%s' deleted\n",
						cfdata.svc_name);
					ret = 0;
				}
				else
				{
					fprintf(stderr, "The service '%s' is "
						"currently running and may "
						"not be deleted\n",
						cfdata.svc_name);
				}
			}
			else
			{
				fprintf(stderr, "Failed to query status of "
					"service '%s' (%s)\n", cfdata.svc_name,
					winerror(GetLastError()));
			}

			CloseServiceHandle(svc);
		}
		else
		{
			fprintf(stderr, "Failed to open service '%s' (%s)\n",
				cfdata.svc_name, winerror(GetLastError()));
		}

		CloseServiceHandle(scm);
	}
	else
	{
		fprintf(stderr, "Failed to open SCManager (%s)\n",
			winerror(GetLastError()));
	}

	return ret;
}

/*
 * LogMessage - Log a status/error message to the Windows NT system log.
 */

void LogMessage(int pri, TCHAR *fmt, ...)
{
	va_list	ap;
	HANDLE	hEventSource;

	va_start(ap, fmt);

	if((hEventSource = RegisterEventSource(NULL, cfdata.svc_name)))
	{
		WORD	winpri;
		TCHAR	msg[512];
		LPCTSTR	szMsg;

		vsnprintf(msg, sizeof(msg), fmt, ap);
		szMsg = msg;

		switch(pri)
		{
			case VLOG_EMERG:
			case VLOG_ALERT:
			case VLOG_CRIT:
			case VLOG_ERR:		winpri = EVENTLOG_ERROR_TYPE;
						break;
			case VLOG_WARNING:	winpri = EVENTLOG_WARNING_TYPE;
						break;
			case VLOG_NOTICE:
			case VLOG_INFO:
			case VLOG_DEBUG:
			default:		winpri = EVENTLOG_INFORMATION_TYPE;
						break;
		}

		ReportEvent
		(
			hEventSource,	/* EventLog handle	*/
			winpri,		/* Event type		*/
			0,		/* Event Category	*/
			1000,		/* Event Identifier	*/
			NULL,		/* No user security ID	*/
			1,		/* #substitution strings*/
			0,		/* No data		*/
			&szMsg,		/* Pointer to strings	*/
			NULL		/* No data		*/
		);

		DeregisterEventSource(hEventSource);
	}

	va_end(ap);
}

/*
 * Format a windows error message and return a pointer
 * to a static buffer containing the resulting error
 * message string.
 */

char *winerror(DWORD code)
{
	static	char	msgbuf[MAX_PATH]	= "\0";
	register char *s;

	FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,	/* System error msg */
		NULL,					/* Message source   */
		code,					/* Error code       */
		0,					/* Language ID      */
		msgbuf,					/* Message buffer   */
		MAX_PATH, NULL);			/* Buffer len       */

	for( ;(s = strrchr(msgbuf, '\n')) != NULL; *s = 0); 
	for( ;(s = strrchr(msgbuf, '\r')) != NULL; *s = 0);

	return msgbuf;
}
#endif /* ----------------------------> WIN32 <---------------------------- */
